/*
 *  Copyright (c) 2019 NetApp
 *  All rights reserved
 */

package nvc

import (
	"context"
	"fmt"
	"math"
	"math/rand"
	"reflect"
	"regexp"
	"strconv"
	"strings"
	"time"

	"github.com/ghodss/yaml"
	"github.com/prometheus/client_golang/prometheus"
	uuid "github.com/satori/go.uuid"
	log "github.com/sirupsen/logrus"
	etcdclient "go.etcd.io/etcd/client/v3"
	appsv1 "k8s.io/api/apps/v1"
	corev1 "k8s.io/api/core/v1"

	// API machinery
	"k8s.io/apimachinery/pkg/api/errors"
	"k8s.io/apimachinery/pkg/api/resource"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/labels"
	"k8s.io/apimachinery/pkg/runtime/schema"
	utilruntime "k8s.io/apimachinery/pkg/util/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes"
	"k8s.io/client-go/kubernetes/scheme"
	typedcorev1 "k8s.io/client-go/kubernetes/typed/core/v1"
	appslisters "k8s.io/client-go/listers/apps/v1"
	corelisters "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
	"k8s.io/client-go/tools/record"
	"k8s.io/client-go/util/workqueue"

	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	quarkclientset "quark.netapp.io/volume-controller/pkg/client/clientset/versioned"
	quarkscheme "quark.netapp.io/volume-controller/pkg/client/clientset/versioned/scheme"
	quarkinformers "quark.netapp.io/volume-controller/pkg/client/informers/externalversions"
	quarklisters "quark.netapp.io/volume-controller/pkg/client/listers/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/cloud"
	quarkregional "quark.netapp.io/volume-controller/pkg/nvc/regional"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
	"quark.netapp.io/volume-controller/pkg/util"
)

const (
	volumeObject = iota
	storagePoolObject
	serviceObject
	clusterPeeringObject
	numObjects
	livenessProbeInitialDelaySeconds     = int32(30)
	livenessProbeSecdInitialDelaySeconds = int32(120)
	livenessProbeTimeoutSeconds          = int32(15)
	livenessProbePeriodSeconds           = int32(15)

	maxVolumePathLen = 80

	// SMB limits
	maxNetBiosLen          = 15
	maxDomainLen           = 255
	maxOrganizationUnitLen = 255
	maxPreferredADSiteLen  = 255
)

type adClientSetupError struct {
	baseError error
}

func (e adClientSetupError) Error() string {
	return fmt.Sprintf("failed to allocate ad client load balancer: %v",
		e.baseError)
}

func newADClientSetupError(err error) adClientSetupError {
	return adClientSetupError{baseError: err}
}

type staticIPAllocationError struct {
	addrName  string
	baseError error
}

func (e staticIPAllocationError) Error() string {
	return fmt.Sprintf("failed to allocate static IP for %s: %v", e.addrName,
		e.baseError)
}

func newStaticIPAllocationError(
	addrName string, err error,
) staticIPAllocationError {
	return staticIPAllocationError{addrName: addrName, baseError: err}
}

func isStaticIPAllocationError(err error) bool {
	_, ok := err.(staticIPAllocationError)
	return ok
}

type InvalidKeyError struct {
}

func (e InvalidKeyError) Error() string {
	return "invalid or stale key"
}

type UpgradeInProgressError struct {
}

func (e UpgradeInProgressError) Error() string {
	return "Upgrade currently in progress"
}

type MigrationInProgressError struct {
}

func (e MigrationInProgressError) Error() string {
	return "Migration-spare deployment in progress"
}

type DiskResizeInProgressError struct {
}

func (e DiskResizeInProgressError) Error() string {
	return "Disk Resize in progress"
}

// ControllerMetrics tracks sundry volume metrics
type ControllerMetrics struct {
	ProvisionedVolumes                util.LockingBoolMap
	FailedVolumes                     util.LockingBoolMap
	KnownVolumes                      util.LockingBoolMap
	FailureCount                      util.LockingUint64
	ProvisionDurationSeconds          util.LockingIntMap
	AverageProvisionDurationSeconds   util.LockingInt
	ProvisionedStoragePool            util.LockingBoolMap
	FailedStoragePool                 util.LockingBoolMap
	KnownStoragePool                  util.LockingBoolMap
	FailureCountStoragePool           util.LockingUint64
	ProvisionDurationSecondsSP        util.LockingIntMap
	AverageProvisionDurationSecondsSP util.LockingInt
}

// NewControllerMetrics initializes a ControllerMetrics struct and returns a pointer to the struct
func NewControllerMetrics() *ControllerMetrics {
	return &ControllerMetrics{
		util.NewLockingBoolMap(),
		util.NewLockingBoolMap(),
		util.NewLockingBoolMap(),
		util.NewLockingUint64(),
		util.NewLockingIntMap(),
		util.NewLockingInt(),
		util.NewLockingBoolMap(),
		util.NewLockingBoolMap(),
		util.NewLockingBoolMap(),
		util.NewLockingUint64(),
		util.NewLockingIntMap(),
		util.NewLockingInt(),
	}
}

// Controller is the controller implementation for NetAppVolume resources
type Controller struct {
	ImagePullPolicy         string
	CCPDImageName           string
	DmapImageName           string
	SecdImageName           string
	EtcdImageName           string
	InitContainerImageName  string
	PodrickImageName        string
	StorageImageName        string
	HASpareImageName        string
	NvramSize               int64
	QuarkVolSize            int64
	MaxPvcSize              int64
	MinPvcSize              int64
	MinDataPvcs             int
	EnableMultiAttach       bool
	Cloud                   string
	LoggingImageName        string
	LoggingOutput           string
	LoadBalancer            bool
	IlbSubnet               string
	IlbTCP                  bool
	IlbUDP                  bool
	SupportabilityImageName string
	DarkContent             string
	IlbGlobalAccess         bool
	QuarkVersion            string
	QuarkVariant            string
	Nodevol                 string
	// We track this so we know whether to set up Podrick with debug=true
	Debug bool

	// Use disk packing, no separate nvram disk, always enabled on DARK_CONTENT
	UseDiskPacking bool

	// kubeclientset is a standard kubernetes clientset
	KubeClientset kubernetes.Interface
	// quarkclientset is a clientset for our own API group
	QuarkClientset quarkclientset.Interface

	// EtcdClient to talk to clusterwide etcd
	EtcdClient *etcdclient.Client

	// metrics is a collection of different metrics that NVC maintains.
	metrics *ControllerMetrics

	cfgMapLister         corelisters.ConfigMapLister
	cfgMapSynced         cache.InformerSynced
	deploysLister        appslisters.DeploymentLister
	deploysSynced        cache.InformerSynced
	nodesLister          corelisters.NodeLister
	nodesSynced          cache.InformerSynced
	pvcsLister           corelisters.PersistentVolumeClaimLister
	pvcsSynced           cache.InformerSynced
	volumesLister        quarklisters.NetAppVolumeLister
	volumesSynced        cache.InformerSynced
	serviceLevelLister   quarklisters.NetAppServiceLevelLister
	serviceLevelsSynced  cache.InformerSynced
	servicesLister       corelisters.ServiceLister
	servicesSynced       cache.InformerSynced
	podsLister           corelisters.PodLister
	podsSynced           cache.InformerSynced
	storagePoolLister    quarklisters.NetAppStoragePoolLister
	storagePoolSynced    cache.InformerSynced
	clusterPeeringLister quarklisters.NetAppClusterPeeringLister
	clusterPeeringSynced cache.InformerSynced

	// workqueue is a rate limited work queue. This is used to queue work to be
	// processed instead of performing it as soon as a change happens. This
	// means we can ensure we only process a fixed amount of resources at a
	// time, and makes it easy to ensure we are never processing the same item
	// simultaneously in two different workers.
	workqueue workqueue.RateLimitingInterface
	// recorder is an event recorder for recording Event resources to the
	// Kubernetes API.
	recorder record.EventRecorder

	// Cloud client
	cloudClient cloud.CloudClient

	thinProvisioningEnabled bool
	regionalMeta            quarkregional.RegionalMetadata
	EnableMultiVolume       bool
	EnableCCPD              bool
	EnableServiceMesh       bool

	// knownObjects is a set of objects that have been seen before
	knownObjects []util.LockingStructMap
}

// Used by event handler
type ObjectKey struct {
	ObjType      string `json:"objType"`
	ObjName      string `json:"objName"`
	ObjNamespace string `json:"ObjNamespace"`
}

// Init does the controller initialization
func (c *Controller) Init(kubeInformerFactory informers.SharedInformerFactory,
	nvInformerFactory quarkinformers.SharedInformerFactory,
	namespace string,
) error {
	var err error

	// enable features
	if c.Cloud != quarkconfig.CloudAzure {
		// Azure doesn't support disk resize.
		// Keep thin-provisioning disabled for Azure.
		c.thinProvisioningEnabled = true
	}

	c.metrics = NewControllerMetrics()

	c.knownObjects = make([]util.LockingStructMap, numObjects)
	for i := 0; i < numObjects; i++ {
		c.knownObjects[i] = util.NewLockingStructMap()
	}

	deployInformer := kubeInformerFactory.Apps().V1().Deployments()
	c.deploysLister = deployInformer.Lister()
	c.deploysSynced = deployInformer.Informer().HasSynced

	pvcInformer := kubeInformerFactory.Core().V1().PersistentVolumeClaims()
	c.pvcsLister = pvcInformer.Lister()
	c.pvcsSynced = pvcInformer.Informer().HasSynced

	volumeInformer := nvInformerFactory.Quark().V1alpha1().NetAppVolumes()
	c.volumesLister = volumeInformer.Lister()
	c.volumesSynced = volumeInformer.Informer().HasSynced

	netAppServiceLevelInformer := nvInformerFactory.Quark().V1alpha1().NetAppServiceLevels()
	c.serviceLevelLister = netAppServiceLevelInformer.Lister()
	c.serviceLevelsSynced = netAppServiceLevelInformer.Informer().HasSynced

	serviceInformer := kubeInformerFactory.Core().V1().Services()
	c.servicesLister = serviceInformer.Lister()
	c.servicesSynced = serviceInformer.Informer().HasSynced

	configMapInformer := kubeInformerFactory.Core().V1().ConfigMaps()
	c.cfgMapLister = configMapInformer.Lister()
	c.cfgMapSynced = configMapInformer.Informer().HasSynced

	podInformer := kubeInformerFactory.Core().V1().Pods()
	c.podsLister = podInformer.Lister()
	c.podsSynced = podInformer.Informer().HasSynced

	nodeInformer := kubeInformerFactory.Core().V1().Nodes()
	c.nodesLister = nodeInformer.Lister()
	c.nodesSynced = nodeInformer.Informer().HasSynced

	storagePoolInformer := nvInformerFactory.Quark().V1alpha1().NetAppStoragePools()
	c.storagePoolLister = storagePoolInformer.Lister()
	c.storagePoolSynced = storagePoolInformer.Informer().HasSynced

	// Create event broadcaster
	// Add sample-controller types to the default Kubernetes Scheme so Events can be
	// logged for sample-controller types.
	utilruntime.Must(quarkscheme.AddToScheme(scheme.Scheme))
	log.Info("Creating event broadcaster.")
	eventBroadcaster := record.NewBroadcaster()
	eventBroadcaster.StartLogging(log.Infof)
	eventBroadcaster.StartRecordingToSink(&typedcorev1.EventSinkImpl{Interface: c.KubeClientset.CoreV1().Events("")})
	c.workqueue = workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "NetAppVolumes")
	c.recorder = eventBroadcaster.NewRecorder(scheme.Scheme, corev1.EventSource{Component: quarkconfig.NVCDeploymentName})

	log.Info("Setting up event handlers.")
	volumeInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: func(obj interface{}) {
			c.enqueueNetAppObj(volumeObject, obj)
		},
		UpdateFunc: func(old, new interface{}) {
			newVolume := new.(*quarkv1alpha1.NetAppVolume)
			oldVolume := old.(*quarkv1alpha1.NetAppVolume)

			// Check if we've seen this volume before, and skip it it the version
			// hasn't changed. It's not safe to skip all updates with no version
			// change because if a volume was added before we started up, it looks
			// like an update with no version change.
			_, ok := c.knownObjects[volumeObject].Get(newVolume.Name)
			if ok && newVolume.ResourceVersion == oldVolume.ResourceVersion {
				return
			}
			c.enqueueNetAppObj(volumeObject, newVolume)
		},
		// We need this for metadata tracking for regional volumes.
		DeleteFunc: func(obj interface{}) {
			c.enqueueNetAppObj(volumeObject, obj)
		},
	})
	deployInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newDeploy := new.(*appsv1.Deployment)
			oldDeploy := old.(*appsv1.Deployment)
			if newDeploy.ResourceVersion == oldDeploy.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})
	pvcInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newPvc := new.(*corev1.PersistentVolumeClaim)
			oldPvc := old.(*corev1.PersistentVolumeClaim)
			if newPvc.ResourceVersion == oldPvc.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	serviceInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newPvc := new.(*corev1.Service)
			oldPvc := old.(*corev1.Service)
			if newPvc.ResourceVersion == oldPvc.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	configMapInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newCfg := new.(*corev1.ConfigMap)
			oldCfg := old.(*corev1.ConfigMap)
			if newCfg.ResourceVersion == oldCfg.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	// We currently only inform for pods for the purposes of reconciling the
	// HA spare, but this has the potential to be heavyweight, so throw it in
	// the work queue.
	podInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: c.handleObject,
		UpdateFunc: func(old, new interface{}) {
			newPod := new.(*corev1.Pod)
			oldPod := new.(*corev1.Pod)
			if newPod.ResourceVersion == oldPod.ResourceVersion {
				return
			}
			c.handleObject(new)
		},
		DeleteFunc: c.handleObject,
	})

	if c.EnableCCPD {
		clusterPeeringInformer := nvInformerFactory.Quark().V1alpha1().NetAppClusterPeerings()
		c.clusterPeeringLister = clusterPeeringInformer.Lister()
		c.clusterPeeringSynced = clusterPeeringInformer.Informer().HasSynced

		clusterPeeringInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
			AddFunc: c.handleObject,
			UpdateFunc: func(old, new interface{}) {
				newClusterPeering := new.(*quarkv1alpha1.NetAppClusterPeering)
				oldClusterPeering := old.(*quarkv1alpha1.NetAppClusterPeering)
				if newClusterPeering.ResourceVersion == oldClusterPeering.ResourceVersion {
					return
				}
				c.handleObject(new)
			},
			DeleteFunc: c.handleObject,
		})
	}
	storagePoolInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
		AddFunc: func(obj interface{}) {
			c.enqueueNetAppObj(storagePoolObject, obj)
		},
		UpdateFunc: func(old, new interface{}) {
			newStoragePool := new.(*quarkv1alpha1.NetAppStoragePool)
			oldStoragePool := old.(*quarkv1alpha1.NetAppStoragePool)
			_, ok := c.knownObjects[storagePoolObject].Get(newStoragePool.Name)
			if ok && newStoragePool.ResourceVersion == oldStoragePool.ResourceVersion {
				return
			}
			c.enqueueNetAppObj(storagePoolObject, newStoragePool)
		},
		DeleteFunc: func(obj interface{}) {
			c.enqueueNetAppObj(storagePoolObject, obj)
		},
	})
	_, err = c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(
		context.TODO(),
		quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
	if err != nil && errors.IsNotFound(err) {
		log.Infof("Regional HA configmap %s not found; assuming zonal cluster.",
			quarkconfig.RegionalHAConfigMapName)
		c.regionalMeta = nil
	} else if err != nil {
		return fmt.Errorf("unable to get regional HA configmap: %v", err)
	} else {
		log.Infof("Found regional HA configmap; initializing regional metadata.")
		c.regionalMeta, err = quarkregional.NewRegionalMetadata(
			c.KubeClientset, c.QuarkClientset, namespace, c.Cloud,
		)
		if err != nil {
			return fmt.Errorf("failed to initialize regional metadata: %v", err)
		}
	}

	rand.Seed(time.Now().UTC().UnixNano())
	// Use shared VIP if using ILB is specified with both protocols
	sharedVIP := (c.IlbSubnet != "") && c.IlbTCP && c.IlbUDP
	switch c.Cloud {
	case quarkconfig.CloudGCP:
		c.cloudClient, err = cloud.NewGCPClient(sharedVIP)
		if err != nil {
			return fmt.Errorf("failed to create GCP Client: %s", err.Error())
		}
	case quarkconfig.CloudAzure:
		c.cloudClient, err = cloud.NewAzureClient()
		if _, ok := err.(quarkconfig.AzureClientENVNotSetError); ok {
			// If env is not set we just log a warning and move on
			// currently azure client isn't used in NVC
			log.Warningf("failed to create Azure client with error %v", err)
		} else if err != nil {
			return fmt.Errorf("failed to create Azure Client: %s", err.Error())
		}
	case quarkconfig.CloudCTL, quarkconfig.CloudNone:
		return nil
	default:
		return fmt.Errorf("unsupported cloud")
	}
	return nil
}

func (c *Controller) enqueueNetAppObj(objectType int, obj interface{}) {
	var objectKey ObjectKey

	object, _ := obj.(metav1.Object)

	objectKey.ObjType = reflect.TypeOf(object).String()
	objectKey.ObjName = object.GetName()
	objectKey.ObjNamespace = object.GetNamespace()

	// Mark this key as known in the map so we can ignore future updates
	// without no version change
	c.knownObjects[objectType].Set(objectKey.ObjName, struct{}{})
	if objectKey.ObjType == reflect.TypeOf(&quarkv1alpha1.NetAppVolume{}).String() {
		c.metrics.KnownVolumes.Set(objectKey.ObjName, true)
	} else {
		c.metrics.KnownStoragePool.Set(objectKey.ObjName, true)
	}
	c.workqueue.Add(objectKey)
}

// handleObject will take any resource implementing metav1.Object and attempt
// to find the NetAppVolume resource that 'owns' it. It does this by looking at the
// objects metadata.ownerReferences field for an appropriate OwnerReference.
// It then enqueues that NetAppVolume resource to be processed. If the object does not
// have an appropriate OwnerReference, it will simply be skipped.
func (c *Controller) handleObject(obj interface{}) {
	var (
		object metav1.Object
		ok     bool
	)
	if object, ok = obj.(metav1.Object); !ok {
		tombstone, ok := obj.(cache.DeletedFinalStateUnknown)
		if !ok {
			utilruntime.HandleError(fmt.Errorf("error decoding object, invalid type"))
			return
		}
		object, ok = tombstone.Obj.(metav1.Object)
		if !ok {
			utilruntime.HandleError(fmt.Errorf("error decoding object tombstone, invalid type"))
			return
		}
		log.WithFields(log.Fields{
			"object": object.GetName(),
		}).Info("Recovered deleted object from tombstone.")
	}

	log.WithFields(log.Fields{
		"name": object.GetName(),
	}).Debug("Processing object.")

	if clusterPeer, ok := object.(*quarkv1alpha1.NetAppClusterPeering); ok {
		c.enqueueNetAppObj(clusterPeeringObject, clusterPeer)
	}
	if service, ok := object.(*corev1.Service); ok {
		spName, ok := service.Labels[quarkconfig.StoragePoolLabel]
		if ok &&
			(service.Name == getADDNSTCPLoadBalancerServiceName(
				service.Namespace, spName) ||
				service.Name == getADDNSUDPLoadBalancerServiceName(
					service.Namespace, spName)) {
			c.enqueueNetAppObj(serviceObject, service)
			// Note that we need to sync the SP afterwards to handle deletion
			// properly.
		}
	}
	if ownerRef := metav1.GetControllerOf(object); ownerRef != nil {
		if ownerRef.Kind == "NetAppVolume" {
			volume, err := c.volumesLister.NetAppVolumes(object.GetNamespace()).Get(ownerRef.Name)
			if err != nil {
				log.Debugf("Ignoring orphaned object '%s' of volume '%s'", object.GetSelfLink(), ownerRef.Name)
				return
			}

			// Handle the volume
			c.enqueueNetAppObj(volumeObject, volume)
		} else if ownerRef.Kind == "NetAppStoragePool" {
			storagePool, err := c.storagePoolLister.NetAppStoragePools(object.GetNamespace()).Get(ownerRef.Name)
			if err != nil {
				log.Debugf("Ignoring orphaned object '%s' of storage pool '%s'", object.GetSelfLink(), ownerRef.Name)
				return
			}
			c.enqueueNetAppObj(storagePoolObject, storagePool)
		}
	}
	// For now, do this in front of the work queue; if we queue it, ha-spare
	// creation can end up well behind the creation of other volumes, which is
	// highly undesirable. If this ends up blocking processing, we'll need to
	// reconcile in a separate thread. For now, though, this appears to be fine.
	// Note that because Go defaults entries not present in the map to "", we
	// don't need to check for membership.
	quarkApp := object.GetLabels()[quarkconfig.AppLabel]
	if app := object.GetLabels()["app"]; quarkApp == "quark" ||
		app == "ha-spare" {
		if err := c.reconcileHASpareDeployment(object.GetNamespace(), c.QuarkVariant); err != nil {
			log.Errorf("failed to reconcile HA spares: %s", err.Error())
		}
	}
}

// Run will set up the event handlers for types we are interested in, as well
// as syncing informer caches and starting workers. It will block until stopCh
// is closed, at which point it will shutdown the workqueue and wait for
// workers to finish processing their current work items.
func (c *Controller) Run(threadiness int, stopCh <-chan struct{}) error {
	defer utilruntime.HandleCrash()
	defer c.workqueue.ShutDown()

	// Start the informer factories to begin populating the informer caches
	log.Info("Starting NetApp Volume controller.")

	// Wait for the caches to be synced before starting workers
	log.Info("Waiting for informer caches to sync")
	if c.EnableCCPD {
		if ok := cache.WaitForCacheSync(stopCh,
			c.deploysSynced,
			c.pvcsSynced,
			c.volumesSynced,
			c.serviceLevelsSynced,
			c.servicesSynced,
			c.cfgMapSynced,
			c.podsSynced,
			c.nodesSynced,
			c.storagePoolSynced,
			c.clusterPeeringSynced); !ok {
			return fmt.Errorf("failed to wait for caches to sync")
		}
	} else {
		if ok := cache.WaitForCacheSync(stopCh,
			c.deploysSynced,
			c.pvcsSynced,
			c.volumesSynced,
			c.serviceLevelsSynced,
			c.servicesSynced,
			c.cfgMapSynced,
			c.podsSynced,
			c.nodesSynced,
			c.storagePoolSynced); !ok {
			return fmt.Errorf("failed to wait for caches to sync")
		}
	}

	log.Infof("Starting %v worker(s).", threadiness)
	// Launch configurable number of workers to process NetAppVolume resources
	for i := 0; i < threadiness; i++ {
		go wait.Until(c.runWorker, time.Second, stopCh)
	}

	log.Info("Started workers.")

	RegisterAndStartMetrics(c)
	<-stopCh
	log.Info("Shutting down workers.")
	StopMetricsServer()

	return nil
}

// RegisterAndStartMetrics will register the Controller which implements collector
//
//	in prometheus and start the metrics server.
func RegisterAndStartMetrics(c *Controller) {
	prometheus.MustRegister(c)
	StartMetricsServer()
}

// runWorker is a long-running function that will continually call the
// processNextWorkItem function in order to read and process a message on the
// workqueue.
func (c *Controller) runWorker() {
	for c.processNextWorkItem() {
	}
}

// processNextWorkItem will read a single work item off the workqueue and
// attempt to process it, by calling the syncHandler.
func (c *Controller) processNextWorkItem() bool {
	obj, shutdown := c.workqueue.Get()

	if shutdown {
		return false
	}

	// We wrap this block in a func so we can defer c.workqueue.Done.
	err := func(obj interface{}) error {
		// We call Done here so the workqueue knows we have finished
		// processing this item. We also must remember to call Forget if we
		// do not want this work item being re-queued. For example, we do
		// not call Forget if a transient error occurs, instead the item is
		// put back on the workqueue and attempted again after a back-off
		// period.
		defer c.workqueue.Done(obj)
		var key ObjectKey
		var ok bool
		// We expect strings to come off the workqueue. These are of the
		// form namespace/name. We do this as the delayed nature of the
		// workqueue means the items in the informer cache may actually be
		// more up to date that when the item was initially put onto the
		// workqueue.
		if key, ok = obj.(ObjectKey); !ok {
			// As the item in the workqueue is actually invalid, we call
			// Forget here else we'd go into a loop of attempting to
			// process a work item that is invalid.
			c.workqueue.Forget(obj)
			utilruntime.HandleError(fmt.Errorf("expected string in workqueue but got %#v", obj))
			return nil
		}
		// Run the syncHandler, passing it the namespace/name string of the
		// NetAppVolume resource to be synced.
		if err := c.syncHandler(key); err != nil {
			// Put the item back on the workqueue to handle any transient errors.
			if _, ok := err.(InvalidKeyError); ok {
				log.WithFields(log.Fields{
					"key": key,
				}).Debugf("Ignoring key: %v", err)
				c.workqueue.Forget(obj)
				return nil
			}
			c.workqueue.AddRateLimited(key)

			if key.ObjType == reflect.TypeOf(&quarkv1alpha1.NetAppVolume{}).String() {
				c.metrics.FailureCount.AddOne()
				c.metrics.FailedVolumes.Set(key.ObjNamespace+"/"+key.ObjName, true)
				//update metrics
				incFailureCount()
				incFailedVolumesCount()
			} else {
				// TODO: Treat failed service updates differently?
				c.metrics.FailureCountStoragePool.AddOne()
				c.metrics.FailedStoragePool.Set(key.ObjNamespace+"/"+key.ObjName, true)
				//update metrics
				incFailureCount()
				incFailedStoragePoolCount()
			}
			return fmt.Errorf("error syncing '%s': %s, requeuing", key, err.Error())
		}
		// Once syncHandler is done with its volume, reconcile HA spares.
		if err := c.reconcileHASpareDeployment(key.ObjNamespace, c.QuarkVariant); err != nil {
			log.Errorf("error reconciling HA spare deployments: %s", err.Error())
		}
		// Finally, if no error occurs we Forget this item so it does not
		// get queued again until another change happens.
		if key.ObjType == reflect.TypeOf(&quarkv1alpha1.NetAppVolume{}).String() && !c.metrics.ProvisionedVolumes.Get(key.ObjNamespace+"/"+key.ObjName) {
			volume, err := c.getNetAppVolumeFromKey(key)
			if err != nil {
				if _, ok := err.(InvalidKeyError); ok || errors.IsNotFound(err) {
					log.WithFields(log.Fields{
						"key": key,
					}).Debugf("Ignoring key: %v", err)
					c.workqueue.Forget(obj)
					return nil
				}
				return err
			}
			if online := util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeOnline); online != nil && online.Status == corev1.ConditionTrue {
				provisionTime := int(online.FirstVolumeOnlineTime.Time.Sub(volume.CreationTimestamp.Time).Seconds())
				provisionedVolLen := c.metrics.ProvisionedVolumes.Len()
				c.metrics.AverageProvisionDurationSeconds.Set(
					(c.metrics.AverageProvisionDurationSeconds.Get()*provisionedVolLen + provisionTime) /
						(provisionedVolLen + 1),
				)
				c.metrics.ProvisionedVolumes.Set(key.ObjNamespace+"/"+key.ObjName, true)
				c.metrics.ProvisionDurationSeconds.Set(key.ObjNamespace+"/"+key.ObjName, provisionTime)
				log.WithFields(log.Fields{
					"netappObj":                   key.ObjName,
					"volumeProvisionTimeSeconds":  provisionTime,
					"provisionedVolumesCount":     c.metrics.ProvisionedVolumes.Len(),
					"failedVolumesCount":          c.metrics.FailedVolumes.Len(),
					"knownVolumesCount":           c.metrics.KnownVolumes.Len(),
					"failureCount":                c.metrics.FailureCount.Get(),
					"averageProvisionTimeSeconds": c.metrics.AverageProvisionDurationSeconds.Get(),
				}).Debug("Updated NVC metrics after successfully provisioning a volume.")
				// Update metrics for prometheus scrape
				updateVolumeProvisionTimeSeconds(provisionTime)
				updateAverageProvisionTimeSeconds(c.metrics.AverageProvisionDurationSeconds.Get())
				incProvisionedVolumesCount()
				updateKnownVolumesCount(c.metrics.KnownVolumes.Len())
			}
		}

		if key.ObjType == reflect.TypeOf(&quarkv1alpha1.NetAppStoragePool{}).String() && !c.metrics.ProvisionedStoragePool.Get(key.ObjNamespace+"/"+key.ObjName) {
			storagePool, err := c.NetAppStoragePoolFromKey(key)
			if err != nil {
				if _, ok := err.(InvalidKeyError); ok || errors.IsNotFound(err) {
					log.WithFields(log.Fields{
						"key": key,
					}).Debugf("Ignoring key: %v", err)
					c.workqueue.Forget(obj)
					return nil
				}
				return err
			}
			if provisioned := util.GetNetAppStoragePoolCondition(storagePool, quarkv1alpha1.NetAppStoragePoolOnline); provisioned != nil && provisioned.Status == corev1.ConditionTrue {
				provisionTime := int(provisioned.FirstStoragePoolOnlineTime.Time.Sub(storagePool.CreationTimestamp.Time).Seconds())
				provisionedSPLen := c.metrics.ProvisionedStoragePool.Len()
				c.metrics.AverageProvisionDurationSecondsSP.Set(
					(c.metrics.AverageProvisionDurationSecondsSP.Get()*provisionedSPLen + provisionTime) /
						(provisionedSPLen + 1),
				)
				c.metrics.ProvisionedStoragePool.Set(key.ObjNamespace+"/"+key.ObjName, true)
				c.metrics.ProvisionDurationSecondsSP.Set(key.ObjNamespace+"/"+key.ObjName, provisionTime)
				log.WithFields(log.Fields{
					"netappObj":                       key.ObjName,
					"storagePoolProvisionTimeSeconds": provisionTime,
					"provisionedStoragePoolCount":     c.metrics.ProvisionedStoragePool.Len(),
					"failedStoragePoolCount":          c.metrics.FailedStoragePool.Len(),
					"knownStoragePoolCount":           c.metrics.KnownStoragePool.Len(),
					"failureCountStoragePool":         c.metrics.FailureCountStoragePool.Get(),
					"averageProvisionTimeSecondsSP":   c.metrics.AverageProvisionDurationSecondsSP.Get(),
				}).Debug("Updated NVC metrics after successfully provisioning a StoragePool.")
				// Update metrics for prometheus scrape
				updateStoragePoolProvisionTimeSeconds(provisionTime)
				updateStoragePoolAverageProvisionTimeSeconds(c.metrics.AverageProvisionDurationSecondsSP.Get())
				incProvisionedStoragePoolsCount()
				updateKnownStoragePoolCount(c.metrics.KnownStoragePool.Len())
			}
		}

		c.workqueue.Forget(obj)
		log.WithFields(log.Fields{
			"netappObj": key,
		}).Debug("Successfully synced NetApp Object.")
		return nil
	}(obj)

	if err != nil {
		utilruntime.HandleError(err)
	}

	return true
}

// getNetAppVolumeFromKey retrieves the NetAppVolume object from the work
// queue given a key.
func (c *Controller) getNetAppVolumeFromKey(key ObjectKey) (*quarkv1alpha1.NetAppVolume, error) {
	// Get the NetAppVolume CRD object
	volume, err := c.volumesLister.NetAppVolumes(key.ObjNamespace).Get(key.ObjName)
	if err != nil {
		return nil, err
	}

	return volume, nil
}

// getNetAppStoragePoolFromKey retrieves the NetAppStoragePool object from the work
// queue given a key.
func (c *Controller) NetAppStoragePoolFromKey(key ObjectKey) (*quarkv1alpha1.NetAppStoragePool, error) {
	// Get the NetAppStoragePool CRD object
	storagePool, err := c.storagePoolLister.NetAppStoragePools(key.ObjNamespace).Get(key.ObjName)
	if err != nil {
		return nil, err
	}

	return storagePool, nil
}

// Retrieves the service object given a key from the work queue.
func (c *Controller) serviceFromKey(key ObjectKey) (*corev1.Service, error) {
	service, err := c.servicesLister.Services(key.ObjNamespace).Get(
		key.ObjName)
	if err != nil {
		return nil, err
	}

	return service, nil
}

// Retrieves the clusterPeering object given a key from the work queue.
func (c *Controller) clusterPeeringFromKey(key ObjectKey) (*quarkv1alpha1.NetAppClusterPeering, error) {
	// Get the NetAppClusterPeering CRD object
	clusterPeer, err := c.clusterPeeringLister.NetAppClusterPeerings(key.ObjNamespace).Get(
		key.ObjName)
	if err != nil {
		return nil, err
	}

	return clusterPeer, nil
}

// NetAppStoragePoolFromVolume retrieves the NetAppStoragePool object from the lister
// queue given a volume.
func NetAppStoragePoolFromVolume(lister quarklisters.NetAppStoragePoolLister, volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppStoragePool, error) {
	metaLabels := volume.ObjectMeta.GetLabels()
	storagePoolUUID := metaLabels[quarkconfig.StoragePoolLabel]
	if storagePoolUUID == "" {
		return nil, fmt.Errorf("storagePoolUUID label not found in volume: %s", volume.Name)
	}
	storagePool, err := lister.NetAppStoragePools(volume.Namespace).Get(storagePoolUUID)
	if err != nil {
		return nil, fmt.Errorf("storage pool not found for volume: %s", volume.Name)
	}
	return storagePool, nil
}

// syncHandler compares the actual state with the desired, and attempts to
// converge the two. It then updates the Status block of the NetAppVolume resource
// with the current status of the resource.
func (c *Controller) syncHandler(key ObjectKey) error {
	var (
		volume      *quarkv1alpha1.NetAppVolume
		storagePool *quarkv1alpha1.NetAppStoragePool
		err         error
	)
	switch key.ObjType {
	case reflect.TypeOf(&quarkv1alpha1.NetAppVolume{}).String():
		volume, err = c.getNetAppVolumeFromKey(key)
		if err != nil && errors.IsNotFound(err) {
			log.WithFields(log.Fields{
				"key": key,
			}).Debug("Failed to find volume; assuming deleted.")
			// This may be a legitimate volume that didn't have a finalizer or whose
			// finalizer was removed early for some reason. We need to ensure that
			// regional metadata tracking is up-to-date.
			if c.regionalMeta != nil {
				c.regionalMeta.DeleteNetAppObj(key.ObjName)
			}
			return nil
		} else if err != nil {
			return err
		}
		storagePool, err = NetAppStoragePoolFromVolume(c.storagePoolLister, volume)
		if err != nil {
			return err
		}
		if nil == volume.ObjectMeta.DeletionTimestamp {
			// Sync DNSIP Update first, as delay in DNSIP update can impact datapath.
			// We don't want errors in other administrative operation delay DNS update.
			// Also retrying DNS update can result in multiple pod restart.
			// So need to avoid retry due to other administrative operation failure.

			// With emptydir changes, we do not have to restart the pod as part of updating the dnsIp
			// The call to syncVolumeDNS should be made only for old volume version
			if isPre1_0VersionedVolume(volume) {
				volume, err = c.syncVolumeDNS(volume)
				if err != nil {
					return err
				}
			}
			err = c.syncNetappObj(volume, storagePool)
		} else {
			// if storagePool is present for the volume do nothing, volume delete will be taken care by podrick.
			// delete should only be invoked for non-upgraded volumes
			if storagePool == nil {
				err = c.syncDeletingNetappObj(volume, nil)
			}
		}
		return err
	case reflect.TypeOf(&quarkv1alpha1.NetAppStoragePool{}).String():
		storagePool, err = c.NetAppStoragePoolFromKey(key)
		if err != nil && errors.IsNotFound(err) {
			log.WithFields(log.Fields{
				"key": key,
			}).Debug("Failed to find storage pool; assuming deleted.")
			return nil
		} else if err != nil {
			return err
		}
		if nil == storagePool.ObjectMeta.DeletionTimestamp {
			err = c.syncNetappObj(volume, storagePool)
		} else {
			err = c.syncDeletingNetappObj(volume, storagePool)
		}
		return err
	case reflect.TypeOf(&corev1.Service{}).String():
		service, err := c.serviceFromKey(key)
		if err != nil && errors.IsNotFound(err) {
			log.WithFields(log.Fields{
				"key": key,
			}).Debug("Failed to find service; assuming deleted.")
			return nil
		} else if err != nil {
			return err
		}
		if service.DeletionTimestamp == nil {
			err = c.syncADClientService(service)
		}
		return err
	case reflect.TypeOf(&quarkv1alpha1.NetAppClusterPeering{}).String():
		clusterPeer, err := c.clusterPeeringFromKey(key)
		if err != nil && errors.IsNotFound(err) {
			log.WithFields(log.Fields{
				"key": key,
			}).Debug("Failed to find clusterpeering; assuming deleted.")
			return nil
		} else if err != nil {
			return err
		}
		if clusterPeer.DeletionTimestamp == nil {
			err = c.syncClusterPeering(clusterPeer)
		}
		return err
	}
	return nil
}

func ValidateSMBConfiguration(volumePath, netbios, domain, organizationUnit, preferredADSite string) error {
	err := ValidateVolumePath(volumePath)
	if err != nil {
		return err
	}
	return ValidateADConfiguration(netbios, domain, organizationUnit, preferredADSite)
}

func ValidateVolumePath(volumePath string) error {
	if len(volumePath) > maxVolumePathLen {
		return fmt.Errorf("length of volume path should not be more than %d characters %s", maxVolumePathLen, volumePath)
	}
	// Validate volume mount path
	matched, err := regexp.MatchString(`^[a-zA-Z][a-zA-Z0-9-]*$`, volumePath)
	if !matched || err != nil {
		return fmt.Errorf("invalid value for volume path %s", volumePath)
	}
	return nil
}

func ValidateADConfiguration(netbios, domain, organizationUnit, preferredADSite string) error {
	if len(netbios) > maxNetBiosLen {
		return fmt.Errorf("invalid netbios length, cannot be greater than %d", maxNetBiosLen)
	}
	netbios_invalid_chars := "@#*()=+[]|;:\",<>/?"
	if strings.ContainsAny(netbios, netbios_invalid_chars) {
		return fmt.Errorf("invalid characters in netbios, cannot contain any of %s", netbios_invalid_chars)
	}
	if len(domain) > maxDomainLen {
		return fmt.Errorf("invalid domain length, cannot be greater than %d", maxDomainLen)
	}
	if len(organizationUnit) > maxOrganizationUnitLen {
		return fmt.Errorf("invalid organization unit length, cannot be greater than %d", maxOrganizationUnitLen)
	}
	if len(preferredADSite) > maxPreferredADSiteLen {
		return fmt.Errorf("invalid preferred AD site length, cannot be greater than %d", maxPreferredADSiteLen)
	}

	return nil
}

func ValidateNFSConfiguration(volumePath string) error {
	return ValidateVolumePath(volumePath)
}

func (c *Controller) validateInput(volumeUUID string, volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) error {
	var (
		err          error
		serviceLevel quarkconfig.ServiceLevel = quarkconfig.ServiceLevel(volume.Spec.ServiceLevel)
	)

	if storagePool != nil {
		serviceLevel = quarkconfig.ServiceLevel(storagePool.Spec.ServiceLevel)
	}

	// Validate service level
	switch serviceLevel {
	case quarkconfig.ServiceLevelStandard, quarkconfig.ServiceLevelRegionalStandard:
	default:
		return fmt.Errorf("invalid value for service level %s", serviceLevel)
	}

	// Validate UUID
	_, err = uuid.FromString(volumeUUID)
	if err != nil {
		return fmt.Errorf("invalid value for UUID %s", volumeUUID)
	}

	if volume.Spec.Protocol == "cifs" {
		err = ValidateSMBConfiguration(volume.Spec.VolumePath, volume.Spec.SMBConfiguration.Netbios, volume.Spec.SMBConfiguration.Domain, volume.Spec.SMBConfiguration.OrganizationUnit, volume.Spec.SMBConfiguration.PreferredADSite)
		if err != nil {
			return err
		}
	} else {
		err = ValidateNFSConfiguration(volume.Spec.VolumePath)
		if err != nil {
			return err
		}
	}

	if volume.Spec.Protocol != "cifs" && volume.Spec.Protocol != "nfs3" && volume.Spec.Protocol != "nfs4" && volume.Spec.Protocol != "" {
		return fmt.Errorf("invalid value for protocol %s", volume.Spec.Protocol)
	}
	return nil
}

func (c *Controller) makeVolumeService(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (string, error) {
	var (
		svc             *corev1.Service
		exportAddress   string
		namespace, name = GetNamespaceNameForNetappObj(volume, storagePool)
		protocol        string
	)
	if storagePool != nil {
		protocol = "all"
	} else {
		protocol = volume.Spec.Protocol
	}

	svc, err := c.servicesLister.Services(namespace).Get(getServiceName(namespace, name))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the service doesn't exist yet, create it
			svc, err = c.KubeClientset.CoreV1().Services(namespace).Create(context.TODO(), c.makeService(volume, storagePool, protocol), metav1.CreateOptions{})
			if err == nil {
				log.WithFields(log.Fields{
					"netappObj": name,
					"service":   getServiceName(namespace, name),
				}).Info("Created service for NetappObj.")
			} else {
				return exportAddress, fmt.Errorf("failed to create service for NetappObj: %s", err.Error())
			}
		} else {
			return exportAddress, fmt.Errorf("failed to get service for NetappObj: %s", err.Error())
		}
	}

	for exportAddress == "" {
		exportAddress = svc.Spec.ClusterIP
		log.WithFields(log.Fields{
			"netappObj":     name,
			"exportAddress": exportAddress,
			"service":       getServiceName(namespace, name),
		}).Debug("Updated service.")
		time.Sleep(1 * time.Second)
		svc, err = c.servicesLister.Services(namespace).Get(getServiceName(namespace, name))
		if err != nil {
			return exportAddress, fmt.Errorf("failed to find service for netappObj: %s", err.Error())
		}
	}
	return exportAddress, nil
}

// Create a k8s service for quark pods that have cluster peering
// The service type is ClusterIP, which is the default, so not bothering to set it here.
// NOTE:  Remove volume from getSelectorNetappObj() & getLabelForNetappObj() and make sure to change it in places where
// it is called
func (c *Controller) makeCRRService(storagePool *quarkv1alpha1.NetAppStoragePool) *corev1.Service {
	var ports = []corev1.ServicePort{
		corev1.ServicePort{
			Name:     "dsmf",
			Protocol: "TCP",
			Port:     11104,
		},
		corev1.ServicePort{
			Name:     "csm",
			Protocol: "TCP",
			Port:     11105,
		},
		corev1.ServicePort{
			Name:     "csm2",
			Protocol: "TCP",
			Port:     7700,
		},
	}

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(nil, storagePool),
		Spec: corev1.ServiceSpec{
			Ports:                    ports,
			Selector:                 getSelectorNetappObj(nil, storagePool),
			PublishNotReadyAddresses: true,
		},
	}
	service.Name = quarkconfig.GetClusterPeeringServiceName(storagePool.Name)
	service.Labels = c.getLabelForNetappObj(nil, storagePool)

	return service
}

func GetNamespaceNameForNetappObj(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (string, string) {
	if storagePool != nil {
		return storagePool.Namespace, storagePool.Name
	}
	return volume.Namespace, volume.Name
}

func (c *Controller) syncADClientService(service *corev1.Service) error {
	var (
		tcpADService, udpADService *corev1.Service
		err                        error
	)

	log.WithFields(log.Fields{
		"service":     service.Name,
		"storagePool": service.Labels[quarkconfig.StoragePoolLabel],
	}).Debug("Got service update")
	spName, ok := service.Labels[quarkconfig.StoragePoolLabel]
	if !ok {
		// Not a storage pool service; this should never happen
		return fmt.Errorf("processed service without a storage pool label: %v;"+
			" this should never happen", service.Name)
	}
	if service.Name == getADDNSTCPLoadBalancerServiceName(service.Namespace,
		spName) {
		tcpADService = service
		udpADService, err = c.servicesLister.Services(service.Namespace).Get(
			getADDNSUDPLoadBalancerServiceName(service.Namespace, spName),
		)
		if err != nil && errors.IsNotFound(err) {
			// If the UDP service doesn't exist yet, this isn't an error;
			// just return nil, since it may not have been created yet
			return nil
		} else if err != nil {
			return fmt.Errorf("failed to get UDP AD/DNS service for SP %s: %v",
				spName, err)
		}
	} else if service.Name == getADDNSUDPLoadBalancerServiceName(
		service.Namespace, spName) {
		udpADService = service
		tcpADService, err = c.servicesLister.Services(service.Namespace).Get(
			getADDNSTCPLoadBalancerServiceName(service.Namespace, spName),
		)
		if err != nil && errors.IsNotFound(err) {
			// If the UDP service doesn't exist yet, this isn't an error;
			// just return nil, since it may not have been created yet
			return nil
		} else if err != nil {
			return fmt.Errorf("failed to get TCP AD/DNS service for SP %s: %v",
				spName, err)
		}
	} else {
		// the service isn't relevant
		return fmt.Errorf("got a non-AD/DNS service: %v; this should never "+
			"happen", service.Name)
	}
	if len(tcpADService.Status.LoadBalancer.Ingress) == 0 ||
		len(udpADService.Status.LoadBalancer.Ingress) == 0 {
		// One or both of the services has yet to come online
		return nil
	}
	tcpService, err := c.servicesLister.Services(service.Namespace).Get(
		GetTCPLoadBalancerServiceName(service.Namespace, spName),
	)
	if err != nil {
		return fmt.Errorf("failed to get tcp svc for SP %s: %v", spName, err)
	}
	udpService, err := c.servicesLister.Services(service.Namespace).Get(
		GetUDPLoadBalancerServiceName(service.Namespace, spName),
	)
	if err != nil {
		return fmt.Errorf("failed to get udp svc for SP %s: %v", spName, err)
	}

	// Get the storage pool here to see if global access is being aborted.
	//
	// For the sake of correctness and to avoid racing with enablement, we need
	// to do this after getting the services. By using older versions of the
	// services here, we ensure that any race with abort will result in a
	// stale version conflict, with abort winning. If we check for abort
	// before getting the services, there is a chance that we could get the
	// AD/DNS services before they are deleted, but the main services after
	// they have been updated, allowing updates to proceed without issues.
	// By getting all services up front, we avoid this problem.
	//
	// We also want to get the storage pool from the server, not from the
	// cache, to avoid any possible issues there (this may be overkill, but
	// this functionality is difficult to test, so the safer option makes more
	// sense here).
	sp, err := c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(
		service.Namespace).Get(context.TODO(), spName, metav1.GetOptions{})
	if err != nil {
		return fmt.Errorf("unable to get storage pool %s: %v", spName, err)
	}
	// Note that this will catch every case in which
	// sp.Spec.GlobalAccesssAborting is true.
	if !sp.Spec.GlobalAccess && !sp.Status.GlobalAccess {
		return nil
	}

	log.WithFields(log.Fields{
		"storagePool":  spName,
		"aborting":     sp.Spec.GlobalAccessAborting,
		"globalAccess": sp.Spec.GlobalAccess,
	}).Info("Updating main service annotations.")
	// Remove the snat annotations for AD/DNS traffic from the main services, if
	// necessary. Note that this will return without updating if the services
	// are already updated.
	err = c.updateServiceAnnotations(tcpService,
		quarkconfig.SNATProtocolTCP,
		quarkconfig.SNATActiveOpenNFSTCPPortRange)
	if err != nil {
		// formatting is taken care of in the function
		return err
	}
	err = c.updateServiceAnnotations(udpService,
		quarkconfig.SNATProtocolUDP,
		quarkconfig.SNATActiveOpenNFSUDPPortRange)
	if err != nil {
		return err
	}

	// Add the annotations for AD/DNS traffic to the auxiliary services
	err = c.updateServiceAnnotations(tcpADService,
		quarkconfig.SNATProtocolTCP,
		quarkconfig.SNATActiveOpenCIFSTCPPortRange)
	if err != nil {
		return err
	}
	err = c.updateServiceAnnotations(udpADService,
		quarkconfig.SNATProtocolUDP,
		quarkconfig.SNATActiveOpenCIFSUDPPortRange)
	if err != nil {
		return err
	}

	globalAccess := c.cloudClient.GlobalAccessEnabled(tcpADService) &&
		c.cloudClient.GlobalAccessEnabled(udpADService)
	adClientIP := tcpADService.Spec.LoadBalancerIP // tcp and UDP should be equal
	log.WithFields(log.Fields{
		"sp":                   sp.Name,
		"spResourceVersion":    sp.ResourceVersion,
		"globalAccess":         globalAccess,
		"globalAccessAborting": sp.Spec.GlobalAccessAborting,
		"adClientIP":           adClientIP,
		"svc":                  service.Name,
	}).Debug("Got AD service update for storage pool.")
	if sp.Status.GlobalAccess == globalAccess &&
		sp.Status.ADClientIP == adClientIP {
		// Either there's nothing to update or we're aborting global access
		// enablement, so don't update the status. If we're aborting, the abort
		// workflow will clean up the snat annotations, so we don't need to do
		// that here
		return nil
	}
	spCopy := sp.DeepCopy()
	spCopy.Status.GlobalAccess = globalAccess
	spCopy.Status.ADClientIP = adClientIP
	_, err = c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(
		spCopy.Namespace).UpdateStatus(context.TODO(), spCopy,
		metav1.UpdateOptions{})
	if err != nil {
		return fmt.Errorf("failed to update storage pool %s global "+
			"access: %v", spName, err)
	}
	log.WithFields(log.Fields{
		"storagePool":             spName,
		"previousResourceVersion": sp.ResourceVersion,
		"globalAccess":            globalAccess,
		"adClientIP":              adClientIP,
		"tcpService":              tcpService.Name,
		"udpService":              udpService.Name,
	}).Debug("Global access for storage pool updated.")
	msg := quarkconfig.MessageResourceGlobalAccessEnabled
	if !globalAccess {
		msg = quarkconfig.MessageResourceGlobalAccessDisabled
	}
	c.recorder.Event(sp, corev1.EventTypeNormal,
		quarkconfig.SuccessSynced,
		msg)
	return nil
}

func (c *Controller) syncClusterPeering(clusterPeer *quarkv1alpha1.NetAppClusterPeering) error {
	// Create the CRR ClusterPeering service if needed
	spName := clusterPeer.Labels[quarkconfig.StoragePoolLabel]
	storagePool, err := c.storagePoolLister.NetAppStoragePools(clusterPeer.GetNamespace()).Get(spName)
	if err != nil {
		return err
	}

	serviceName := quarkconfig.GetClusterPeeringServiceName(spName)
	_, err = c.servicesLister.Services(clusterPeer.Namespace).Get(serviceName)
	if err != nil {
		if errors.IsNotFound(err) {
			// The service needs to be created.
			_, err = c.KubeClientset.CoreV1().Services(clusterPeer.Namespace).Create(context.TODO(), c.makeCRRService(storagePool), metav1.CreateOptions{})
			if err != nil {
				log.WithFields(log.Fields{
					"error": err.Error(),
				}).Debug("ClusterPeeringService create failed.")
				return err
			}
		} else {
			return err
		}
	}
	return nil
}

func (c *Controller) deleteClusterPeeringService(storagePool *quarkv1alpha1.NetAppStoragePool) error {
	// Delete the CRR ClusterPeering service if it exists

	serviceName := quarkconfig.GetClusterPeeringServiceName(storagePool.Name)
	_, err := c.servicesLister.Services(storagePool.Namespace).Get(serviceName)
	if err != nil {
		if errors.IsNotFound(err) {
			return nil
		} else {
			return err
		}
	}
	// The service needs to be created.
	err = c.KubeClientset.CoreV1().Services(storagePool.Namespace).Delete(context.TODO(), serviceName, metav1.DeleteOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"error": err.Error(),
		}).Debug("ClusterPeeringService delete failed.")
		return err
	}
	return nil
}

func (c *Controller) syncNetappObj(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) error {
	var (
		err                      error
		updateStatus             = false
		quarkUUID                string
		quarkNvramID             string
		volumeUUID               string
		ownerID                  string
		vserverUUID              string
		nbladeUUID               string
		volumeStateEvent         *corev1.Event
		totalActualSize          int64
		desiredSize              int64
		diskPackingEnabled       bool
		failureDomain            = ""
		nodeSelector             = map[string]string{}
		namespace, name          = GetNamespaceNameForNetappObj(volume, storagePool)
		migSpareName             = GetQuarkMigrationSpareDeployName(name)
		conditions               []quarkv1alpha1.NetAppVolumeCondition
		spConditions             []quarkv1alpha1.NetAppStoragePoolCondition
		updateNodeSelector       bool
		svl                      string
		specSize                 int64
		statusProvisionedSize    int64
		specProvisionedSize      int64
		exportAddress            string
		isPreVol                 bool
		specZone                 string
		specAvailableZones       []string
		expectedAvailableZones   = quarkconfig.AvailableZonesGCP
		specQuarkVersion         string
		statusQuarkVersion       string
		connectionTrackingPolicy string
	)
	if storagePool != nil {
		quarkUUID = storagePool.Status.QuarkUUID
		quarkNvramID = storagePool.Status.QuarkNvramID
		vserverUUID = storagePool.Status.QuarkVserverUUID
		nbladeUUID = storagePool.Status.QuarkNbladeUUID
		svl = storagePool.Spec.ServiceLevel
		specSize = storagePool.Spec.Size.Value()
		// Assuming the storagePool.Spec.Zone and storagePool.Spec.AvailableZones
		// cannot be wrong.
		specZone = storagePool.Spec.Zone
		specAvailableZones = storagePool.Spec.AvailableZones
		if svl == string(quarkconfig.ServiceLevelRegionalStandard) {
			switch c.Cloud {
			case quarkconfig.CloudGCP:
				expectedAvailableZones = quarkconfig.AvailableZonesGCP
			case quarkconfig.CloudAzure:
				expectedAvailableZones = quarkconfig.AvailableZonesAzure
			}
			if !(len(specAvailableZones) == expectedAvailableZones || len(specAvailableZones) == 0) {
				// This should not happen
				return fmt.Errorf("incorrect number of zones specified for available "+
					"zones, expected %d, got %d", expectedAvailableZones,
					len(specAvailableZones))
			}
		}
		if specSize != storagePool.Status.Size.Value() && (specSize%quarkconfig.GiB) != 0 {
			errMsg := fmt.Sprintf("Storage pool size should be a mulitple of %d", quarkconfig.GiB)
			c.recorder.Event(storagePool, corev1.EventTypeWarning, "ErrStoragePoolSize", errMsg)
			return fmt.Errorf(errMsg)
		}
		exportAddress = storagePool.Status.ExportAddress
		metaLabels := storagePool.ObjectMeta.GetLabels()
		ownerID = metaLabels[quarkconfig.OwnerIDLabel]
		if ownerID == "" {
			errMsg := "storagePool OwnerID not found"
			c.recorder.Event(storagePool, corev1.EventTypeWarning, "ErrStoragePoolOwnerId", errMsg)
			return fmt.Errorf(errMsg)
		}
		specQuarkVersion = storagePool.Spec.QuarkVersion
		statusQuarkVersion = storagePool.Status.QuarkVersion
		spConditions = storagePool.Status.Conditions

		if c.EnableCCPD {
			// Create the CRR ClusterPeering service if needed
			clusterPeering, err := NetappClusterPeeringsByStoragePool(c.clusterPeeringLister, namespace, storagePool.Name)
			if err != nil {
				return err
			}
			if len(clusterPeering) > 0 {
				err = c.syncClusterPeering(clusterPeering[0])
				if err != nil {
					return err
				}
			} else {
				// Clean up any left over services if there are not cluster peerings
				err = c.deleteClusterPeeringService(storagePool)
				if err != nil {
					return err
				}
			}
		}
	} else {
		quarkUUID = volume.Status.QuarkUUID
		quarkNvramID = volume.Status.QuarkNvramID
		volumeUUID = volume.Status.VolumeUUID
		svl = volume.Spec.ServiceLevel
		specSize = volume.Spec.Size.Value()
		exportAddress = volume.Status.ExportAddress
		if specSize != volume.Status.Size.Value() && (specSize%quarkconfig.GiB) != 0 {
			errMsg := fmt.Sprintf("Volume size should be a mulitple of %d", quarkconfig.GiB)
			c.recorder.Event(volume, corev1.EventTypeWarning, "ErrVolumeSize", errMsg)
			return fmt.Errorf(errMsg)
		}
		isPreVol = isPre1_0VersionedVolume(volume)
		specQuarkVersion = volume.Spec.QuarkVersion
		statusQuarkVersion = volume.Status.QuarkVersion
	}

	updateUUIDs := false
	if quarkUUID == "" {
		quarkUUID = uuid.NewV1().String()
		updateUUIDs = true
	}
	if quarkNvramID == "" {
		quarkNvramID = c.generateNvramID()
		updateUUIDs = true
	}
	if volumeUUID == "" && volume != nil {
		volumeUUID = volume.Name
		updateUUIDs = true
	}
	if vserverUUID == "" {
		vserverUUID = uuid.NewV1().String()
		updateUUIDs = true
	}
	if nbladeUUID == "" {
		nbladeUUID = uuid.NewV1().String()
		updateUUIDs = true
	}

	if volume != nil {
		err = c.validateInput(volumeUUID, volume, storagePool)
		if err != nil {
			c.recorder.Event(volume, corev1.EventTypeWarning, "ErrSyncVolume", err.Error())
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"error":  err.Error(),
			}).Error("Volume failed input validation.")
			return nil
		}

		metaLabels := volume.ObjectMeta.GetLabels()
		ownerID = metaLabels[quarkconfig.OwnerIDLabel]
		if ownerID == "" {
			errMsg := "volume OwnerID not found"
			c.recorder.Event(volume, corev1.EventTypeWarning, "ErrVolumeOwnerId", errMsg)
			return fmt.Errorf(errMsg)
		}
	}

	if updateUUIDs {
		volume, storagePool, err = c.updateNetAppObjectUUIDStatus(volume, storagePool, quarkUUID, quarkNvramID, volumeUUID, vserverUUID, nbladeUUID)
		if err != nil {
			log.WithFields(log.Fields{
				"netappObj": name,
			}).Error("UpdateNetAppObjectUUIDStatus Failed.")
			return err
		}
	}
	serviceLevel, err := c.QuarkClientset.QuarkV1alpha1().NetAppServiceLevels(
		namespace).Get(context.TODO(), svl, metav1.GetOptions{})
	if err != nil {
		var errMsg string
		if errors.IsNotFound(err) {
			errMsg = fmt.Sprintf("unknown service level provided for netappObj: %s",
				svl)
		} else {
			errMsg = fmt.Sprintf("unable to retrieve service level: %s %v",
				svl, err)
		}
		if storagePool != nil {
			c.recorder.Event(storagePool, corev1.EventTypeWarning, "ErrServiceLevel", errMsg)
		} else {
			c.recorder.Event(volume, corev1.EventTypeWarning, "ErrServiceLevel", errMsg)
		}
		return fmt.Errorf(errMsg)
	}

	if serviceLevel.Spec.RightSize {
		nodeSelector, err = getServiceLevelNodeSelector(specSize, serviceLevel)
		if err != nil {
			return err
		}
	}

	if specZone != "" {
		nodeSelector[quarkconfig.ZoneTopologyKey] = specZone
	}

	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(namespace).List(c.getDataPvcSelector(volume, storagePool))
	if err != nil {
		return fmt.Errorf("failed to list PVCs: %s", err.Error())
	}
	dataCount := len(dataPvcs)

	// This is the earliest point at which we can tell if the volume has already
	// been assigned a failure domain if we don't already know about it.
	// TODO: Check to see if this can be moved into a function
	if c.regionalMeta != nil {
		failureDomain, err = c.regionalMeta.ProcessNewNetAppObj(c.KubeClientset,
			name, serviceLevel.Spec.Regional, dataPvcs, specAvailableZones)
		if err != nil {
			return fmt.Errorf("unable to determine failure domain for netappObj %s: %v",
				failureDomain, err)
		}
		log.WithFields(log.Fields{
			"failureDomain": failureDomain,
			"netappObj":     name,
		}).Debug("Got failure domain for netappObj.")
		// TODO: Get rid of this unless creating a volume? It's redundant with
		// what's going on in NHC, and it'd save us some effort
		haCM, err := c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(
			context.TODO(), quarkconfig.RegionalHAConfigMapName, metav1.GetOptions{})
		if err != nil {
			return fmt.Errorf("unable to get regional HA configmap; unable " +
				"to determine active zone")
		}
		if storagePool != nil && storagePool.Spec.Zone != "" && haCM.Data[storagePool.Status.Zone] == quarkconfig.RegionalHAActiveZone {
			nodeSelector[quarkconfig.ZoneTopologyKey] = storagePool.Spec.Zone
		}
		fdZones := util.GetFailureDomainZones(failureDomain)
		if len(fdZones) > 1 {
			needsSelector := false
			validZone := ""
			for _, z := range fdZones {
				if haCM.Data[z] == quarkconfig.RegionalHAActiveZone {
					// We can schedule in this zone
					validZone = z
				} else {
					// There's at least one failed zone in the failure domain, so need
					// a node selector if it's a volume or the SP hasn't yet landed in
					// a zone. If it *has*, the NHC will take care of it if it hasn't
					// already.
					needsSelector = storagePool == nil || storagePool.Status.Zone == ""
				}
			}
			if needsSelector && validZone != "" {
				nodeSelector[quarkconfig.ZoneTopologyKey] = validZone
			} else if needsSelector {
				log.WithFields(log.Fields{
					"failureDomain": failureDomain,
					"netappObj":     name,
				}).Warn("No active zone found for volume. Proceeding, but netappObj " +
					"may not be runnable.")
				// TODO: Return an error here instead? Currently erring on the side
				// of creating or updating the volume, regardless, so that when
				// zonal outages recover, we won't have lost any updates.
			}
		}
	} else {
		log.WithFields(log.Fields{
			"netappObj": name,
		}).Debug("Skipping failure domain for zonal volume.")
	}

	if serviceLevel.Spec.RightSize || (c.regionalMeta != nil && storagePool != nil) {
		// check if the deployment definition has correct node selector, skip migration if it does
		updateNodeSelector, err = c.needDeploymentUpdate(namespace, name, nodeSelector)
		if err != nil {
			return err
		}
		if !updateNodeSelector {
			// ensure no migration-spare exists, when quark deployment is up-to date with correct node-selector
			err = c.reconcileDeleteMigrationSpare(namespace, migSpareName)
			if err != nil {
				return err
			}
		}
	}

	// Try to find the NVRAM PVC
	// Create NVRAM PVC if not present and diskpacking is not enabled
	diskPackingEnabled, updateStatus, err = c.syncNvramPvc(volume, storagePool, serviceLevel, failureDomain)
	if err != nil {
		return err
	}

	// Allocate the address ahead of the deployment so we can pass the IP into
	// the pod for NSM callbacks.
	// TODO: Investigate if it is possible to do syncronize between two controller.
	if exportAddress == "" {
		// Mark to update the netappObj after the exportAddress is populated
		updateStatus = true
		volume, storagePool, exportAddress, err = c.syncExportAddress(volume, storagePool)
	} else if storagePool != nil && storagePool.Spec.GlobalAccess != storagePool.Status.GlobalAccess {
		storagePool, err = c.updateExistingServices(storagePool)
		if err == nil {
			updateStatus = true
		}
	} else if storagePool != nil && storagePool.Spec.GlobalAccessAborting {
		log.WithFields(log.Fields{
			"storagePool":     storagePool.Name,
			"ADClientIP":      storagePool.Status.ADClientIP,
			"resourceVersion": storagePool.ResourceVersion,
		}).Debug("Aborting global access enablement.")
		err = c.deleteADClientServicesIP(storagePool)
		if err == nil {
			updateStatus = true
		}
	}
	if isStaticIPAllocationError(err) {
		if !existsSpAddressAllocationErrorCondition(storagePool) {
			log.WithFields(log.Fields{
				"storagePool":  storagePool.Name,
				"updateStatus": updateStatus,
				"error":        err,
			}).Error("Failed to allocate static IP")
			timestamp := time.Now()
			spConditions = append(spConditions,
				quarkv1alpha1.NetAppStoragePoolCondition{
					Type:               quarkv1alpha1.NetAppStoragePoolAddressAllocationError,
					Status:             corev1.ConditionTrue,
					LastHeartbeatTime:  metav1.NewTime(timestamp),
					LastTransitionTime: metav1.NewTime(timestamp),
					Reason:             "Failed to allocate static IP address",
					Message:            err.Error(),
				},
			)
			updateStatus = true
		}
	} else if err != nil {
		return err
	} else if existsSpAddressAllocationErrorCondition(storagePool) {
		spConditions = util.RemoveStoragePoolConditions(spConditions,
			quarkv1alpha1.NetAppStoragePoolAddressAllocationError)
		updateStatus = true
	}

	// update connection tracking policy in loadbalancer only for gke
	if c.LoadBalancer && exportAddress != "" && c.Cloud == quarkconfig.CloudGCP {
		updateStatus, connectionTrackingPolicy, err = c.updateConnectionTrackingPolicy(storagePool)
		if err != nil {
			return err
		}
	}

	// Add finalizer for Etcd Key Cleanup
	if storagePool != nil && !HasFinalizer(storagePool.Finalizers, quarkconfig.EtcdKeyFinalizer) {
		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"finalizers":  strings.Join(storagePool.Finalizers, ", "),
			"finalizer":   quarkconfig.EtcdKeyFinalizer,
		}).Info("Adding etcdKeyFinalizer for storagePool.")
		storagePool, err = c.ensureEtcdKeyFinalizer(nil, storagePool, quarkconfig.EtcdKeyFinalizer)
		if err != nil {
			return err
		}
	}

	// Check if we are in upgrade path and if so go through upgrade workflow
	if statusQuarkVersion != "" && specQuarkVersion != statusQuarkVersion && err == nil {
		// TODO: Make sure we don't migrate zones during an upgrade or that it
		// still works
		err = c.syncNetappObjectUpgrade(volume, storagePool,
			serviceLevel, dataCount, quarkUUID, quarkNvramID, volumeUUID, ownerID,
			vserverUUID, nbladeUUID, exportAddress, nodeSelector, diskPackingEnabled)
		if _, ok := err.(UpgradeInProgressError); ok {
			// If upgrade in progress , check if a storage expansion is needed
			// If storage expansion is needed return error
			if storagePool != nil && storagePool.Spec.Size.Value() != storagePool.Status.Size.Value() {
				return fmt.Errorf("cannot expand storage while upgrade is in progress")
			}
			if volume != nil && volume.Spec.Size.Value() != volume.Status.Size.Value() {
				return fmt.Errorf("cannot expand storage while upgrade is in progress")
			}
		} else if err != nil {
			return err
		}
	}

	// sync volume creation and expansion operations
	if err == nil {
		volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err = c.syncVolumeCreateExpansion(volume, storagePool, serviceLevel,
			failureDomain, quarkUUID, quarkNvramID, volumeUUID, ownerID,
			vserverUUID, nbladeUUID, exportAddress, nodeSelector, diskPackingEnabled, &updateStatus)
		if err != nil {
			if _, ok := err.(DiskResizeInProgressError); ok {
				// Resize ongoing , just return , don't requeue
				// We will come back in when we get a pvc update event
				if storagePool != nil {
					log.WithFields(log.Fields{
						"storagePool": storagePool.Name,
					}).Info("Disk resize in progress.")
				} else {
					log.WithFields(log.Fields{
						"volume": volumeUUID,
					}).Info("Disk resize in progress.")
				}
				return nil
			} else if _, ok := err.(MigrationInProgressError); ok {
				if storagePool != nil {
					log.WithFields(log.Fields{
						"storagePool": storagePool.Name,
					}).Info("Migration in progress.")
				} else {
					log.WithFields(log.Fields{
						"volume": volumeUUID,
					}).Info("Migration in progress.")
				}
				return nil
			}
			return err
		}
	}
	// Finally, we update the status block of the NetAppVolume resource to reflect the
	// current state of the world.
	if updateStatus {
		var updateErr error

		// totalActualSize is unused in beta/GA version in updateNetAppObjectStatus() routine.
		// Don't need to set it for old volumes
		if storagePool != nil {
			statusProvisionedSize = storagePool.Status.ProvisionedSize.Value()
			specProvisionedSize = storagePool.Spec.ProvisionedSize.Value()
		} else {
			statusProvisionedSize = volume.Status.ProvisionedSize.Value()
			specProvisionedSize = volume.Spec.ProvisionedSize.Value()
		}
		if !isPreVol {
			if totalActualSize < desiredSize {
				totalActualSize = statusProvisionedSize
			} else {
				totalActualSize = specProvisionedSize
			}
		}
		volume, storagePool, updateErr = c.updateNetAppObjectStatus(
			volume, storagePool, totalActualSize, dataPvcs, exportAddress,
			conditions, spConditions, diskPackingEnabled, serviceLevel, connectionTrackingPolicy)
		if updateErr != nil {
			return fmt.Errorf("failed to update the NetAppVolume status: %v",
				updateErr)
		}
		if storagePool != nil {
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"namespace":   storagePool.Namespace,
			}).Debug("Emitting sync event.")
			c.recorder.Event(storagePool, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSyncedSP)
		}
		if storagePool == nil && volumeStateEvent != nil {
			c.recorder.Event(volume, volumeStateEvent.Type, volumeStateEvent.Reason, volumeStateEvent.Message)
		}
	}
	return err
}

func (c *Controller) updateNetappVolumeResourcesLabelForStoragePool(namespace, name string) error {
	labelSelector := fmt.Sprintf(quarkconfig.VolumeLabel+"=%s", name)
	listOptions := metav1.ListOptions{
		LabelSelector: labelSelector,
	}

	snapshotPolicies, err := c.QuarkClientset.QuarkV1alpha1().NetAppSnapshotPolicies(namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to get snapshot policy objects: %s", err.Error())
	}
	for _, snapshotPolicy := range snapshotPolicies.Items {
		if _, ok := snapshotPolicy.ObjectMeta.Labels[quarkconfig.StoragePoolLabel]; !ok {
			snapshotPolicy.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = name
			_, err = c.QuarkClientset.QuarkV1alpha1().NetAppSnapshotPolicies(namespace).Update(context.TODO(), &snapshotPolicy, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update snapshot policy object: %s", err.Error())
			}
		}
	}

	exportPolicies, err := c.QuarkClientset.QuarkV1alpha1().NetAppExportPolicies(namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to get export policy objects: %s", err.Error())
	}
	for _, exportPolicy := range exportPolicies.Items {
		if _, ok := exportPolicy.ObjectMeta.Labels[quarkconfig.StoragePoolLabel]; !ok {
			exportPolicy.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = name
			_, err = c.QuarkClientset.QuarkV1alpha1().NetAppExportPolicies(namespace).Update(context.TODO(),
				&exportPolicy, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update export policy object: %s", err.Error())
			}
		}
	}

	volumeCloudSnapshots, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumeCloudSnapshots(namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to get volumecloudsnapshots objects: %s", err.Error())
	}
	for _, volumeCloudSnapshot := range volumeCloudSnapshots.Items {
		if _, ok := volumeCloudSnapshot.ObjectMeta.Labels[quarkconfig.StoragePoolLabel]; !ok {
			volumeCloudSnapshot.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = name
			_, err = c.QuarkClientset.QuarkV1alpha1().NetAppVolumeCloudSnapshots(namespace).Update(context.TODO(), &volumeCloudSnapshot, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update volumeCloudSnapshot object: %s", err.Error())
			}
		}
	}

	volumeSnapshots, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumeSnapshots(namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to get volumeSnapshots objects: %s", err.Error())
	}
	for _, volumeSnapshot := range volumeSnapshots.Items {
		if _, ok := volumeSnapshot.ObjectMeta.Labels[quarkconfig.StoragePoolLabel]; !ok {
			volumeSnapshot.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = name
			_, err = c.QuarkClientset.QuarkV1alpha1().NetAppVolumeSnapshots(namespace).Update(context.TODO(), &volumeSnapshot, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update volumeSnapshot object: %s", err.Error())
			}
		}
	}

	volumeSnapshotLists, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumeSnapshotsLists(namespace).List(context.TODO(),
		listOptions)
	if err != nil {
		return fmt.Errorf("failed to get volumeSnapshotLists objects: %s", err.Error())
	}
	for _, volumeSnapshotList := range volumeSnapshotLists.Items {
		if _, ok := volumeSnapshotList.ObjectMeta.Labels[quarkconfig.StoragePoolLabel]; !ok {
			volumeSnapshotList.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = name
			_, err = c.QuarkClientset.QuarkV1alpha1().NetAppVolumeSnapshotsLists(namespace).Update(context.TODO(), &volumeSnapshotList, metav1.UpdateOptions{})
			if err != nil {
				return fmt.Errorf("failed to update volumeSnapshotList object: %s", err.Error())
			}
		}
	}
	return nil
}

func (c *Controller) constructStoragePoolFromVolume(
	volume *quarkv1alpha1.NetAppVolume,
	dataPvcs []*corev1.PersistentVolumeClaim,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	metaLabels := volume.ObjectMeta.GetLabels()
	ownerID := metaLabels[quarkconfig.OwnerIDLabel]
	storagePool := &quarkv1alpha1.NetAppStoragePool{
		ObjectMeta: metav1.ObjectMeta{
			Name: volume.Name,
			Labels: map[string]string{
				quarkconfig.StoragePoolLabel: volume.Name,
				quarkconfig.OwnerIDLabel:     ownerID,
			},
		},
		Spec: quarkv1alpha1.NetAppStoragePoolSpec{
			Size:            volume.Spec.Size,
			DisplayName:     volume.Spec.DisplayName,
			ServiceLevel:    volume.Spec.ServiceLevel,
			ProvisionedSize: volume.Spec.ProvisionedSize,
			QuarkVersion:    volume.Spec.QuarkVersion,
		},
	}
	if c.regionalMeta != nil {
		// Assume that the desired zone is the current zone of the volume
		// SDE only upgrades volumes that have come online, so we don't need to
		// handle an empty Zone string
		storagePool.Spec.Zone = volume.Status.Zone
		availableZones := []string{volume.Status.Zone}
		fd, err := c.regionalMeta.GetFailureDomainForVolume(c.KubeClientset,
			volume.Name, dataPvcs)
		if err != nil {
			return nil, fmt.Errorf("unable to look up zones for volume: %v", err)
		}
		for _, zone := range fd {
			if zone != volume.Status.Zone {
				availableZones = append(availableZones, zone)
			}
		}
		storagePool.Spec.AvailableZones = availableZones
	}
	if volume.Spec.Protocol == "cifs" {
		sMBConfiguration := quarkv1alpha1.SMBServerConfiguration{
			Netbios:          volume.Spec.SMBConfiguration.Netbios,
			Domain:           volume.Spec.SMBConfiguration.Domain,
			OrganizationUnit: volume.Spec.SMBConfiguration.OrganizationUnit,
			PreferredADSite:  volume.Spec.SMBConfiguration.PreferredADSite,
		}
		smbADConfiguration := quarkv1alpha1.SMBADConfiguration{
			EncryptedADCommunication: volume.Spec.SMBConfiguration.EncryptedADCommunication,
		}
		storagePool.Spec.ADDNSIPs = volume.Spec.SMBConfiguration.DNSIPs
		storagePool.Spec.SMBServerConfiguration = sMBConfiguration
		storagePool.Spec.SMBADConfiguration = smbADConfiguration
		storagePool.Spec.SMBPasswordResetCount = volume.Status.SMBPasswordResetCount
	}
	return storagePool, nil
}

func constructStoragePoolStatusFromVolume(volume *quarkv1alpha1.NetAppVolume) quarkv1alpha1.NetAppStoragePoolStatus {
	storagePoolStatus := quarkv1alpha1.NetAppStoragePoolStatus{
		Size:              volume.Status.Size,
		ProvisionedSize:   volume.Status.ProvisionedSize,
		DisplayName:       volume.Status.VolumeUUID,
		ServiceLevel:      volume.Status.ServiceLevel,
		QuarkVersion:      volume.Spec.QuarkVersion,
		ExportAddress:     volume.Status.ExportAddress,
		DataPvcs:          volume.Status.DataPvcs,
		NvramPvc:          volume.Status.NvramPvc,
		NvramSize:         volume.Status.NvramSize,
		StorageDeployment: volume.Status.StorageDeployment,
		QuarkUUID:         volume.Status.QuarkUUID,
		QuarkNvramID:      volume.Status.QuarkNvramID,
		QuarkDeployment:   volume.Status.QuarkDeployment,
		Zone:              volume.Status.Zone,
	}
	if volume.Spec.Protocol == "cifs" {
		sMBConfiguration := quarkv1alpha1.SMBServerConfiguration{
			Netbios:          volume.Status.SMBConfiguration.Netbios,
			Domain:           volume.Status.SMBConfiguration.Domain,
			OrganizationUnit: volume.Status.SMBConfiguration.OrganizationUnit,
			PreferredADSite:  volume.Status.SMBConfiguration.PreferredADSite,
		}
		smbADConfiguration := quarkv1alpha1.SMBADConfiguration{
			EncryptedADCommunication: volume.Status.SMBConfiguration.EncryptedADCommunication,
		}
		storagePoolStatus.SMBServerConfiguration = sMBConfiguration
		storagePoolStatus.SMBADConfiguration = smbADConfiguration
		storagePoolStatus.SMBPasswordResetCount = volume.Status.SMBPasswordResetCount
	}
	return storagePoolStatus
}

func (c *Controller) syncVolumeDNS(volume *quarkv1alpha1.NetAppVolume) (vol *quarkv1alpha1.NetAppVolume, err error) {
	if volume.Spec.Protocol != "cifs" {
		return volume, nil
	}
	if len(volume.Status.SMBConfiguration.DNSIPs) > 0 && !util.IsEqualStrings(volume.Spec.SMBConfiguration.DNSIPs, volume.Status.SMBConfiguration.DNSIPs) {
		err = c.deleteQuarkPod(volume)
		if err != nil {
			return nil, err
		}
		conditions := volume.Status.Conditions
		if existsVolOnlineCondition(volume) {
			conditions = MakeNetAppVolumeConditionOffline(volume, quarkconfig.MessageDNSUpdate)
		}
		c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.NetAppVolumeOfflined, quarkconfig.MessageDNSUpdate)
		volume, err = c.updateNetAppVolumeDNSStatus(volume, conditions)
	}
	return volume, err
}

// Upgrade netappObj workflow
func (c *Controller) syncNetappObjectUpgrade(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	volumeUUID string,
	ownerID string,
	vserverUUID string,
	nbladeUUID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) error {
	var (
		specQuarkVersion string
		specSize         int64
		namespace, name  = GetNamespaceNameForNetappObj(volume, storagePool)
		migSpareName     = GetQuarkMigrationSpareDeployName(name)
	)
	if storagePool != nil {
		specQuarkVersion = storagePool.Spec.QuarkVersion
		specSize = storagePool.Spec.Size.Value()
	} else {
		specQuarkVersion = volume.Spec.QuarkVersion
		specSize = volume.Spec.Size.Value()
	}

	// check if upgrade configmap is already present
	cfgMap, err := c.getUpgradeConfigMap(volume, storagePool)
	if err != nil {
		if !errors.IsNotFound(err) {
			log.WithFields(log.Fields{
				"netappObj": name,
				"error":     err,
			}).Error("Failed to get the upgrade configmap.")
			return fmt.Errorf("failed to get the upgrade configmap")
		}
		// currently there is a race condition when podrick deletes the configmap
		// and nvc enters the upgrade workflow after SP upgrade is complete.
		// checking here if the status and spec quarkVersion of SP are same to verify if upgrade has already completed
		storagePool, err = c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(namespace).Get(context.TODO(),
			name, metav1.GetOptions{})
		if err != nil {
			log.WithFields(log.Fields{
				"storagepool": storagePool.Name,
				"error":       err,
			}).Error("Failed to get storage pool when upgrade configmap is not found.")
			return err
		}
		if storagePool.Status.QuarkVersion == storagePool.Spec.QuarkVersion {
			return nil
		}
	} else {
		// upgrade configmap found, check the quarkVersion in configmap with netappObj's spec
		// if quarkVersion is found and matches, podrick would be using it, just return
		// if not, delete the configmap and recreate it
		cfgMapQuarkVersion, ok := cfgMap.Data[quarkconfig.QuarkVersion]
		if ok && cfgMapQuarkVersion == specQuarkVersion {
			if storagePool != nil {
				c.recorder.Event(storagePool, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)
			} else {
				c.recorder.Event(volume, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)
			}
			return UpgradeInProgressError{}
		} else {
			if err := c.deleteConfigMap(cfgMap.Name, cfgMap.Namespace); err != nil {
				log.WithFields(log.Fields{
					"configmap": cfgMap.Name,
					"error":     err,
				}).Error("Failed to delete upgrade configmap.")
				return err
			}
		}
	}

	// if we are here, no upgrade configmap is present in the Quark Pod(volume's) namespace in the cluster
	// We deploy a migration-spare pod to allow quarkpod to schedule on the correct size node(defined by service level) after upgrade
	slNodeSelector, err := getServiceLevelNodeSelector(specSize, serviceLevel)
	if err != nil {
		return err
	}
	needsMigration, err := c.needQuarkPodMigration(volume, storagePool, nodeSelector)
	if err != nil {
		return err
	}
	// deploy migration-spare pod on right-sized VM when node selectors do not match
	if needsMigration {
		// update the node pool in the quark deployment node selector
		for k, v := range slNodeSelector {
			nodeSelector[k] = v
		}
		err = c.deployMigrationSpare(nodeSelector, volume, storagePool, serviceLevel)
		if err != nil {
			if _, ok := err.(MigrationInProgressError); ok {
				if storagePool != nil {
					log.WithFields(log.Fields{
						"storagePool": storagePool.Name,
					}).Info("Migration in progress.")
				} else {
					log.WithFields(log.Fields{
						"volume": volumeUUID,
					}).Info("Migration in progress.")
				}
				return nil
			}
			return fmt.Errorf("failed to deploy %s during upgrade: %s", migSpareName, err)
		}
	} else {
		err = c.reconcileDeleteMigrationSpare(namespace, migSpareName)
		if err != nil {
			return err
		}
	}

	deploymentDefinition, err := c.makeRevisedQuarkDeployment(volume, storagePool,
		serviceLevel, dataCount, quarkUUID, quarkNvramID, volumeUUID, ownerID,
		vserverUUID, nbladeUUID, exportAddress, nodeSelector, diskPackingEnabled)
	if err != nil {
		log.WithFields(log.Fields{
			"netappObj": name,
			"error":     err,
		}).Error("Failed to generate Quark deployment definition for upgrade.")
		return fmt.Errorf("failed to generate the Quark deployment definition for upgrade")
	}
	deploymentDefinitionYaml, err := yaml.Marshal(deploymentDefinition)
	if err != nil {
		log.WithFields(log.Fields{
			"netappObj": name,
			"error":     err,
		}).Error("Failed to marshal the Quark deployment definition for upgrade.")
		return fmt.Errorf("failed to marshal the Quark deployment definition for upgrade")
	}
	// Create a configmap, Podrick will watch for creation of this configmap and trigger upgrade workflow from its side
	if err := c.createQuarkPodUpgradeConfigMap(volume, storagePool, string(deploymentDefinitionYaml)); err != nil {
		log.WithFields(log.Fields{
			"netappObj": name,
			"error":     err,
		}).Error("Failed to create Upgrade configmap.")
		return err
	}
	log.WithFields(log.Fields{
		"netappObj": name,
		"error":     err,
	}).Debug("Created Upgrade configmap.")
	if storagePool != nil {
		c.recorder.Event(storagePool, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)
	} else {
		c.recorder.Event(volume, corev1.EventTypeNormal, quarkconfig.SuccessSynced, quarkconfig.MessageResourceSynced)
	}

	return nil
}

// Check and create Quark vol PVC is needed
func (c *Controller) syncQuarkvolPvc(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string) error {
	// TO-DO: Remove this code when nodevol becomes the default
	// Try to find the QuarkVol PVC
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	quarkVolPvc, err := c.pvcsLister.PersistentVolumeClaims(namespace).Get(
		GetQuarkVolPvcName(namespace, name))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the QuarkVol PVC doesn't yet exist, create it
			log.WithFields(log.Fields{
				"netappObj": name,
				"size":      c.QuarkVolSize,
			}).Info("Creating QuarkVol PVC for volume.")

			quarkVolPvc, err = c.KubeClientset.CoreV1().PersistentVolumeClaims(
				namespace).Create(
				context.TODO(),
				makeQuarkVolPvc(
					volume, storagePool, util.GetSCNameForFailureDomain(
						serviceLevel.Spec.QuarkVolumeStorageClass, failureDomain),
					c.QuarkVolSize, c.EnableMultiAttach,
				),
				metav1.CreateOptions{},
			)
			if err != nil {
				return fmt.Errorf("failed to create QuarkVol PVC %s: %s", quarkVolPvc.Name, err.Error())
			}
		} else {
			return fmt.Errorf("failed to get QuarkVol PVC %s: %s", GetQuarkVolPvcName(namespace, name), err.Error())
		}
	}

	// Make sure the PVC we found is controlled by us
	if !c.isControlledByNetappObj(quarkVolPvc, volume, storagePool) {
		msg := fmt.Sprintf(quarkconfig.MessageResourceExists+"%s", quarkVolPvc.Name, quarkVolPvc.Kind)
		if storagePool != nil {
			c.recorder.Event(storagePool, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		} else {
			c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		}
		return fmt.Errorf(msg)
	}

	return nil
}

func (c *Controller) updateServiceAnnotations(
	svc *corev1.Service,
	protocol quarkconfig.SNATProtocol,
	portString string,
) error {
	annotation := fmt.Sprintf("%s:%s", protocol, portString)

	if svc.Annotations[quarkconfig.SNATPortRangeAnnotation] == annotation {
		return nil
	}
	svcCopy := svc.DeepCopy()
	svcCopy.Annotations[quarkconfig.SNATPortRangeAnnotation] = annotation
	_, err := c.KubeClientset.CoreV1().Services(svc.Namespace).Update(
		context.TODO(), svcCopy, metav1.UpdateOptions{})
	if err != nil {
		return fmt.Errorf("unable to update annotations for service %s: %v",
			svc.Name, err)
	}
	return nil

}

func (c *Controller) deleteADClientServicesIP(
	sp *quarkv1alpha1.NetAppStoragePool,
) error {
	svcNames := []string{
		getADDNSTCPLoadBalancerServiceName(sp.Namespace, sp.Name),
		getADDNSUDPLoadBalancerServiceName(sp.Namespace, sp.Name),
	}
	// Delete the services before deleting the IP address
	servicesDeleted := true
	for _, name := range svcNames {
		err := c.KubeClientset.CoreV1().Services(sp.Namespace).Delete(
			context.TODO(), name, metav1.DeleteOptions{})
		if err != nil {
			if !errors.IsNotFound(err) {
				return fmt.Errorf("failed to delete AD client service %s: %v",
					name, err)
			}
		} else {
			servicesDeleted = false
		}
	}
	if !servicesDeleted {
		// Ensure services are deleted
		for _, name := range svcNames {
			_, err := c.KubeClientset.CoreV1().Services(sp.Namespace).Get(
				context.TODO(), name, metav1.GetOptions{})
			if err == nil {
				return fmt.Errorf("awaiting %s service deletion", name)
			} else if !errors.IsNotFound(err) {
				return fmt.Errorf("failed to check if service %s is deleted: "+
					"%v", name, err)
			}
		}
	}
	// restore the original SNAT port ranges
	svcDescriptors := []struct {
		name           string
		annotationFunc func(string, bool) string
	}{
		{
			GetTCPLoadBalancerServiceName(sp.Namespace, sp.Name),
			getMainServiceTCPSNATAnnotations,
		},
		{
			GetUDPLoadBalancerServiceName(sp.Namespace, sp.Name),
			getMainServiceUDPSNATAnnotations,
		},
	}
	for _, d := range svcDescriptors {
		svc, err := c.KubeClientset.CoreV1().Services(sp.Namespace).Get(
			context.TODO(), d.name, metav1.GetOptions{})
		if err != nil {
			return fmt.Errorf("failed to get svc %s: %v", d.name, err)
		}
		if snatAnnotations := d.annotationFunc(
			"all", true,
		); snatAnnotations != "" {
			if svc.Annotations == nil {
				svc.Annotations = make(map[string]string)
			}
			svc.Annotations[quarkconfig.SNATPortRangeAnnotation] = snatAnnotations
		}
		_, err = c.KubeClientset.CoreV1().Services(sp.Namespace).Update(
			context.TODO(), svc, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update svc %s: %v", svc.Name, err)
		}
	}
	// Both services are now deleted; delete the address
	// Set GlobalAccessAborting to false; it will get updated when the finalizer
	// is removed.
	sp.Spec.GlobalAccessAborting = false
	log.WithFields(log.Fields{
		"storagePool":     sp.Name,
		"resourceVersion": sp.ResourceVersion,
	}).Debug("Finished deleting services; deleting address")
	return c.cleanUpAddress(nil, sp, getADClientAddrName(sp.Namespace, sp.Name),
		quarkconfig.ADClientAddressFinalizer)
}

func (c *Controller) updateExistingServices(
	sp *quarkv1alpha1.NetAppStoragePool,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	// This is the first time we're enabling global access
	if sp.Status.ADClientIP == "" && sp.Spec.GlobalAccess {
		log.WithFields(log.Fields{
			"storagePool":     sp.Name,
			"resourceVersion": sp.ResourceVersion,
		}).Debug("Beginning global access enablement")
		updatedSP, _, err := c.setupGCPILB(
			nil, sp,
			sp.Finalizers,
			quarkconfig.ADClientAddressFinalizer,
			getADClientAddrName(sp.Namespace, sp.Name),
			getADDNSTCPLoadBalancerServiceName(sp.Namespace, sp.Name),
			func(addr string) *corev1.Service {
				return c.makeADClientTCPLoadBalancerService(sp, addr)
			},
			getADDNSUDPLoadBalancerServiceName(sp.Namespace, sp.Name),
			func(addr string) *corev1.Service {
				return c.makeADClientUDPLoadBalancerService(sp, addr)
			},
		)
		if err == nil {
			log.WithFields(log.Fields{
				"storagePool":     sp.Name,
				"resourceVersion": sp.ResourceVersion,
			}).Debug("Finished setting up ILB")
		}
		return updatedSP, err
	}
	log.WithFields(log.Fields{
		"storagePool":     sp.Name,
		"globalAccess":    sp.Spec.GlobalAccess,
		"value":           sp.Spec.GlobalAccess,
		"resourceVersion": sp.ResourceVersion,
	}).Debug("AD/DNS services exist; updating existing annotations.")
	for _, svcName := range []string{
		getADDNSTCPLoadBalancerServiceName(sp.Namespace, sp.Name),
		getADDNSUDPLoadBalancerServiceName(sp.Namespace, sp.Name),
	} {
		svc, err := c.servicesLister.Services(sp.Namespace).Get(svcName)
		if err != nil {
			return nil, fmt.Errorf("unable to retrieve service for updating "+
				"%s: %v", svcName, err)
		}
		svcCopy := svc.DeepCopy()
		c.cloudClient.AddCloudAnnotations(svcCopy, c.IlbSubnet,
			sp.Spec.GlobalAccess)
		_, err = c.KubeClientset.CoreV1().Services(sp.Namespace).Update(
			context.TODO(), svcCopy, metav1.UpdateOptions{})
		if err != nil {
			return nil, fmt.Errorf("unable to update service %s: %v", svcName,
				err)
		}
	}
	return sp, nil
}

// Setup ILB based on the cloud hyperscaler we are running on
func (c *Controller) syncExportAddress(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
) (
	*quarkv1alpha1.NetAppVolume, *quarkv1alpha1.NetAppStoragePool,
	string, error,
) {
	// Mark to update the volume after the exportAddress is populated
	var (
		err           error
		exportAddress string
	)
	// If LoadBalancer only then do all of the following:
	// For GCP deployments:
	//	* Get a static LoadBalancer IP
	//  * Use this IP to create the TCP & UDP ILB Loadbalancer Service and use that as the
	//    volume's exportAddress
	// For Azure deployments and CTL deployments:
	//  * Create ILB service
	//  * Use service ingress IP as the exportAddress
	// Else
	//  * Skip to just create a ClusterIP Service and use that as the volume exportAddress
	if !c.LoadBalancer {
		exportAddress, err = c.makeVolumeService(volume, storagePool)
		if err != nil {
			return volume, storagePool, exportAddress, fmt.Errorf(
				"failed to create volume service for volume: %s", err.Error())
		}
	} else {
		switch c.Cloud {
		case quarkconfig.CloudGCP:
			storagePool, exportAddress, err = c.setupGCPILBs(volume, storagePool)
			if err != nil {
				return volume, storagePool, exportAddress, err
			}
		case quarkconfig.CloudAzure:
			exportAddress, err = c.setupAzureILB(volume, storagePool)
			if err != nil {
				return volume, storagePool, exportAddress, err
			}
		default:
			exportAddress, err = c.setupDefaultILB(volume, storagePool)
			if err != nil {
				return volume, storagePool, exportAddress, err
			}
		}
	}
	volume, storagePool, err = c.updateNetAppObjectExportAddress(volume, storagePool, exportAddress)
	if err != nil {
		return volume, storagePool, exportAddress, fmt.Errorf(
			`failed to update the NetAppVolume "exportAddress" status: %v`, err)
	}
	return volume, storagePool, exportAddress, nil
}

// Returns a staticIPAllocation error if IP address allocation fails.
func (c *Controller) checkAndAllocateIPAddress(
	ctx context.Context,
	addrName string,
) (string, error) {
	addr, err := c.cloudClient.GetLoadBalancerAddress(ctx, addrName)
	if err != nil && c.cloudClient.IsNotFound(err) {
		// If the address doesn't exist yet, create it
		err = c.cloudClient.CreateLoadBalancerAddress(ctx, addrName,
			c.IlbSubnet)
		if err != nil {
			return "", newStaticIPAllocationError(addrName, err)
		}
		addr, err = c.cloudClient.GetLoadBalancerAddress(ctx, addrName)
		if err != nil {
			return "", fmt.Errorf(
				"static IP for netappObj isn't ready yet; retrying")
		}
	} else if err != nil {
		return "", fmt.Errorf(
			"failed to get load balancer static IP: %v", err)
	}
	return addr, nil
}

// TODO: Simplify the parameter list once volumes are gone.
func (c *Controller) setupGCPILB(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	finalizers []string,
	finalizer string,
	addrName string,
	// The closure parameters should be closures that wrap a constructor
	// function (make*LoadBalancerService), passing in address, as that's not
	// known at the time this function is called.
	tcpServiceName string,
	tcpServiceClosure func(addr string) *corev1.Service,
	udpServiceName string,
	udpServiceClosure func(addr string) *corev1.Service,
) (*quarkv1alpha1.NetAppStoragePool, string, error) {
	var err error

	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)

	if !HasFinalizer(finalizers, finalizer) {
		log.WithFields(log.Fields{
			"netappObj":  name,
			"finalizers": strings.Join(finalizers, ", "),
			"finalizer":  finalizer,
		}).Info("Adding finalizer for netappObj.")
		volume, storagePool, err = AddFinalizer(c.QuarkClientset, volume,
			storagePool, finalizer)
		if err != nil {
			return nil, "", fmt.Errorf(
				"failed to add finalizer for netappObj: %v", err)
		}
	}
	ctx := context.Background()
	// Get the static IP address associated with this volume
	addr, err := c.checkAndAllocateIPAddress(ctx, addrName)
	if err != nil {
		if isStaticIPAllocationError(err) {
			// Don't remove the finalizer here, as removing it will cause an
			// endless cycle of adding and readding it. Remove the finalizer
			// either on abort or SP deletion. In the worst case, this will
			// cause us to make a superfluous GCP API call, which shouldn't
			// hurt anything
			return storagePool, "", err
		}
		return nil, "", fmt.Errorf(
			"unable to allocate address: %v", err)
	}
	log.WithFields(log.Fields{
		"netappObj": name,
		"address":   addr,
	}).Info("Created static IP for netappObj.")

	// At the moment ILB doesn't support multi-protocol
	if c.IlbTCP {
		// Get the TCP load balancer service associated with this volume
		_, err = c.servicesLister.Services(namespace).Get(tcpServiceName)
		if err != nil {
			if errors.IsNotFound(err) {
				// If the service doesn't exist yet, create it
				_, err = c.KubeClientset.CoreV1().Services(namespace).Create(
					context.TODO(), tcpServiceClosure(addr),
					metav1.CreateOptions{})
				if err == nil {
					log.WithFields(log.Fields{
						"netapObj": tcpServiceName,
					}).Info("Created TCP load balancer service for netapObj.")
				} else {
					return nil, addr, fmt.Errorf(
						"failed to create TCP load balancer service: %v", err)
				}
			} else {
				return nil, addr, fmt.Errorf(
					"failed to get TCP load balancer service: %v", err)
			}
		}
	}

	if c.IlbUDP {
		// Get the UDP load balancer service associated with this volume
		_, err = c.servicesLister.Services(namespace).Get(udpServiceName)
		if err != nil {
			if errors.IsNotFound(err) {
				// If the service doesn't exist yet, create it
				_, err = c.KubeClientset.CoreV1().Services(namespace).Create(
					context.TODO(), udpServiceClosure(addr),
					metav1.CreateOptions{})
				if err == nil {
					log.WithFields(log.Fields{
						"netappObj": name,
					}).Info("Created UDP load balancer service for netappObj.")
				} else {
					return nil, addr, fmt.Errorf(
						"failed to create UDP load balancer service: %v", err)
				}
			} else {
				return nil, addr, fmt.Errorf(
					"failed to get UDP load balancer service: %v", err)
			}
		}
	}
	return storagePool, addr, nil
}

// Setup ILB for google cloud
func (c *Controller) setupGCPILBs(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
) (*quarkv1alpha1.NetAppStoragePool, string, error) {
	var (
		finalizers   []string
		protocol     string
		globalAccess = false
	)
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	if storagePool == nil {
		finalizers = volume.ObjectMeta.Finalizers
		protocol = volume.Spec.Protocol
	} else {
		protocol = "all"
		globalAccess = storagePool.Spec.GlobalAccess
		finalizers = storagePool.ObjectMeta.Finalizers
	}
	if globalAccess {
		// We fail SP create if we can't allocate an ILB for AD traffic,
		// so it's safe to create that ILB before the main export one.
		storagePool, _, err := c.setupGCPILB(volume, storagePool,
			finalizers, quarkconfig.ADClientAddressFinalizer,
			getADClientAddrName(namespace, name),
			getADDNSTCPLoadBalancerServiceName(namespace, name),
			func(addr string) *corev1.Service {
				return c.makeADClientTCPLoadBalancerService(storagePool, addr)
			},
			getADDNSUDPLoadBalancerServiceName(namespace, name),
			func(addr string) *corev1.Service {
				return c.makeADClientUDPLoadBalancerService(storagePool, addr)
			},
		)
		if err != nil && isStaticIPAllocationError(err) {
			return storagePool, "", err
		} else if err != nil {
			return nil, "", err
		}
	}
	storagePool, exportAddress, err := c.setupGCPILB(
		volume, storagePool,
		finalizers, quarkconfig.AddressFinalizer,
		getAddrName(namespace, name),
		GetTCPLoadBalancerServiceName(namespace, name),
		func(addr string) *corev1.Service {
			return c.makeTCPLoadBalancerService(
				volume, storagePool, addr, true, protocol)
		},
		GetUDPLoadBalancerServiceName(namespace, name),
		func(addr string) *corev1.Service {
			return c.makeUDPLoadBalancerService(
				volume, storagePool, addr, true, protocol)
		},
	)
	if err != nil && isStaticIPAllocationError(err) {
		return storagePool, "", err
	} else if err != nil {
		return nil, "", fmt.Errorf(
			"failed to set up export address: %v", err)
	}
	return storagePool, exportAddress, nil
}

// Setup ILB for Azure
func (c *Controller) setupAzureILB(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (exportAddress string, err error) {
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	protocol := "all"
	if storagePool == nil {
		exportAddress = volume.Status.ExportAddress
	} else {
		exportAddress = storagePool.Status.ExportAddress
	}
	if volume != nil {
		protocol = volume.Spec.Protocol
	}
	internal := c.IlbSubnet != ""
	var svc *corev1.Service
	serviceExists := true
	if c.IlbTCP && c.IlbUDP {
		// Check if we have a multiprotocol service
		// Create one if not present
		// Make a multiprotocol service
		svc, err = c.servicesLister.Services(namespace).Get(getServiceName(namespace, name))
		if err != nil {
			if errors.IsNotFound(err) {
				svc = c.makeMultiProtocolILB(volume, storagePool, "", protocol)
				serviceExists = false
			} else {
				return exportAddress, fmt.Errorf("failed to get multiprotocol load balancer service: %s", err.Error())
			}
		}

	} else {
		// Check if we have an existing tcp svc
		// Create one if not present
		svc, err = c.servicesLister.Services(namespace).Get(GetTCPLoadBalancerServiceName(namespace, name))
		if err != nil {
			if errors.IsNotFound(err) {
				svc = c.makeTCPLoadBalancerService(volume, storagePool, "", internal, protocol)
				serviceExists = false
			} else {
				return exportAddress, fmt.Errorf("failed to get TCP load balancer service: %s", err.Error())
			}
		}
	}

	// Create the service if it doesn't exist
	if !serviceExists {
		svc, err = c.KubeClientset.CoreV1().Services(namespace).Create(context.TODO(),
			svc,
			metav1.CreateOptions{})
		if err == nil {
			log.WithFields(log.Fields{
				"netappObj": name,
			}).Info("Created load balancer service for netappObj.")
		} else {
			return exportAddress, fmt.Errorf("failed to create load balancer service: %s", err.Error())
		}
	}
	if ingress := svc.Status.LoadBalancer.Ingress; len(ingress) > 0 {
		exportAddress = ingress[0].IP
		log.WithFields(log.Fields{
			"netappObj":     name,
			"exportAddress": exportAddress,
		}).Debug("Found load balancer IP for Azure netappObj")
	} else {
		return exportAddress, fmt.Errorf("failed to get exportAddress from service")
	}
	return exportAddress, nil
}

// Setup ILB for CTL deployments and cloud providers other than GCP and Azure
func (c *Controller) setupDefaultILB(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (exportAddress string, err error) {
	// For any other cloud including CTL
	var svc *corev1.Service
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	protocol := "all"
	if storagePool == nil {
		exportAddress = volume.Status.ExportAddress
	} else {
		exportAddress = storagePool.Status.ExportAddress
	}
	if volume != nil {
		protocol = volume.Spec.Protocol
	}
	if c.IlbTCP {
		// Get the service associated with this volume
		svc, err = c.servicesLister.Services(namespace).Get(GetTCPLoadBalancerServiceName(namespace, name))
		if err != nil {
			if errors.IsNotFound(err) {
				// If the service doesn't exist yet, create it
				svc, err = c.KubeClientset.CoreV1().Services(namespace).Create(context.TODO(),
					c.makeTCPLoadBalancerService(volume, storagePool, "", false, protocol),
					metav1.CreateOptions{})
				if err == nil {
					log.WithFields(log.Fields{
						"netappObj": name,
					}).Info("Created TCP load balancer service for netappObj.")
				} else {
					return exportAddress, fmt.Errorf("failed to create TCP load balancer service: %s", err.Error())
				}
			} else {
				return exportAddress, fmt.Errorf("failed to get TCP load balancer service: %s", err.Error())
			}
		}

		for exportAddress == "" {
			if len(svc.Status.LoadBalancer.Ingress) > 0 {
				exportAddress = svc.Status.LoadBalancer.Ingress[0].IP
				log.WithFields(log.Fields{
					"netappObj":     name,
					"exportAddress": exportAddress,
					"service":       GetTCPLoadBalancerServiceName(namespace, name),
				}).Debug("Updated service.")
				break
			} else {
				log.WithFields(log.Fields{
					"netappObj": name,
					"service":   GetTCPLoadBalancerServiceName(namespace, name),
				}).Debug("Waiting for load balancer address...")
				time.Sleep(2 * time.Second)
				svc, err = c.servicesLister.Services(namespace).Get(GetTCPLoadBalancerServiceName(namespace, name))
				if err != nil {
					return exportAddress, fmt.Errorf("failed to find service for netappObj: %s", err.Error())
				}
			}
		}
	}
	return exportAddress, nil
}

// updateConnectionTrackingPolicy updates the connection tracking policy on gke internal load balancers
func (c *Controller) updateConnectionTrackingPolicy(
	sp *quarkv1alpha1.NetAppStoragePool,
) (bool, string, error) {
	connectionTrackingPolicyValue := quarkconfig.ConnectionTrackingPolicyNeverPersist
	ctx := context.Background()
	gcpClient, ok := c.cloudClient.(*cloud.GCPClient)
	if !ok {
		return false, "", fmt.Errorf("must use a gcp cloud client when updating connection tracking policy")
	}
	patchStatusEachService := false
	policyUpdated := true
	for _, svcName := range []string{
		GetTCPLoadBalancerServiceName(sp.Namespace, sp.Name),
		GetUDPLoadBalancerServiceName(sp.Namespace, sp.Name)} {
		svc, err := c.servicesLister.Services(sp.Namespace).Get(svcName)
		if err != nil {
			return false, "", fmt.Errorf("unable to retrieve service for updating connection tracking policy %s: %v",
				svcName, err)
		}
		if svc.Spec.LoadBalancerIP != "" {
			patchStatusEachService, err = gcpClient.UpdateILBConnectionTrackingPolicy(ctx, svc, connectionTrackingPolicyValue)
			if err != nil {
				return false, "", fmt.Errorf("failed to update the connection tracking policy: %v", err)
			}
			if !patchStatusEachService {
				policyUpdated = false
			}
		} else {
			return false, "", fmt.Errorf("service doesnt has a load balancer ip yet serviceName : %s ", svcName)
		}
	}
	return policyUpdated, connectionTrackingPolicyValue, nil
}

// Check if storage deployments are created , if not create them
func (c *Controller) syncStorageDeployment(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	updateStatus *bool,
	dataCount int,
	dataPvcs []*corev1.PersistentVolumeClaim,
	diskPackingEnabled bool) (err error) {
	// Try to find the storage deployment
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	storageDeploy, err := c.deploysLister.Deployments(namespace).Get(getStorageDeployName(namespace, name))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the storage deployment doesn't exist yet, create it
			*updateStatus = true
			log.WithFields(log.Fields{
				"netappObj": name,
			}).Debug("Updated deployment for multiattach.")
			storageDeploy, err = c.KubeClientset.AppsV1().Deployments(namespace).Create(context.TODO(),
				c.makeStorageDeployment(volume, storagePool, dataCount, diskPackingEnabled),
				metav1.CreateOptions{})
			if err != nil {
				return fmt.Errorf("failed to create storage deployment: %s", err.Error())
			}
		} else {
			return fmt.Errorf("failed to get storage deployment: %s", err.Error())
		}
	}

	// Make sure the deployment we found is controlled by us
	if !c.isControlledByNetappObj(storageDeploy, volume, storagePool) {
		msg := fmt.Sprintf(quarkconfig.MessageResourceExists, storageDeploy.Name)
		if storagePool != nil {
			c.recorder.Event(storagePool, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		} else {
			c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		}
		return fmt.Errorf(msg)
	}

	// Validate that the PVCs in the storage deployment template match the
	// actual list of PVCs
	// XXX this just compares the list lengths, we should actually compare the list contents
	if len(storageDeploy.Spec.Template.Spec.Volumes) != len(dataPvcs)+1 {
		_, err := c.KubeClientset.AppsV1().Deployments(namespace).Update(context.TODO(),
			c.makeStorageDeployment(volume, storagePool, dataCount, diskPackingEnabled),
			metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to update the storage deployment PVCs: %s", err.Error())
		}
	}

	return nil
}

// Check if appropriate nvram PVC are configured
func (c *Controller) syncNvramPvc(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string) (bool, bool, error) {
	diskPackingEnabled := false
	updateStatus := false
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	nvramPvc, err := c.pvcsLister.PersistentVolumeClaims(namespace).Get(getNvramPvcName(namespace, name))
	if err != nil {
		if !errors.IsNotFound(err) {
			return diskPackingEnabled, updateStatus, fmt.Errorf("failed to get NVRAM PVC %s: %s", getNvramPvcName(namespace, name), err.Error())
		}
		// Need the nvram size for either disk packing, or separate nvram disk
		nvramSize, err := c.getNvramSize(volume, storagePool, serviceLevel)
		if err != nil {
			msg := fmt.Sprintf("Invalid value of %d NVRAM Size obtained from service level spec: %v", nvramSize, err)
			if storagePool != nil {
				c.recorder.Event(storagePool, corev1.EventTypeWarning, "ErrSyncStoragePool", msg)
			} else {
				c.recorder.Event(volume, corev1.EventTypeWarning, "ErrSyncVolume", msg)
			}
			log.Errorf("Failed to determine the quark pod NVRAM size: %s", msg)
			return diskPackingEnabled, updateStatus, fmt.Errorf("%s", msg)
		}

		if c.UseDiskPacking {
			diskPackingEnabled = true
			// We're doing disk packing, don't create the NVRAM PVC
			log.WithFields(log.Fields{
				"netappObj": name,
				"nvramSize": nvramSize,
			}).Debug("Using disk packing.")
		} else {
			// If the NVRAM PVC doesn't yet exist, create it
			log.WithFields(log.Fields{
				"netappObj": name,
				"size":      nvramSize,
			}).Debug("Creating NVRAM PVC.")
			updateStatus = true
			log.WithFields(log.Fields{
				"netappObj": name,
			}).Debug("Updated NVRAM.")
			// using namespace to use the routine for either volume or storagePool
			nvramPvc, err = c.KubeClientset.CoreV1().PersistentVolumeClaims(
				namespace).Create(
				context.TODO(),
				makeNvramPvc(
					volume, storagePool, util.GetSCNameForFailureDomain(
						serviceLevel.Spec.NvramVolumeStorageClass, failureDomain,
					), nvramSize,
					c.EnableMultiAttach),
				metav1.CreateOptions{})
			if err != nil {
				return diskPackingEnabled, updateStatus, fmt.Errorf("failed to create NVRAM PVC %s: %s", nvramPvc.Name, err.Error())
			}
		}
	}

	if !diskPackingEnabled {
		// Make sure the PVC we found is controlled by us
		if !c.isControlledByNetappObj(nvramPvc, volume, storagePool) {
			msg := fmt.Sprintf(quarkconfig.MessageResourceExists, nvramPvc.Name)
			if storagePool != nil {
				c.recorder.Event(storagePool, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
			} else {
				c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
			}
			return diskPackingEnabled, updateStatus, fmt.Errorf(msg)
		}
	}

	return diskPackingEnabled, updateStatus, nil

}

// Check if volume needs to be expanded , if needed resize existing PVCs
// create new PVCs if existing pvcs can't be expanded further
// return list of new data pvcs after going through expansion
func (c *Controller) syncVolumeExpansionGCP(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	totalRequestedSize int64,
	totalActualSize int64,
	desiredSpecSize int64,
	desiredSize int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string,
) ([]*corev1.PersistentVolumeClaim, []corev1.Volume, []corev1.VolumeDevice, bool, error) {
	var (
		newDataVolumesList       []corev1.Volume
		newDataVolumeDevicesList []corev1.VolumeDevice
		dataPvcs                 []*corev1.PersistentVolumeClaim
		diskAdd                  bool
		namespace, name          = GetNamespaceNameForNetappObj(volume, storagePool)
	)

	diskAdd = false
	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(namespace).List(c.getDataPvcSelector(volume, storagePool))
	if err != nil {
		return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to list PVCs: %s", err.Error())
	}
	if len(dataPvcs) > 0 {
		log.WithFields(log.Fields{
			"netappObj":       name,
			"requestedSize":   totalRequestedSize,
			"actualSize":      totalActualSize,
			"desiredSpecSize": desiredSpecSize,
			"desiredSize":     desiredSize,
		}).Info("ProvisionedSize changed")
	}
	needSize := desiredSize - totalRequestedSize
	// If the size doesn't match, we'll either need to resize one of the existing PVCs,
	// or add another one.
	// If the resizePvcs feature is turned on, scan the existing PVCs and resize any that
	// are below the maximum size to try to achieve the new size
	// [ARDALAN] We should check with Rajesh to see if ONTAP/WAFL can handle simultaneous (online) resizing of Random disks.
	for _, dataPvc := range dataPvcs {
		oldSize := dataPvc.Spec.Resources.Requests[corev1.ResourceStorage]
		if c.MaxPvcSize <= oldSize.Value() {
			continue
		}
		growSize := needSize
		newSize := oldSize.Value() + growSize
		if newSize > c.MaxPvcSize {
			growSize = c.MaxPvcSize - oldSize.Value()
			newSize = c.MaxPvcSize
		}
		log.WithFields(log.Fields{
			"PVC":     dataPvc.Name,
			"oldSize": oldSize.Value(),
			"newSize": newSize,
		}).Info("Changing size of PVC.")
		err := c.resizePvc(dataPvc, newSize)
		if err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to resize PVC %s: %s", dataPvc.Name, err.Error())
		}
		needSize -= growSize
		if needSize == 0 {
			break
		}
	}
	// If after all the resizing, we still need more space, start making more PVCs until
	// we finally have enough space.
	for 0 < needSize {
		newSize := needSize
		if newSize > c.MaxPvcSize {
			newSize = c.MaxPvcSize
		}
		pvcName := fmt.Sprintf("data%d", len(dataPvcs)+1)
		log.WithFields(log.Fields{
			"netappObj": name,
			"PVC":       pvcName,
			"size":      newSize,
		}).Info("Creating a new data PVC for NetApp Object.")

		dataPvc, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(
			namespace).Create(
			context.TODO(),
			makeDataPvc(
				volume, storagePool, util.GetSCNameForFailureDomain(
					serviceLevel.Spec.DataVolumeStorageClass, failureDomain),
				len(dataPvcs), newSize, c.EnableMultiAttach,
			),
			metav1.CreateOptions{})
		if err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to create PVC for extra space requirement: %s", err.Error())
		}
		newDataVolume := corev1.Volume{
			Name: fmt.Sprintf("data%d", len(dataPvcs)+1),
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: dataPvc.Name,
				},
			},
		}
		newDataVolumesList = append(newDataVolumesList, newDataVolume)
		newDataDevice := makeDataVolumeDevice(len(dataPvcs) + 1)
		newDataVolumeDevicesList = append(newDataVolumeDevicesList, newDataDevice)
		dataPvcs = append(dataPvcs, dataPvc)
		needSize -= newSize
		diskAdd = true
	}

	return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, nil
}

// Check if volume needs to be expanded , if needed resize existing PVCs
// create new PVCs if it is a new deployment and there are no data pvcs yet
// return list of new data pvcs after going through expansion
func (c *Controller) syncVolumeExpansionAzure(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	totalRequestedSize int64,
	totalActualSize int64,
	desiredSpecSize int64,
	desiredSize int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string,
	diskPackingEnabled bool,
) ([]*corev1.PersistentVolumeClaim, []corev1.Volume, []corev1.VolumeDevice, bool, error) {
	var (
		newDataVolumesList       []corev1.Volume
		newDataVolumeDevicesList []corev1.VolumeDevice
		dataPvcs                 []*corev1.PersistentVolumeClaim
		diskAdd                  bool
		namespace, name          = GetNamespaceNameForNetappObj(volume, storagePool)
	)

	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(namespace).List(c.getDataPvcSelector(volume, storagePool))
	if err != nil {
		return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to list PVCs: %s", err.Error())
	}
	if len(dataPvcs) > 0 {
		log.WithFields(log.Fields{
			"netappObj":       name,
			"requestedSize":   totalRequestedSize,
			"actualSize":      totalActualSize,
			"desiredSpecSize": desiredSpecSize,
			"desiredSize":     desiredSize,
		}).Info("ProvisionedSize changed")
	}

	// Get nvram size
	nvramSize, err := c.getNvramSize(volume, storagePool, serviceLevel)
	if err != nil {
		return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err
	}

	// Get the layout of what sizes the data pvcs need to be at
	diskLayout := c.getAzureDiskLayout(volume, storagePool, len(dataPvcs), desiredSize, nvramSize, diskPackingEnabled)
	log.WithFields(log.Fields{
		"diskLayout": diskLayout,
	}).Info("diskLayout is")

	if len(dataPvcs) == 0 {
		// New deployment, create the required PVCs
		i := 0
		for pvcName, newSize := range diskLayout {
			log.WithFields(log.Fields{
				"netappObj": name,
				"PVC":       pvcName,
				"size":      newSize,
			}).Info("Creating a new data PVC for NetApp Volume.")
			pvcObj := makeDataPvc(
				volume, storagePool, util.GetSCNameForFailureDomain(
					serviceLevel.Spec.DataVolumeStorageClass, failureDomain),
				i, int64(newSize), c.EnableMultiAttach,
			)
			// Assign the correct name since maps aren't ordered
			pvcObj.Name = pvcName
			dataPvc, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(
				namespace).Create(context.TODO(), pvcObj,
				metav1.CreateOptions{})
			if err != nil {
				return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to create PVC for extra space requirement: %s", err.Error())
			}
			newDataVolume := corev1.Volume{
				Name: fmt.Sprintf("data%d", len(dataPvcs)+1),
				VolumeSource: corev1.VolumeSource{
					PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
						ClaimName: dataPvc.Name,
					},
				},
			}
			newDataVolumesList = append(newDataVolumesList, newDataVolume)
			newDataDevice := makeDataVolumeDevice(len(dataPvcs) + 1)
			newDataVolumeDevicesList = append(newDataVolumeDevicesList, newDataDevice)
			dataPvcs = append(dataPvcs, dataPvc)
			i++
		}
		diskAdd = true
	} else {
		// Resize scenario
		// Check if resize is needed
		// If not needed then just skip scaling and just wait for resize to complete
		if isResizeRequired(diskLayout, dataPvcs) {
			// Resize all the PVCs
			for _, pvc := range dataPvcs {
				log.Info("Resizing pvcs")
				err = c.resizePvc(pvc, int64(diskLayout[pvc.Name]))
				if err != nil {
					return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err
				}
			}
		}

		isResizeDone, err := c.isResizeDone(volume, storagePool)()
		if err != nil {
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, fmt.Errorf("failed to check if resize is done: %v", err)
		}
		if !isResizeDone {
			log.Info("resize not complete, will come back in when done")
			return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, DiskResizeInProgressError{}
		}

		log.Info("All pvcs have been resized")
	}

	return dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, nil
}

// Deploy Quark pod for volume if not present
// Check if existing quark pod volume needs to be expanded
// if new PVCs need to be added as part of expansion , update quark deployment
func (c *Controller) syncVolumeCreateExpansion(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	failureDomain string,
	quarkUUID string,
	quarkNvramID string,
	volumeUUID string,
	ownerID string,
	vserverUUID string,
	nbladeUUID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
	updateStatus *bool) (*quarkv1alpha1.NetAppVolume, *quarkv1alpha1.NetAppStoragePool, int64, int64, []quarkv1alpha1.NetAppVolumeCondition, *corev1.Event, error) {
	// Recalculate totalActualSize every time since we don't store allocated size which is delta + requestedSize
	// Add up the sizes of the PVCs
	var (
		totalRequestedSize, totalActualSize int64
		newDataVolumesList                  []corev1.Volume
		newDataVolumeDevicesList            []corev1.VolumeDevice
		volumeStateEvent                    *corev1.Event
		conditions                          []quarkv1alpha1.NetAppVolumeCondition
		desiredSize                         int64
		diskAdd                             bool
		needsMigration                      bool
		isPre10Vol                          bool
		desiredSpecSize                     int64
		specProvisionedSize                 int64
		statusProvisionedSize               int64
		specSize                            int64
		statusSize                          int64
		namespace, name                     = GetNamespaceNameForNetappObj(volume, storagePool)
		quarkVersion                        string
	)
	if storagePool == nil {
		conditions = volume.Status.Conditions
		isPre10Vol = isPre1_0VersionedVolume(volume)
		desiredSpecSize = volume.Spec.Size.Value()
		specSize = volume.Spec.Size.Value()
		statusSize = volume.Status.Size.Value()
		specProvisionedSize = volume.Spec.ProvisionedSize.Value()
		statusProvisionedSize = volume.Status.ProvisionedSize.Value()
		quarkVersion = volume.Status.QuarkVersion
	} else {
		desiredSpecSize = storagePool.Spec.Size.Value()
		specSize = storagePool.Spec.Size.Value()
		statusSize = storagePool.Status.Size.Value()
		specProvisionedSize = storagePool.Spec.ProvisionedSize.Value()
		statusProvisionedSize = storagePool.Status.ProvisionedSize.Value()
		quarkVersion = storagePool.Status.QuarkVersion
	}

	// Look up all the PVCs associated with this volume
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(namespace).List(c.getDataPvcSelector(volume, storagePool))
	if err != nil {
		return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to list PVCs: %s", err.Error())
	}

	for _, dataPvc := range dataPvcs {
		requestedSize := dataPvc.Spec.Resources.Requests[corev1.ResourceStorage]
		totalRequestedSize += requestedSize.Value()
		actualSize, ok := dataPvc.Status.Capacity[corev1.ResourceStorage]
		if ok {
			totalActualSize += actualSize.Value()
		}
	}

	volume, provisionedSize, err := c.getProvisionedSize(volume, storagePool, serviceLevel)
	if err != nil {
		return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get aggr size: %s", err.Error())
	}

	// During upgrade actualSize maybe higher. Set the correct provisionedSize.
	if !isPre10Vol && totalActualSize > provisionedSize && specProvisionedSize == 0 {
		provisionedSize = totalActualSize
	}
	if !isPre10Vol && provisionedSize > specProvisionedSize {
		if storagePool == nil {
			volume.Spec.ProvisionedSize = *resource.NewQuantity(provisionedSize, resource.BinarySI)
			volume, err = c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(namespace).Update(context.TODO(),
				volume, metav1.UpdateOptions{})
			if err != nil {
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to update provisionedSize: %s", err.Error())
			}
		} else {
			storagePool.Spec.ProvisionedSize = *resource.NewQuantity(provisionedSize, resource.BinarySI)
			storagePool, err = c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(namespace).Update(context.TODO(),
				storagePool, metav1.UpdateOptions{})
			if err != nil {
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to update provisionedSize: %s", err.Error())
			}
		}
	}

	volume, desiredSize, err = c.getTotalDiskSize(volume, storagePool, diskPackingEnabled, serviceLevel)
	if err != nil {
		return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get aggr size: %s", err.Error())
	}

	if totalRequestedSize < desiredSize || totalRequestedSize != totalActualSize {
		log.WithFields(log.Fields{
			"netappObj":       name,
			"requestedSize":   totalRequestedSize,
			"actualSize":      totalActualSize,
			"desiredSpecSize": desiredSpecSize,
			"desiredSize":     desiredSize,
		}).Info("ProvisionedSize changed")
		// Check if we are at the required size
		// if not at the required size expand current existing PVCs
		// if max size of the PVC is reached calculate how many new PVCs need to be added and create those PVCs
		// return pvc list after new pvcs are added if necessary
		switch c.Cloud {
		case "gcp", "ctl":
			dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err = c.syncVolumeExpansionGCP(volume, storagePool, totalRequestedSize, totalActualSize, desiredSpecSize, desiredSize, serviceLevel, failureDomain)
			if err != nil {
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err
			}
		case "azure":
			dataPvcs, newDataVolumesList, newDataVolumeDevicesList, diskAdd, err = c.syncVolumeExpansionAzure(volume, storagePool, totalRequestedSize, totalActualSize, desiredSpecSize, desiredSize, serviceLevel, failureDomain, diskPackingEnabled)
			if err != nil {
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err
			}
		default:
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("unknown cloud")
		}
	}

	if storagePool == nil {
		specProvisionedSize = volume.Spec.ProvisionedSize.Value()
		statusProvisionedSize = volume.Status.ProvisionedSize.Value()
	} else {
		specProvisionedSize = storagePool.Spec.ProvisionedSize.Value()
		statusProvisionedSize = storagePool.Status.ProvisionedSize.Value()
	}

	if specProvisionedSize != statusProvisionedSize && totalActualSize >= desiredSize {
		*updateStatus = true
	}

	dataCount := len(dataPvcs)

	// Wait for volume to be upgraded , if quark has been upgraded , before kicking off migration
	// If kicked off inbetween there will be a mismatch between the deployment yaml and image
	// Handle zone migration here too; if there's overlap, we should accomplish
	// them at the same time.
	if quarkVersion == "" || serviceLevel.Spec.QuarkVersion == quarkVersion {
		needsMigration, err = c.needQuarkPodMigration(volume, storagePool, nodeSelector)
		if err != nil {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	}
	if needsMigration {
		// To migrate quark pod to the right-sized VM, update node selector on quark deployment
		err := c.migrateQuarkPodVMRightSizing(volume, storagePool,
			serviceLevel, nodeSelector)
		if err != nil {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	} else if specSize != statusSize && totalActualSize >= desiredSize {
		// set volResize to true if it's not an initial deployment and if vol resize error condition does not exist.
		if storagePool == nil && statusSize != 0 && !existsVolResizeErrorCondition(volume) && !existsVolResizedCondition(volume) {
			timestamp := time.Now()
			newCondition := quarkv1alpha1.NetAppVolumeCondition{
				Type:               quarkv1alpha1.NetAppVolumeResized,
				Status:             corev1.ConditionTrue,
				LastHeartbeatTime:  metav1.NewTime(timestamp),
				LastTransitionTime: metav1.NewTime(timestamp),
				Reason:             "VolumeResized",
				Message:            "Resized PVCs",
			}
			conditions = append(conditions, newCondition)
		}
		*updateStatus = true
	}

	if !strings.EqualFold(c.Nodevol, "true") {
		// TODO: Remove this code when nodevol becomes the default
		err = c.syncQuarkvolPvc(volume, storagePool, serviceLevel, failureDomain)
		if err != nil {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	}

	if c.EnableMultiAttach {
		err = c.syncStorageDeployment(volume, storagePool, updateStatus, dataCount, dataPvcs, diskPackingEnabled)
		if err != nil {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, err
		}
	}
	// Try to find the Quark deployment
	quarkDeploy, err := c.deploysLister.Deployments(namespace).Get(GetQuarkDeployName(namespace, name))
	if err != nil {
		if errors.IsNotFound(err) {
			// If the Quark deployment doesn't exist yet, create it
			deploymentDefinition, err := c.makeInitialQuarkDeployment(volume,
				storagePool, serviceLevel, dataCount, quarkUUID, quarkNvramID, volumeUUID, ownerID,
				vserverUUID, nbladeUUID, exportAddress, nodeSelector, diskPackingEnabled)
			if err != nil {
				msg := fmt.Sprintf("failed to generate the definition for Quark deployment: %s", err.Error())
				c.recorder.Event(volume, corev1.EventTypeWarning, "ErrSyncVolume", msg)
				log.Errorf("Failed to generate the Quark deployment definition: %s", msg)
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, nil
			}
			quarkDeploy, err = c.KubeClientset.AppsV1().Deployments(namespace).Create(context.TODO(),
				deploymentDefinition, metav1.CreateOptions{})
			if err != nil {
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to create the quark deployment: %s", err.Error())
			}
			log.WithFields(log.Fields{
				"netappObj": name,
			}).Info("Created Quark deployment for netappObj.")
			*updateStatus = true
		} else {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to find quark deployment: %s", err.Error())
		}
	}

	// Make sure the deployment we found is controlled by us
	if !c.isControlledByNetappObj(quarkDeploy, volume, storagePool) {
		msg := fmt.Sprintf(quarkconfig.MessageResourceExists+",%s", quarkDeploy.Name, quarkDeploy.Kind)
		if storagePool != nil {
			c.recorder.Event(storagePool, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		} else {
			c.recorder.Event(volume, corev1.EventTypeWarning, quarkconfig.ErrResourceExists, msg)
		}
		return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf(msg)
	}
	// Make sure Deployment is scaled up
	// This can be due to resize timing out on azure and then succeeding between the functions calls
	if *quarkDeploy.Spec.Replicas == int32(0) {
		log.Debug("Scaling up deployment")
		// scale up deployment
		if err := c.scaleQuarkDeployment(namespace, name, int32(1)); err != nil {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to scale quark deployment : %s", err.Error())
		}
	}
	// Calculate the number of non-data volumes that exist in the deployment
	// example: nvram, quarkvol, supportabilitySecretsVolume, loggingVolume, exporter-config
	nonDataVolumes := countNonDataVolumes(quarkDeploy.Spec.Template.Spec.Volumes)

	// If the list of PVCs is not up to date, update the deployment
	if len(quarkDeploy.Spec.Template.Spec.Volumes) != len(dataPvcs)+nonDataVolumes {
		if diskAdd {
			// If the increase in provisioned size resulted in disk add, mark the volume offline before updating the deployment.
			// When pod restarts, podrick will mark volume as online
			if storagePool == nil && existsVolOnlineCondition(volume) {
				conditions = MakeNetAppVolumeConditionOffline(volume, quarkconfig.MessageDiskAdd)
				*updateStatus = true
				volumeStateEvent = &corev1.Event{
					Type:    corev1.EventTypeWarning,
					Reason:  quarkconfig.NetAppVolumeOfflined,
					Message: quarkconfig.MessageDiskAdd,
				}
			} else if storagePool != nil {
				// Marking the storagepool as offline.
				// TODO : Remove the volume offline part for storagepool configuration
				if existsSpOnlineCondition(storagePool) {
					spConditions := MakeNetAppStoragePoolConditionOffline(storagePool, quarkconfig.MessageDiskAdd)
					storagePoolStateEvent := &corev1.Event{
						Type:    corev1.EventTypeWarning,
						Reason:  quarkconfig.NetAppStoragePoolOfflined,
						Message: quarkconfig.MessageDiskAdd,
					}
					c.recorder.Event(storagePool, storagePoolStateEvent.Type, storagePoolStateEvent.Reason, storagePoolStateEvent.Message)
					storagePool.Status.Conditions = spConditions

					// Update the object status on the server
					storagePool, err = c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(storagePool.Namespace).UpdateStatus(context.TODO(),
						storagePool, metav1.UpdateOptions{})
					if err != nil {
						return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to add condition to storagepool status: %s", err.Error())
					}
				}
				volumes, err := NetappVolumesByStoragePool(c.volumesLister, storagePool.Namespace, storagePool.Name)
				if err != nil {
					return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get volumes from the quark deployment: %s", err.Error())
				}
				for _, volume := range volumes {
					if existsVolOnlineCondition(volume) {
						conditions = MakeNetAppVolumeConditionOffline(volume, quarkconfig.MessageDiskAdd)
						volumeStateEvent = &corev1.Event{
							Type:    corev1.EventTypeWarning,
							Reason:  quarkconfig.NetAppVolumeOfflined,
							Message: quarkconfig.MessageDiskAdd,
						}
						c.recorder.Event(volume, volumeStateEvent.Type, volumeStateEvent.Reason, volumeStateEvent.Message)
						volume.Status.Conditions = conditions

						// Update the object status on the server
						volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
							volume, metav1.UpdateOptions{})
						if err != nil {
							return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to add condition to volume status: %s", err.Error())
						}
					}
				}
			}
			// Update the Volume list and dmap's VolumeDevices list
			quarkDeploy.Spec.Template.Spec.Volumes = append(quarkDeploy.Spec.Template.Spec.Volumes,
				newDataVolumesList...)
			dmapContainer := util.GetContainerSpec(quarkDeploy.Spec.Template.Spec.Containers, quarkconfig.QuarkContainerDMAP)
			if dmapContainer == nil {
				return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to get dmap container spec")
			}
			dmapContainer.VolumeDevices = append(dmapContainer.VolumeDevices, newDataVolumeDevicesList...)
			// Update the QUARK_VOLUME_SIZE environment variable in case volume size changed
			if storagePool == nil {
				dmapContainer.Env = setEnvVar(dmapContainer.Env, "QUARK_VOLUME_SIZE", fmt.Sprintf("%d", volume.Spec.Size.Value()))
			}
			diskAdd = false
		}
		_, err = c.KubeClientset.AppsV1().Deployments(namespace).Update(context.TODO(),
			quarkDeploy, metav1.UpdateOptions{})
		if err != nil {
			return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, fmt.Errorf("failed to update the quark deployment: %s", err.Error())
		}
		log.WithFields(log.Fields{
			"currentVolumes": len(quarkDeploy.Spec.Template.Spec.Volumes),
			"desiredVolumes": len(dataPvcs) + nonDataVolumes,
			"netappObj":      name,
		}).Debug("Updated the Quark deployment.")
	}
	return volume, storagePool, totalActualSize, desiredSize, conditions, volumeStateEvent, nil
}

func (c *Controller) cleanUpAddress(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	addrName, finalizer string,
) error {
	var name, resourceVersion string
	if storagePool != nil {
		name = storagePool.Name
		resourceVersion = storagePool.Name
	} else {
		name = volume.Name
		resourceVersion = volume.Name
	}
	ctx := context.Background()
	// Ensure the static IP address associated with this volume is deleted
	_, err := c.cloudClient.GetLoadBalancerAddress(ctx, addrName)
	if err == nil {
		// If the address still exists, delete it
		log.WithFields(log.Fields{
			"netappObj":   name,
			"addressName": addrName,
			"finalizer":   finalizer,
		}).Info("Deleting static IP for netappObj.")
		if err = c.cloudClient.DeleteLoadBalancerAddress(ctx, addrName); err != nil {
			return fmt.Errorf("failed to delete the static IP address %s for "+
				"object %s: %v", addrName, name, err)
		}
	} else if !c.cloudClient.IsNotFound(err) {
		return fmt.Errorf("failed to get the static IP address %s for object %v: "+
			"%v", addrName, name, err)
	}

	// Remove the finalizer at last
	log.WithFields(log.Fields{
		"netappObj":       name,
		"resourceVersion": resourceVersion,
		"addressName":     addrName,
		"finalizer":       finalizer,
	}).Debug("Deleted address; removing finalizer.")
	if _, err = RemoveFinalizer(c.QuarkClientset, volume, storagePool,
		finalizer); err != nil {
		return fmt.Errorf("failed to remove address finalizer: %v", err)
	}
	log.WithFields(log.Fields{
		"netappObj":       name,
		"resourceVersion": resourceVersion,
		"addressName":     addrName,
		"finalizer":       finalizer,
	}).Debug("Deleted address; removed finalizer.")
	return nil
}

func (c *Controller) syncDeletingNetappObj(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) error {
	var (
		err             error
		netBios         string
		finalizer       []string
		namespace, name = GetNamespaceNameForNetappObj(volume, storagePool)
		volumes         []*quarkv1alpha1.NetAppVolume
		labelSelector   string
	)

	// protocol := "nfs3"
	deleteOptions := metav1.DeleteOptions{}

	if storagePool != nil {
		finalizer = storagePool.ObjectMeta.Finalizers
		netBios = storagePool.Status.SMBServerConfiguration.Netbios
		volumes, err = NetappVolumesByStoragePool(c.volumesLister, storagePool.Namespace, storagePool.Name)
		if err != nil {
			if !errors.IsNotFound(err) {
				log.WithFields(log.Fields{
					"storagePool": storagePool.Name,
				}).Error("Failed to get volumes from storage Pool")
				return err
			}
		}
	} else {
		// protocol = volume.Spec.Protocol
		finalizer = volume.ObjectMeta.Finalizers
		netBios = volume.Status.SMBConfiguration.Netbios
	}

	if c.regionalMeta != nil {
		c.regionalMeta.DeleteNetAppObj(name)
	}

	if storagePool == nil || (storagePool != nil && len(volumes) == 0) {
		switch c.Cloud {
		case quarkconfig.CloudGCP, quarkconfig.CloudAzure:
			if c.LoadBalancer {
				// Ensure the services are deleted only after finalizer is deleted for SMB in Podrick
				if HasFinalizer(finalizer, quarkconfig.SMBFinalizer) && netBios != "" {
					return nil
				}
				if storagePool == nil {
					labelSelector = fmt.Sprintf("%s=%s", quarkconfig.VolumeLabel, volume.Name)
				} else {
					labelSelector = fmt.Sprintf("%s=%s", quarkconfig.StoragePoolLabel, storagePool.Name)
				}
				listOptions := metav1.ListOptions{
					LabelSelector: labelSelector,
				}

				services, err := c.KubeClientset.CoreV1().Services(namespace).List(context.TODO(),
					listOptions)
				if err != nil {
					return fmt.Errorf("failed to retrieve load Balancer services: %s", err.Error())
				}

				for _, svc := range services.Items {
					_, err = c.servicesLister.Services(namespace).Get(svc.Name)
					if nil == err {
						// If the service still exists, delete it
						log.WithFields(log.Fields{
							"netappObj":   name,
							"serviceName": svc.Name,
						}).Info("Deleting load balancer service.")
						err = c.KubeClientset.CoreV1().Services(namespace).Delete(context.TODO(),
							svc.Name, deleteOptions)
						if nil != err {
							return fmt.Errorf("failed to delete the Load Balancer service: %s, err: %s", svc.Name, err.Error())
						}
					} else if !errors.IsNotFound(err) {
						return fmt.Errorf("failed to find the Load Balancer service for deletion: %s, err: %s", svc.Name, err.Error())
					}
				}

				// If any of service delete are still pending, do not try to delete ILB IP
				// It will be deleted on service delete callback, no need to requeue

				services, err = c.KubeClientset.CoreV1().Services(namespace).List(context.TODO(),
					listOptions)
				if err != nil {
					return fmt.Errorf("failed to retrieve load Balancer services: %s", err.Error())
				}
				if len(services.Items) > 0 {
					return nil
				}

				if HasFinalizer(finalizer,
					quarkconfig.ADClientAddressFinalizer) {
					if err = c.cleanUpAddress(volume, storagePool,
						getADClientAddrName(namespace, name),
						quarkconfig.ADClientAddressFinalizer,
					); err != nil {
						return err
					}
				}

				if err = c.cleanUpAddress(volume, storagePool,
					getAddrName(namespace, name), quarkconfig.AddressFinalizer,
				); err != nil {
					return err
				}
			}
		default:
			// We don't have to do anything here because for the non-load balancer
			// service, the ownerRef is set? Shouldn't it be the same for cloud load
			// balancer services?
		}

		//Cleanup Etcd Keys
		if storagePool != nil && HasFinalizer(finalizer, quarkconfig.EtcdKeyFinalizer) {
			if err = c.cleanUpEtcdKeysWithRetries(storagePool); err != nil {
				return err
			}
		}
	} else if len(volumes) > 0 {
		return fmt.Errorf("failed to delete storage pool, volume exist")
	}

	// TODO-SP: Remove it once ownerRef is added during upgrade, move this to podrick
	if volume != nil {
		labelSelector := fmt.Sprintf("quark.netapp.io/volume=%s", volume.Name)
		listOptions := metav1.ListOptions{
			LabelSelector: labelSelector,
		}
		// Auto-delete snapshot policy objects during volume deletion
		// This code is needed to remove Beta.0 policy objects
		// Owner reference is only added to Beta.1 policy objects
		// Keeping this code for backwards compatability
		snapshotPolicies, err := c.QuarkClientset.QuarkV1alpha1().NetAppSnapshotPolicies(namespace).List(context.TODO(),
			listOptions)
		if err != nil {
			return fmt.Errorf("failed to auto-delete snapshot policy objects: %s", err.Error())
		}
		for _, snapshotPolicy := range snapshotPolicies.Items {
			_ = c.QuarkClientset.QuarkV1alpha1().NetAppSnapshotPolicies(namespace).Delete(context.TODO(),
				snapshotPolicy.Name, metav1.DeleteOptions{})
			log.WithFields(log.Fields{
				"snapshotPolicy": snapshotPolicy.Name,
			}).Debug("Removed the snapshot policy object.")
		}
		// Auto-delete export policy objects during volume deletion
		exportPolicies, err := c.QuarkClientset.QuarkV1alpha1().NetAppExportPolicies(namespace).List(context.TODO(),
			listOptions)
		if err != nil {
			return fmt.Errorf("failed to auto-delete export policy objects: %s", err.Error())
		}
		for _, exportPolicy := range exportPolicies.Items {
			_ = c.QuarkClientset.QuarkV1alpha1().NetAppExportPolicies(namespace).Delete(context.TODO(),
				exportPolicy.Name, metav1.DeleteOptions{})
			log.WithFields(log.Fields{
				"exportPolicy": exportPolicy.Name,
			}).Debug("Removed the export policy object.")
		}
		c.metrics.KnownVolumes.Delete(volume.Name)
	}
	if storagePool != nil {
		c.metrics.KnownStoragePool.Delete(storagePool.Name)
	}
	return nil
}

func (c *Controller) getCloudSnapshotRestoreReserve(volume *quarkv1alpha1.NetAppVolume) (*quarkv1alpha1.NetAppVolume, int64, error) {
	// Check for cache size in this order:
	// 1. Status (we are done)
	// 2. Spec (copy to status)
	// 3. None (calculate based on cloud snapshot size, copy to status)
	if !volume.Status.RestoreCacheSize.IsZero() {
		return volume, volume.Status.RestoreCacheSize.Value(), nil
	}

	// Copy from spec
	restoreCacheSize := volume.Spec.RestoreCacheSize

	if restoreCacheSize.IsZero() {
		// Calculate based on cloud snapshot size
		cloudSnapshot, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumeCloudSnapshots(volume.Namespace).Get(context.TODO(),
			volume.Spec.CloudSnapshot, metav1.GetOptions{})
		if err != nil {
			return volume, 0, err
		}

		bytes := calculateCloudSnapshotRestoreReserve(cloudSnapshot.Status.LogicalSize)

		restoreCacheSize.Set(bytes)

		log.WithFields(log.Fields{
			"volume":            volume.Name,
			"cloudSnapshot":     cloudSnapshot.Name,
			"cloudSnapshotSize": cloudSnapshot.Status.LogicalSize,
			"cacheSize":         restoreCacheSize.String(),
		}).Debug("Calculated cloud snapshot restore cache.")
	}

	// Set cache size in status
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.RestoreCacheSize = restoreCacheSize
	volumeCopy, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(), volumeCopy, metav1.UpdateOptions{})

	return volumeCopy, restoreCacheSize.Value(), err
}

// Returns a map of data pvcs and the size they need to be at depending on the needSize
func (c *Controller) getAzureDiskLayout(volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	numOfDataPvcs int,
	desiredSize int64,
	nvramSize int64,
	diskPackingEnabled bool) map[string]float64 {
	diskLayout := make(map[string]float64)

	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)

	nvramSizeGB := nvramSize / quarkconfig.GiB
	if numOfDataPvcs == 0 {
		// This is an initial deployment , set to intial num of data disks
		numOfDataPvcs = c.MinDataPvcs
	}

	minPvcSizeGB := float64(c.MinPvcSize) / quarkconfig.GiB
	maxPvcSizeGB := float64(c.MaxPvcSize) / quarkconfig.GiB
	// We need to accomadate for nvram when diskpacking is enabled
	// We need closest power of 2 to minPvcSize + nvram size rounded up of space on the first disk
	minTotalFirstDisksize := minPvcSizeGB + float64(nvramSizeGB)
	MinFirstDiskDiskpackingSize := util.NearestPowerofTwo(minTotalFirstDisksize, quarkconfig.CeilRoundFunction) * quarkconfig.GiB
	// Calculate the size needed in GB
	needSize := float64(desiredSize) / quarkconfig.GiB
	// Calculate the size of each disk that would be needed
	averageDiskSize := needSize / float64(numOfDataPvcs)
	// Calculate the nearest power of 2 to the size of each disk
	// Minimum base disk size needs to be atleast MinPvcSize
	// Minimum base disk size can atmost be MaxPvcSize
	baseDiskSize := math.Min(maxPvcSizeGB, math.Max(minPvcSizeGB, util.NearestPowerofTwo(averageDiskSize, quarkconfig.FloorRoundFunction)))
	// Calculate how many disks we will need at the baseDiskSize to reach needSize
	// It needs to be atleast minDataPvcs
	numberofDisksWithBaseSize := math.Max(float64(c.MinDataPvcs), (needSize / baseDiskSize))
	// If numberofDisksWithBaseSize is greater than numOfDataPvcs , Calculate number of larger disks needed
	// Azure disks are always to the power of two.See https://confluence.ngage.netapp.com/display/CVO/Managed+Disk+Strategy for more details
	numberOfLargerDisks := int(math.Ceil(numberofDisksWithBaseSize - float64(numOfDataPvcs)))

	log.WithFields(log.Fields{
		"minPvcSizeGB":              minPvcSizeGB,
		"needSize":                  needSize,
		"averageDiskSize":           averageDiskSize,
		"baseDiskSize":              baseDiskSize,
		"numberofDisksWithBaseSize": numberofDisksWithBaseSize,
		"numberOfLargerDisks":       numberOfLargerDisks,
	}).Debug("Calculated sizes")

	// Build a map of the disk layout with size of each disk
	// The size of the disk is in Bytes
	for i := 0; i < numOfDataPvcs; i++ {
		if numberOfLargerDisks > 0 {
			// If we need larger disks , make all the disks larger
			// Disk size can be atmost maxPvcSize
			if i == 0 && diskPackingEnabled {
				// If diskpacking is enabled ,first disk needs to be atleast MinFirstDiskDiskpackingSize
				diskLayout[getDataPvcName(namespace, name, i)] = math.Max(MinFirstDiskDiskpackingSize, math.Min(float64(c.MaxPvcSize), 2*baseDiskSize*quarkconfig.GiB))
			} else {
				diskLayout[getDataPvcName(namespace, name, i)] = math.Min(float64(c.MaxPvcSize), 2*baseDiskSize*quarkconfig.GiB)
			}
		} else {
			if i == 0 && diskPackingEnabled {
				// If diskpacking is enabled ,first disk needs to be atleast MinFirstDiskDiskpackingSize
				diskLayout[getDataPvcName(namespace, name, i)] = math.Max(MinFirstDiskDiskpackingSize, math.Min(float64(c.MaxPvcSize), baseDiskSize*quarkconfig.GiB))
			} else {
				diskLayout[getDataPvcName(namespace, name, i)] = baseDiskSize * quarkconfig.GiB
			}
		}
	}
	return diskLayout
}

// Checks if the PVCs have been resized to expected size
func (c *Controller) isResizeDone(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) func() (bool, error) {
	return func() (bool, error) {
		namespace, _ := GetNamespaceNameForNetappObj(volume, storagePool)
		dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(namespace).List(c.getDataPvcSelector(volume, storagePool))
		if err != nil {
			return false, err
		}
		for _, pvc := range dataPvcs {
			requestedSize := pvc.Spec.Resources.Requests[corev1.ResourceStorage]
			actualSize, ok := pvc.Status.Capacity[corev1.ResourceStorage]
			if ok && actualSize.Value() == requestedSize.Value() {
				log.Infof("Pvc %s Resized successfully ", pvc.Name)
				continue
			}
			log.Infof("PVC %s size mismatch actualSize %d: expectedSize %d", pvc.Name, actualSize.Value(), requestedSize.Value())
			return false, nil
		}

		return true, nil
	}
}

func calculateCloudSnapshotRestoreReserve(size int64) int64 {
	var (
		snapGB              float64
		minSnapSize         float64
		maxSnapSize         float64
		sizeRatio           float64
		minSnapCachePercent float64
		maxSnapCachePercent float64
		minCacheSize        float64
		maxCacheSize        float64
		cacheRatio          float64
		a                   float64
		b                   float64
		cacheSize           float64
	)
	// Formula from: https://confluence.eng.netapp.com/display/~abhisar/Cbmap+cache+space+provisioning
	// All sizes need to be in GBs
	snapGB = float64(size) / quarkconfig.GiB

	// Someday we can parameterize these, but for now they are constant
	minSnapSize = 128.0
	maxSnapSize = 128 * float64(quarkconfig.TiB) / quarkconfig.GiB
	sizeRatio = maxSnapSize / minSnapSize
	minSnapCachePercent = 25.0
	maxSnapCachePercent = 0.78125

	// If less than minimum size, use minSize percent
	if snapGB <= minSnapSize {
		cacheSize = snapGB * minSnapCachePercent / 100
		return util.RoundUp(cacheSize*quarkconfig.GiB, quarkconfig.MiB)
	}

	minCacheSize = minSnapCachePercent * minSnapSize / 100
	maxCacheSize = maxSnapCachePercent * maxSnapSize / 100
	cacheRatio = maxCacheSize / minCacheSize

	b = math.Pow(2, math.Log2(cacheRatio)/math.Log2(sizeRatio))
	a = minCacheSize / math.Pow(b, math.Log2(minSnapSize))

	cacheSize = a * math.Pow(b, math.Log2(snapGB))

	return util.RoundUp(cacheSize*quarkconfig.GiB, quarkconfig.MiB)
}

func calculateCloudSnapshotCreateReserve(volSize int64) int64 {
	var volFloat = float64(volSize)

	vmapPercent := 0.2
	vmapSize := volFloat * vmapPercent / 100
	L1 := (volFloat / quarkconfig.MiB) * 8
	L2 := ((volFloat / quarkconfig.MiB) / 255) * 8
	objid := float64(quarkconfig.GiB)

	reserve := vmapSize + L1 + L2 + objid

	return int64(reserve)
}

// Get the required total disk size.
// This includes spaces needed for all the volumes and NVRAM size
// when disk packing will be enabled
func (c *Controller) getTotalDiskSize(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, diskPackingEnabled bool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*quarkv1alpha1.NetAppVolume, int64, error) {
	volume, aggrSize, err := c.getProvisionedSize(volume, storagePool, serviceLevel)
	if err == nil && diskPackingEnabled {
		// With disk packing, need to count the size of nvram as well
		nvramSize, err := c.getNvramSize(volume, storagePool, serviceLevel)
		if err != nil {
			return volume, 0, fmt.Errorf("invalid value of %d NVRAM Size obtained from service level spec: %s", nvramSize, err.Error())
		}
		aggrSize += nvramSize
	}
	return volume, aggrSize, err
}

// Get the required provisioned size for DMAP aggregate
func (c *Controller) getProvisionedSize(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, serviceLevel *quarkv1alpha1.NetAppServiceLevel) (*quarkv1alpha1.NetAppVolume, int64, error) {
	var (
		err                     error
		vol0Size                int64
		waflReserve             int64
		cloudSnapReserve        int64
		cloudSnapCreateReserve  int64
		cloudSnapRestoreReserve int64
		aggrSize                float64
		ntapVolSize             int64
		isPre10Vol              bool
		svl                     string
		specSize                int64
		provisionedSizeSpec     int64
	)

	if storagePool == nil {
		ntapVolSize = volume.Spec.Size.Value()
		isPre10Vol = isPre1_0VersionedVolume(volume)
		svl = volume.Spec.ServiceLevel
		specSize = volume.Spec.Size.Value()
		provisionedSizeSpec = volume.Spec.ProvisionedSize.Value()
	} else {
		ntapVolSize = storagePool.Spec.Size.Value()
		svl = storagePool.Spec.ServiceLevel
		specSize = storagePool.Spec.Size.Value()
		provisionedSizeSpec = storagePool.Spec.ProvisionedSize.Value()
	}

	if isPre10Vol || !c.thinProvisioningEnabled {

		// Add extra size for vol0 as follows
		// MIN(20GB, 12% of volume size)
		vol0Size = (ntapVolSize / 1024) * 123
		if vol0Size > quarkconfig.MaxVol0Size {
			vol0Size = quarkconfig.MaxVol0Size
		}

		// Allocate 16.8% (15% for wafl reserve + 1.8 for checksum and label)for
		// extra space based on the following calculation.
		// Each disk added will lose about 40 MB for storing labels, etc.
		// 1 block in 64 will be used for azcs checksum, that’s about 1.6%
		// 15%  of total volume size will be for wafl reserve.
		waflReserve = (ntapVolSize / 1024) * 172

		// Cloud snapshots need some extra space for performance reasons.
		// CS restore consumes much more space than CS creation and the two
		// operations cannot happen at the same time. So if we're doing a
		// restore, provision more, otherwise just provision what is needed
		// for CS creation.
		needCloudSnapRestoreReserve := false
		if storagePool == nil {
			if volume.Spec.CloudSnapshot != "" {
				needCloudSnapRestoreReserve = true
				for _, condition := range volume.Status.Conditions {
					if condition.Type == quarkv1alpha1.NetAppVolumeCloudSnapshotRestored {
						if condition.Status == corev1.ConditionTrue {
							needCloudSnapRestoreReserve = false
						}
						break
					}
				}
			}
			if needCloudSnapRestoreReserve {
				// Calculate size for cloud snapshot restore cache
				volume, cloudSnapRestoreReserve, err = c.getCloudSnapshotRestoreReserve(volume)
				if err != nil {
					return volume, 0, fmt.Errorf("failed to populate restore cache size: %s", err.Error())
				}
			}

			cloudSnapCreateReserve = calculateCloudSnapshotCreateReserve(ntapVolSize)

			if cloudSnapRestoreReserve > cloudSnapCreateReserve {
				cloudSnapReserve = cloudSnapRestoreReserve
			} else {
				cloudSnapReserve = cloudSnapCreateReserve
			}
		}

		// Total it up
		aggrSize = float64(ntapVolSize + vol0Size + waflReserve + cloudSnapReserve)
	} else {
		// With thin-provisioning, we don't need to reserve space for cloud snapshot restore cache but
		// we still need to set env QUARK_PHOTON_COPY_CACHE_SIZE
		if storagePool == nil && volume.Spec.CloudSnapshot != "" {
			// Calculate size for cloud snapshot restore cache
			volume, _, err = c.getCloudSnapshotRestoreReserve(volume)
			if err != nil {
				return volume, 0, fmt.Errorf("failed to populate restore cache size: %s", err.Error())
			}
		}
		quarkIOPSPerTiB, _, quarkCapacityLimit, diskIOPSPerGiB, _, tputLimit, quarkTpAmpFactor, err := util.GetServiceLevelLimits(svl, serviceLevel)
		if err != nil {
			return volume, 0, err
		}
		if quarkIOPSPerTiB == 0 || quarkCapacityLimit.Value() == 0 || diskIOPSPerGiB == 0 || quarkTpAmpFactor == 0 {
			return volume, 0, fmt.Errorf("serviceLevelLimits can not be zero, quarkIOPSPerTiB: %v, quarkCapacityLimit:%v, diskIOPSPerGiB:%v, quarkTpAmpFactor:%v ", quarkIOPSPerTiB, quarkCapacityLimit, diskIOPSPerGiB, quarkTpAmpFactor)
		}
		// using maxIOPS32k to determine promisedIOPS since it's the dominating factor among different models ->
		// https://confluence.ngage.netapp.com/display/CVO/STP+calculation+for+n1-32
		maxIOPS32k := tputLimit * 1000 / 32
		quarkIOPSPerGiB := (float64(quarkIOPSPerTiB) / float64(1024))
		promisedIOPS := (quarkIOPSPerGiB) * float64(specSize/(quarkconfig.GiB))
		promisedIOPS = math.Min(promisedIOPS, float64(maxIOPS32k))
		aggrSizeGiB := ((float64(promisedIOPS) * float64(quarkTpAmpFactor)) / float64(diskIOPSPerGiB)) * quarkconfig.AggrTpFactor
		aggrSize = aggrSizeGiB * quarkconfig.GiB
		if float64(provisionedSizeSpec) > aggrSize {
			aggrSize = float64(provisionedSizeSpec)
		}
	}

	// Round off to the nearest multiple of c.MinPvcSize
	roundedSize := util.RoundUp(aggrSize, c.MinPvcSize)

	return volume, roundedSize, nil
}

func (c *Controller) scaleQuarkDeployment(namespace string, name string, replicas int32) error {
	quarkDeployName := GetQuarkDeployName(namespace, name)
	var (
		d   *appsv1.Deployment
		err error
	)
	if d, err = c.deploysLister.Deployments(namespace).Get(quarkDeployName); err != nil {
		return fmt.Errorf("failed to scale deployment %s: %s", name, err.Error())
	}
	d.Spec.Replicas = &replicas
	if _, err = c.KubeClientset.AppsV1().Deployments(namespace).Update(context.TODO(), d, metav1.UpdateOptions{}); err != nil {
		return fmt.Errorf("failed to scale deployment %s: %s", name, err.Error())
	}
	return err
}

// Change the size on an existing PVC
func (c *Controller) resizePvc(pvc *corev1.PersistentVolumeClaim, newSize int64) error {
	pvcCopy := pvc.DeepCopy()
	// Check if a resize is needed
	if actualSize, ok := pvcCopy.Spec.Resources.Requests[corev1.ResourceStorage]; !ok || actualSize.Value() != newSize {
		pvcCopy.Spec.Resources.Requests[corev1.ResourceStorage] = *resource.NewQuantity(newSize, resource.BinarySI)
		_, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(pvc.Namespace).Update(context.TODO(),
			pvcCopy, metav1.UpdateOptions{})
		if err != nil {
			return fmt.Errorf("failed to resize PVC %s: %s", pvc.Name, err.Error())
		}
	}
	return nil
}

// isDetachDoneFunc fetches node with `nodeName` and returns false if
// volume is listed in the node.status.attachedVolumes list, or true if the volume
// is not in the list volume names in attachedVolumes use fully qualified path
// e.g. /subscriptionID/resourceGroupID/.../pvcName
func isDetachDoneFunc(client kubernetes.Interface, nodeName, pvcName string) func() (bool, error) {
	return func() (bool, error) {
		node, err := client.CoreV1().Nodes().Get(context.TODO(),
			nodeName, metav1.GetOptions{})
		if err != nil {
			return false, fmt.Errorf("error getting node: %v", err)
		}
		for _, disk := range node.Status.VolumesAttached {
			matched, err := regexp.MatchString(pvcName, string(disk.Name))
			if err != nil {
				return false, fmt.Errorf("error matching string: %v", err)
			}
			if matched {
				log.WithFields(log.Fields{
					"nodeName": nodeName,
					"pvcName":  pvcName,
					"diskName": disk.Name,
				}).Infof("Volume is attached to node.")
				return false, nil
			}
		}
		log.WithFields(log.Fields{
			"nodeName": nodeName,
			"pvcName":  pvcName,
		}).Infof("Volume is detached from node.")
		return true, nil
	}
}

// updateNetAppVolumeUUIDStatus updates UUIDs in status of a NetAppVolume CRD
func (c *Controller) updateNetAppObjectUUIDStatus(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	quarkUUID string,
	quarkNvramID string,
	volumeUUID string,
	vserverUUID string,
	nbladeUUID string) (*quarkv1alpha1.NetAppVolume, *quarkv1alpha1.NetAppStoragePool, error) {
	if storagePool != nil {
		storagePoolCopy := storagePool.DeepCopy()
		storagePoolCopy.Status.QuarkUUID = quarkUUID
		storagePoolCopy.Status.QuarkNvramID = quarkNvramID
		storagePoolCopy.Status.QuarkVserverUUID = vserverUUID
		storagePoolCopy.Status.QuarkNbladeUUID = nbladeUUID
		// Update the object status on the server
		storagePool, err := c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(storagePool.Namespace).UpdateStatus(context.TODO(),
			storagePoolCopy, metav1.UpdateOptions{})
		if err != nil {
			return volume, storagePool, fmt.Errorf("failed to update UUIDs in NetAppstoragePool CRD status: %s", err.Error())
		}
	} else {
		volumeCopy := volume.DeepCopy()
		volumeCopy.Status.QuarkUUID = quarkUUID
		volumeCopy.Status.QuarkNvramID = quarkNvramID
		volumeCopy.Status.VolumeUUID = volumeUUID
		// Update the object status on the server
		volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
			volumeCopy, metav1.UpdateOptions{})
		if err != nil {
			return volume, storagePool, fmt.Errorf("failed to update UUIDs in NetAppVolume CRD status: %s", err.Error())
		}
	}
	return volume, storagePool, nil
}

// updateNetAppObjectStatus updates the status of a NetAppObject CRD
func (c *Controller) updateNetAppObjectStatus(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	provisionedSize int64,
	dataPvcs []*corev1.PersistentVolumeClaim,
	exportAddress string,
	conditions []quarkv1alpha1.NetAppVolumeCondition,
	spConditions []quarkv1alpha1.NetAppStoragePoolCondition,
	diskPackingEnabled bool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	connectionTrackingPolicy string,
) (*quarkv1alpha1.NetAppVolume, *quarkv1alpha1.NetAppStoragePool, error) {
	if storagePool != nil {
		storagePoolCopy := storagePool.DeepCopy()
		storagePoolCopy.Status.DataPvcs = make([]string, len(dataPvcs))
		for i, dataPvc := range dataPvcs {
			storagePoolCopy.Status.DataPvcs[i] = dataPvc.Name
		}

		if !diskPackingEnabled {
			storagePoolCopy.Status.NvramPvc = getNvramPvcName(storagePool.Namespace, storagePool.Name)
		}

		if storagePoolCopy.Status.NvramSize == 0 {
			// Need to get the nvram size out of the serviceLevel
			nvramSize, err := c.getNvramSize(nil, storagePool, serviceLevel)
			if err != nil {
				storagePoolCopy.Status.NvramSize = 0
				return volume, storagePool, fmt.Errorf("failed to determine the quark pod NVRAM size: %s", err.Error())
			} else {
				storagePoolCopy.Status.NvramSize = nvramSize
			}
		}

		log.WithFields(log.Fields{
			"storagePool": storagePool.Name,
			"conditions":  len(spConditions),
		}).Debug("Updating storage pool status.")

		storagePoolCopy.Status.ServiceLevel = storagePool.Spec.ServiceLevel
		storagePoolCopy.Status.ExportAddress = exportAddress
		storagePoolCopy.Status.Conditions = spConditions
		storagePoolCopy.Status.Size = storagePool.Spec.Size
		if c.EnableMultiAttach {
			storagePoolCopy.Status.StorageDeployment = getStorageDeployName(storagePool.Namespace, storagePool.Name)
		}
		storagePoolCopy.Status.QuarkDeployment = GetQuarkDeployName(storagePool.Namespace, storagePool.Name)
		storagePoolCopy.Status.ProvisionedSize = *resource.NewQuantity(provisionedSize, resource.BinarySI)
		storagePoolCopy.Status.IlbConnectionTrackingPolicy = connectionTrackingPolicy

		// Update the object status on the server
		storagePool, err := c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(storagePool.Namespace).UpdateStatus(context.TODO(),
			storagePoolCopy, metav1.UpdateOptions{})
		if err != nil {
			return volume, storagePool, fmt.Errorf("failed to update UUIDs in NetAppVolume CRD status: %s", err.Error())
		}
		return volume, storagePool, nil
	} else {
		namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
		volumeCopy := volume.DeepCopy()
		volumeCopy.Status.StorageDeployment = getStorageDeployName(namespace, name)
		volumeCopy.Status.QuarkDeployment = GetQuarkDeployName(namespace, name)
		if !diskPackingEnabled {
			volumeCopy.Status.NvramPvc = getNvramPvcName(namespace, name)
		}
		if volumeCopy.Status.NvramSize == 0 {
			// Need to get the nvram size out of the serviceLevel
			nvramSize, err := c.getNvramSize(volume, storagePool, serviceLevel)
			if err != nil {
				volumeCopy.Status.NvramSize = 0
				return volume, storagePool, fmt.Errorf("failed to determine the quark pod NVRAM size: %s", err.Error())
			} else {
				volumeCopy.Status.NvramSize = nvramSize
			}
		}
		volumeCopy.Status.DataPvcs = make([]string, len(dataPvcs))
		volumeCopy.Status.ServiceLevel = volume.Spec.ServiceLevel
		volumeCopy.Status.ExportAddress = exportAddress
		volumeCopy.Status.Conditions = conditions
		if !isPre1_0VersionedVolume(volume) {
			volumeCopy.Status.ProvisionedSize = *resource.NewQuantity(provisionedSize, resource.BinarySI)
		}
		for i, dataPvc := range dataPvcs {
			volumeCopy.Status.DataPvcs[i] = dataPvc.Name
		}

		// Update the object status on the server
		volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
			volumeCopy, metav1.UpdateOptions{})
		if err != nil {
			return volume, storagePool, fmt.Errorf("failed to update object status on the server: %s", err.Error())
		}
		return volume, storagePool, nil
	}
}

// updateNetAppVolumeDNSStatus updates the status of a NetAppVolume CRD
func (c *Controller) updateNetAppVolumeDNSStatus(volume *quarkv1alpha1.NetAppVolume,
	conditions []quarkv1alpha1.NetAppVolumeCondition) (*quarkv1alpha1.NetAppVolume, error) {
	volumeCopy := volume.DeepCopy()
	volumeCopy.Status.SMBConfiguration.DNSIPs = volumeCopy.Spec.SMBConfiguration.DNSIPs
	volumeCopy.Status.Conditions = conditions
	// Update the object status on the server
	volume, err := c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return volume, fmt.Errorf("failed to update NetAppVolume status on the server: %s", err.Error())
	}
	return volume, nil
}

// updateNetAppObjectExportAddress updates only the export address status of a NetAppVolume CR.
func (c *Controller) updateNetAppObjectExportAddress(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, exportAddress string) (*quarkv1alpha1.NetAppVolume, *quarkv1alpha1.NetAppStoragePool, error) {
	var (
		updatedVolume      *quarkv1alpha1.NetAppVolume
		updatedStoragePool *quarkv1alpha1.NetAppStoragePool
		err                error
	)
	if storagePool != nil {
		// Copy and replace exportAddress
		storagePoolCopy := storagePool.DeepCopy()
		storagePoolCopy.Status.ExportAddress = exportAddress
		updatedStoragePool, err = c.QuarkClientset.QuarkV1alpha1().NetAppStoragePools(storagePool.Namespace).UpdateStatus(context.TODO(),
			storagePoolCopy, metav1.UpdateOptions{})

		if err != nil {
			return nil, nil, fmt.Errorf("failed to update object export address status on the server: %s", err.Error())
		}

		log.WithFields(log.Fields{
			"storagePool":   storagePool.Name,
			"exportAddress": storagePoolCopy.Status.ExportAddress,
			"success":       (err == nil),
		}).Debug("Updated Netapp StoragePool exportAddress.")
	} else {
		// Copy and replace exportAddress
		volumeCopy := volume.DeepCopy()
		volumeCopy.Status.ExportAddress = exportAddress
		updatedVolume, err = c.QuarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).UpdateStatus(context.TODO(),
			volumeCopy, metav1.UpdateOptions{})

		if err != nil {
			return nil, nil, fmt.Errorf("failed to update object export address status on the server: %s", err.Error())
		}

		// Log action outcome
		log.WithFields(log.Fields{
			"volume":        volume.Name,
			"exportAddress": volumeCopy.Status.ExportAddress,
			"success":       (err == nil),
		}).Debug("Updated Netapp Volume exportAddress.")

	}
	return updatedVolume, updatedStoragePool, nil
}

// Add the finalizer to the CRD object
func AddFinalizer(quarkClientset quarkclientset.Interface, volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, finalizer string) (*quarkv1alpha1.NetAppVolume, *quarkv1alpha1.NetAppStoragePool, error) {
	if storagePool != nil {
		storagePoolCopy := storagePool.DeepCopy()
		// Find if the finalizer is already present
		idx := util.FindStringInSlice(storagePoolCopy.ObjectMeta.Finalizers, finalizer)
		if 0 <= idx {
			// It was found
			return volume, storagePool, nil
		}
		// Append the finalizer
		storagePoolCopy.ObjectMeta.Finalizers = append(storagePoolCopy.ObjectMeta.Finalizers, finalizer)
		log.WithFields(log.Fields{
			"storagePool":     storagePoolCopy.Name,
			"resourceVersion": storagePoolCopy.ResourceVersion,
			"finalizer":       finalizer,
		}).Debug("Adding finalizer to SP")
		// Update the object on the server
		storagePool, err := quarkClientset.QuarkV1alpha1().NetAppStoragePools(storagePool.Namespace).Update(context.TODO(),
			storagePoolCopy, metav1.UpdateOptions{})
		if err != nil {
			return volume, storagePool, fmt.Errorf("failed to append finalizer %s to the CRD object: %s", finalizer, err.Error())
		}
		return volume, storagePool, nil
	}

	volumeCopy := volume.DeepCopy()
	// Find if the finalizer is already present
	idx := util.FindStringInSlice(volumeCopy.ObjectMeta.Finalizers, finalizer)
	if 0 <= idx {
		// It was found
		return volume, storagePool, nil
	}
	// Append the finalizer
	volumeCopy.ObjectMeta.Finalizers = append(volumeCopy.ObjectMeta.Finalizers, finalizer)
	// Update the object on the server
	volume, err := quarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).Update(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return volume, storagePool, fmt.Errorf("failed to append finalizer %s to the CRD object: %s", finalizer, err.Error())
	}
	return volume, storagePool, nil
}

// Remove the finalizer from the CRD object
func RemoveFinalizer(
	quarkClientset quarkclientset.Interface,
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	finalizer string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	var err error
	if storagePool != nil {
		// Find if the finalizer is still present
		idx := util.FindStringInSlice(storagePool.ObjectMeta.Finalizers,
			finalizer)
		if 0 > idx {
			// It wasn't found
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"finalizer":   finalizer,
			}).Debug("Finalizer already removed.")
			return storagePool, nil
		}
		storagePoolCopy := storagePool.DeepCopy()
		// Remove the array element
		storagePoolCopy.ObjectMeta.Finalizers = append(
			storagePoolCopy.ObjectMeta.Finalizers[:idx],
			storagePoolCopy.ObjectMeta.Finalizers[idx+1:]...)
		// Update the object on the server
		storagePool, err = quarkClientset.QuarkV1alpha1().NetAppStoragePools(
			storagePool.Namespace).Update(context.TODO(), storagePoolCopy,
			metav1.UpdateOptions{})
		if err != nil {
			return nil, fmt.Errorf("failed to remove finalizer %s from storage"+
				" pool %s: %v", finalizer, storagePool.Name, err)
		}
		return storagePool, nil
	}
	volumeCopy := volume.DeepCopy()
	// Find if the finalizer is still present
	idx := util.FindStringInSlice(volumeCopy.ObjectMeta.Finalizers, finalizer)
	if 0 > idx {
		// It wasn't found
		return nil, nil
	}
	// Remove the array element
	volumeCopy.ObjectMeta.Finalizers = append(volumeCopy.ObjectMeta.Finalizers[:idx], volumeCopy.ObjectMeta.Finalizers[idx+1:]...)
	// Update the object on the server
	_, err = quarkClientset.QuarkV1alpha1().NetAppVolumes(volume.Namespace).Update(context.TODO(),
		volumeCopy, metav1.UpdateOptions{})
	if err != nil {
		return nil, fmt.Errorf("failed to remove finalizer %s from CRD object: %s", finalizer, err.Error())
	}
	return nil, nil
}

// Generate a QuarkVol PVC definition
func makeQuarkVolPvc(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, scName string, size int64, enableMultiAttach bool) *corev1.PersistentVolumeClaim {
	mode := corev1.PersistentVolumeFilesystem
	accessMode := corev1.ReadWriteOnce
	if enableMultiAttach {
		accessMode = corev1.ReadWriteMany
	}
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{accessMode},
			VolumeMode:  &mode,
			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceStorage: *resource.NewQuantity(size, resource.BinarySI),
				},
			},
		},
	}
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	pvc.Name = GetQuarkVolPvcName(namespace, name)
	if scName != "" {
		pvc.Spec.StorageClassName = &scName
	}
	if storagePool == nil {
		pvc.Labels = map[string]string{
			quarkconfig.TypeLabel:   "quarkVol",
			quarkconfig.VolumeLabel: name,
		}
	} else {
		pvc.Labels = map[string]string{
			quarkconfig.TypeLabel:        "quarkVol",
			quarkconfig.StoragePoolLabel: name,
		}
	}

	return pvc

}

// Generate an NVRAM PVC definition
func makeNvramPvc(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, scName string, size int64, enableMultiAttach bool) *corev1.PersistentVolumeClaim {
	mode := corev1.PersistentVolumeBlock
	accessMode := corev1.ReadWriteOnce
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)

	if enableMultiAttach {
		accessMode = corev1.ReadWriteMany
	}
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{accessMode},
			VolumeMode:  &mode,
			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceStorage: *resource.NewQuantity(size, resource.BinarySI),
				},
			},
		},
	}
	pvc.Name = getNvramPvcName(namespace, name)
	if scName != "" {
		pvc.Spec.StorageClassName = &scName
	}
	if storagePool == nil {
		pvc.Labels = map[string]string{
			quarkconfig.TypeLabel:   "nvram",
			quarkconfig.VolumeLabel: name,
		}
	} else {
		pvc.Labels = map[string]string{
			quarkconfig.TypeLabel:        "nvram",
			quarkconfig.StoragePoolLabel: name,
		}
	}

	return pvc
}

// Generate a data PVC definition
func makeDataPvc(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, scName string, n int, size int64, enableMultiAttach bool) *corev1.PersistentVolumeClaim {
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	mode := corev1.PersistentVolumeBlock
	accessMode := corev1.ReadWriteOnce
	if enableMultiAttach {
		accessMode = corev1.ReadWriteMany
	}
	pvc := &corev1.PersistentVolumeClaim{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: corev1.PersistentVolumeClaimSpec{
			AccessModes: []corev1.PersistentVolumeAccessMode{accessMode},
			VolumeMode:  &mode,
			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceStorage: *resource.NewQuantity(size, resource.BinarySI),
				},
			},
		},
	}
	pvc.Name = getDataPvcName(namespace, name, n)
	if scName != "" {
		pvc.Spec.StorageClassName = &scName
	}

	if storagePool == nil {
		pvc.Labels = map[string]string{
			quarkconfig.TypeLabel:   "data",
			quarkconfig.VolumeLabel: name,
		}
	} else {
		pvc.Labels = map[string]string{
			quarkconfig.TypeLabel:        "data",
			quarkconfig.StoragePoolLabel: name,
		}
	}
	return pvc
}

// Generate a storage deployment definition
func (c *Controller) makeStorageDeployment(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, dataCount int, diskPackingEnabled bool) *appsv1.Deployment {
	var (
		protocol string
		labels   map[string]string
	)
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	if storagePool == nil {
		labels = map[string]string{
			quarkconfig.AppLabel:    "storage",
			quarkconfig.VolumeLabel: name,
		}
		protocol = volume.Spec.Protocol
	} else {
		labels = map[string]string{
			quarkconfig.AppLabel:         "storage",
			quarkconfig.StoragePoolLabel: name,
		}
	}
	replicas := int32(3)
	deploy := &appsv1.Deployment{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: labels,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels: labels,
				},
				Spec: corev1.PodSpec{
					Containers: []corev1.Container{
						{
							Name:  "storage",
							Image: c.StorageImageName,

							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),
							VolumeDevices:   c.makeVolumeDevices(dataCount, diskPackingEnabled),
						},
					},
					Volumes: c.makeVolumes(namespace, name, protocol, dataCount, diskPackingEnabled),
					Affinity: &corev1.Affinity{
						PodAntiAffinity: &corev1.PodAntiAffinity{
							RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
								{
									LabelSelector: &metav1.LabelSelector{
										MatchLabels: labels,
									},
									TopologyKey: "kubernetes.io/hostname",
								},
							},
						},
					},
				},
			},
		},
	}
	deploy.Name = getStorageDeployName(namespace, name)
	deploy.Labels = labels
	return deploy
}

func (c *Controller) generateNvramID() string {
	r := rand.Uint32()
	s := fmt.Sprintf("%d", uint32(r))
	return s
}

func (c *Controller) makeMrootVolumeMount(name string) corev1.VolumeMount {
	if strings.EqualFold(c.Nodevol, "true") {
		return corev1.VolumeMount{
			Name:      "quarkvol",
			MountPath: "/mroot",
			SubPath:   name,
		}
	} else {
		return corev1.VolumeMount{
			Name:      "quarkvol",
			MountPath: "/mroot",
		}
	}
}

// Generate a Quark deployment definition
// If runDmap is true, this starts dmap with standard boot arguments, liveness
// probe, and prestop.  If false, it starts dmap with a no-op command and
// removes the liveness probe and prestop.
func (c *Controller) makeQuarkDeployment(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	quarkVolumeUUID string,
	ownerID string,
	vserverUUID string,
	nbladeUUID string,
	exportAddress string,
	podPriorityClass string,
	nodeSelector map[string]string,
	runDmap bool,
	diskPackingEnabled bool,
) (*appsv1.Deployment, error) {
	var (
		dmapProtocolPorts []corev1.ContainerPort
		svl               string
		labels            map[string]string
		quarkVolumeSize   string
		quarkVersion      string
		specSize          int64
		quarkVolumeName   string
		selectorLabel     map[string]string
	)

	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	protocol := "nfs3"
	if storagePool == nil {
		svl = volume.Spec.ServiceLevel
		labels = map[string]string{
			quarkconfig.AppLabel:    "quark",
			quarkconfig.VolumeLabel: name,
		}
		selectorLabel = labels
		quarkVolumeSize = fmt.Sprintf("%d", volume.Spec.Size.Value())
		quarkVolumeName = strings.Replace("vol_"+volume.Spec.VolumePath, "-", "_", -1)
		specSize = volume.Spec.Size.Value()
		protocol = volume.Spec.Protocol
		quarkVersion = volume.Spec.QuarkVersion
	} else {
		svl = storagePool.Spec.ServiceLevel
		labels = map[string]string{
			quarkconfig.AppLabel:         "quark",
			quarkconfig.StoragePoolLabel: name,
			quarkconfig.VolumeLabel:      name,
		}
		selectorLabel = map[string]string{
			quarkconfig.AppLabel:    "quark",
			quarkconfig.VolumeLabel: name,
		}
		quarkVolumeSize = fmt.Sprintf("%d", storagePool.Spec.Size.Value())
		specSize = storagePool.Spec.Size.Value()
		quarkVersion = storagePool.Spec.QuarkVersion
		// TODO: workaround needs to be removed after DMAP changes
		quarkVolumeName = strings.Replace("vol_"+storagePool.Spec.DisplayName, "-", "_", -1)
	}

	// TODO: Should it be volume.Spec.DisplayName_VolumePath?

	tolerationSecondsFive := int64(5)
	tolerationSecondsThirty := int64(30)

	quarkIOPSPerTiB, quarkThroughput, quarkCapacityLimit, _, quarkIopsLimit, quarkTputLimit, _, err := util.GetServiceLevelLimits(
		svl, serviceLevel)
	log.WithFields(log.Fields{
		"name":          name,
		"serviceLevel":  serviceLevel,
		"IOPSPerTiB":    quarkIOPSPerTiB,
		"MiBsPerTiB":    quarkThroughput,
		"CapacityLimit": quarkCapacityLimit,
		"IopsLimit":     quarkIopsLimit,
		"TputLimit":     quarkTputLimit,
	}).Debug("Determined serviceLevel limits.")
	if err != nil {
		return nil, fmt.Errorf("invalid QoS limits provided: %s", err)
	}

	annotations := map[string]string{
		quarkconfig.PrometheusScrapeAnnotation: "true",
		quarkconfig.PrometheusPortAnnotation:   "9340",
	}
	dmapCPUsRequest, dmapCPUsLimit, err := c.getCPUs(name, specSize, serviceLevel,
		quarkconfig.QuarkContainerDMAP)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", dmapCPUsRequest.String(), dmapCPUsLimit.String(), err)
	}
	dmapMemRequest, dmapMemLimit, err := c.getMem(name, specSize, serviceLevel,
		quarkconfig.QuarkContainerDMAP)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", dmapMemRequest.String(), dmapMemLimit.String(), err)
	}
	podrickCPUsRequest, podrickCPUsLimit, err := c.getCPUs(name, specSize, serviceLevel,
		quarkconfig.QuarkContainerPodrick)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", podrickCPUsRequest.String(), podrickCPUsLimit.String(), err)
	}
	podrickMemRequest, podrickMemLimit, err := c.getMem(name, specSize, serviceLevel,
		quarkconfig.QuarkContainerPodrick)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", podrickMemRequest.String(), podrickMemLimit.String(), err)
	}
	supportabilityCPUsReq, supportabilityCPUsLimit, err := c.getCPUs(name, specSize,
		serviceLevel, quarkconfig.QuarkContainerSupportability)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", supportabilityCPUsReq.String(), supportabilityCPUsLimit.String(), err)
	}
	supportabilityMemReq, supportabilityMemLimit, err := c.getMem(name, specSize,
		serviceLevel, quarkconfig.QuarkContainerSupportability)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", supportabilityMemReq.String(), supportabilityMemLimit.String(), err)
	}
	loggingCPUsReq, loggingCPUsLimit, err := c.getCPUs(name, specSize, serviceLevel,
		quarkconfig.QuarkContainerLogging)
	if err != nil {
		return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", loggingCPUsReq.String(), loggingCPUsLimit.String(), err)
	}
	loggingMemReq, loggingMemLimit, err := c.getMem(name, specSize, serviceLevel,
		quarkconfig.QuarkContainerLogging)
	if err != nil {
		return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", loggingMemReq.String(), loggingMemLimit.String(), err)
	}

	schedulerName, err := c.getSchedulerName(serviceLevel)
	if err != nil {
		return nil, fmt.Errorf("invalid value of %s scheduler name for the Quark deployment: %v", schedulerName, err)
	}

	nvramSize, err := c.getNvramSize(volume, storagePool, serviceLevel)
	if err != nil {
		return nil, fmt.Errorf("invalid value of %d NVRAM Size obtained from service level spec: %v", nvramSize, err)
	}

	dmsCPUs, err := c.getDmsCPUs(volume, storagePool, serviceLevel)
	if err != nil {
		return nil, fmt.Errorf("invalid value of %d DMS CPUs obtained from service level spec: %v", dmsCPUs, err)
	}

	imageLookup := c.getImageTags(quarkVersion)

	// A pod with low priority will trigger a new node allocation without displacing
	// one of the spare Quark pods.
	priorityClass := podPriorityClass

	// Set up the shell commands that dmap should be using
	dmapLivenessProbe := GetDmapLivenessProbe(c.QuarkVariant)
	dmapPrestop := GetDmapPrestop()
	dmapArgs := GetDmapArgs()

	// If we're not running dmap (currently, when the Quark volume is first
	// deployed), run nop commands instead.
	if !runDmap {
		dmapLivenessProbe = nil
		dmapPrestop = nil
		dmapArgs = []string{"while true; do sleep 1; done"}
	}

	// Only 1 replica
	replicas := int32(1)
	deploy := &appsv1.Deployment{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: appsv1.DeploymentSpec{
			Replicas: &replicas,
			Selector: &metav1.LabelSelector{
				MatchLabels: selectorLabel,
			},
			Template: corev1.PodTemplateSpec{
				ObjectMeta: metav1.ObjectMeta{
					Labels:      labels,
					Annotations: annotations,
				},
				Spec: corev1.PodSpec{
					SchedulerName:      schedulerName,
					ServiceAccountName: "podrick-account",
					PriorityClassName:  priorityClass,
					SecurityContext: &corev1.PodSecurityContext{
						Sysctls: []corev1.Sysctl{
							{
								Name:  "net.ipv4.ip_local_port_range",
								Value: "32768 60999",
							},
						},
					},
					NodeSelector: nodeSelector,
					Containers: []corev1.Container{
						{
							Name:  quarkconfig.QuarkContainerDMAP,
							Image: imageLookup[quarkconfig.QuarkContainerDMAP],

							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),
							VolumeDevices:   c.makeVolumeDevices(dataCount, diskPackingEnabled),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    dmapCPUsRequest,
									corev1.ResourceMemory: dmapMemRequest,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    dmapCPUsLimit,
									corev1.ResourceMemory: dmapMemLimit,
								},
							},
							LivenessProbe: dmapLivenessProbe,
							Lifecycle: &corev1.Lifecycle{
								PreStop: dmapPrestop,
							},
						},
						{
							Name:            quarkconfig.QuarkContainerPodrick,
							Image:           imageLookup[quarkconfig.QuarkContainerPodrick],
							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    podrickCPUsRequest,
									corev1.ResourceMemory: podrickMemRequest,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    podrickCPUsLimit,
									corev1.ResourceMemory: podrickMemLimit,
								},
							},
							LivenessProbe: &corev1.Probe{
								ProbeHandler: corev1.ProbeHandler{
									Exec: &corev1.ExecAction{
										Command: []string{
											"/bin/sh",
											"-ec",
											"[ $(expr `date +%s` - `stat -c %Y /tmp/liveness`) -le 30 ]",
										},
									},
								},
								InitialDelaySeconds: livenessProbeInitialDelaySeconds,
								PeriodSeconds:       livenessProbePeriodSeconds,
								TimeoutSeconds:      livenessProbeTimeoutSeconds,
							},
						},
						{
							Name:            quarkconfig.QuarkContainerLogging,
							Image:           imageLookup[quarkconfig.QuarkContainerLogging],
							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    loggingCPUsReq,
									corev1.ResourceMemory: loggingMemReq,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    loggingCPUsLimit,
									corev1.ResourceMemory: loggingMemLimit,
								},
							},
						},
						{
							Name:            quarkconfig.QuarkContainerSupportability,
							Image:           imageLookup[quarkconfig.QuarkContainerSupportability],
							ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

							Resources: corev1.ResourceRequirements{
								Requests: corev1.ResourceList{
									corev1.ResourceCPU:    supportabilityCPUsReq,
									corev1.ResourceMemory: supportabilityMemReq,
								},
								Limits: corev1.ResourceList{
									corev1.ResourceCPU:    supportabilityCPUsLimit,
									corev1.ResourceMemory: supportabilityMemLimit,
								},
							},
							Lifecycle: &corev1.Lifecycle{
								PreStop: &corev1.LifecycleHandler{
									Exec: &corev1.ExecAction{
										Command: []string{
											"pkill",
											"-TERM",
											"exporter",
										},
									},
								},
							},
						},
					},
					Tolerations: []corev1.Toleration{
						// Tolerate NoExecute on a not-ready node for five seconds
						// so that the node has some time to remove the NoExecute taint
						// in the case it becomes Ready with the taint still attached.
						// This prevents the TaintManager from deleting the newly scheduled pod
						// on a node it should otherwise not need to be deleted from.
						{
							Key:               "node.kubernetes.io/not-ready",
							Operator:          corev1.TolerationOpExists,
							Effect:            corev1.TaintEffectNoExecute,
							TolerationSeconds: &tolerationSecondsFive,
						},
						// Tolerate NoExecute on an unreachable node for thirty seconds,
						// so that the NHC has some time to delete it. If this fails,
						// let the taint manager at least try.
						{
							Key:               "node.kubernetes.io/unreachable",
							Operator:          corev1.TolerationOpExists,
							Effect:            corev1.TaintEffectNoExecute,
							TolerationSeconds: &tolerationSecondsThirty,
						},
					},
					Volumes: c.makeVolumes(namespace, name, protocol, dataCount, diskPackingEnabled),
				},
			},
			Strategy: appsv1.DeploymentStrategy{
				Type: appsv1.RecreateDeploymentStrategyType,
			},
		},
	}

	spec := &deploy.Spec.Template.Spec
	if protocol == "cifs" || storagePool != nil {
		var createResolvConfCommand, resolvData string
		if storagePool != nil {
			createResolvConfCommand = "touch /tmp/resolv.conf"
		} else {
			resolvData = util.MakeResolvConfString(volume.Spec.SMBConfiguration.DNSIPs)
			createResolvConfCommand = "touch /tmp/resolv.conf && printf '" + resolvData + "' >> /tmp/resolv.conf"
		}
		spec.InitContainers = append(
			spec.InitContainers,
			corev1.Container{
				Name:  quarkconfig.QuarkContainerInit,
				Image: imageLookup[quarkconfig.QuarkContainerInit],
				Command: []string{
					"/bin/sh",
					"-c",
					createResolvConfCommand,
				},
				VolumeMounts: []corev1.VolumeMount{
					{
						Name:      "secd-dns-volume-emptydir",
						MountPath: "/tmp",
					},
				},
			},
		)

		secdCPUsRequest, secdCPUsLimit, err := c.getCPUs(name, specSize, serviceLevel,
			quarkconfig.QuarkContainerSECD)
		if err != nil {
			return nil, fmt.Errorf("failed to get CPUs request: %s and limit: %s for the Quark pod: %v", secdCPUsRequest.String(), secdCPUsLimit.String(), err)
		}
		secdMemRequest, secdMemLimit, err := c.getMem(name, specSize, serviceLevel,
			quarkconfig.QuarkContainerSECD)
		if err != nil {
			return nil, fmt.Errorf("failed to get Memory request: %s and limit: %s for the Quark pod: %v", secdMemRequest.String(), secdMemLimit.String(), err)
		}
		secdContainer := corev1.Container{
			Name:  quarkconfig.QuarkContainerSECD,
			Image: imageLookup[quarkconfig.QuarkContainerSECD],

			ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),

			Resources: corev1.ResourceRequirements{
				Requests: corev1.ResourceList{
					corev1.ResourceCPU:    secdCPUsRequest,
					corev1.ResourceMemory: secdMemRequest,
				},
				Limits: corev1.ResourceList{
					corev1.ResourceCPU:    secdCPUsLimit,
					corev1.ResourceMemory: secdMemLimit,
				},
			},
			LivenessProbe: &corev1.Probe{
				ProbeHandler: corev1.ProbeHandler{
					Exec: &corev1.ExecAction{
						Command: []string{
							"/bin/bash",
							"-ec",
							"/usr/local/bin/secd_health_check.sh",
						},
					},
				},
				InitialDelaySeconds: livenessProbeSecdInitialDelaySeconds,
				PeriodSeconds:       livenessProbePeriodSeconds,
				TimeoutSeconds:      livenessProbeTimeoutSeconds,
			},
			Command: []string{"/bin/bash", "-c"},
			Args:    []string{"PATH=$PATH:/scripts start_secd.sh"},
			Env: []corev1.EnvVar{
				{
					Name:  "QUARK_LOAD_BALANCER_IP",
					Value: exportAddress,
				},
				{
					Name:  "QUARK_DARK",
					Value: c.DarkContent,
				},
				{
					Name:  "QUARK_SECD_ACTIVE_OPEN_PORT_RANGE",
					Value: "61001 61094",
				},
				{
					Name:  "QUARK_SECD_NSUPDATE_ACTIVE_OPEN_PORT_RANGE",
					Value: "61095 61098",
				},
				{
					Name:  "QUARK_SECD_DIG_ACTIVE_OPEN_PORT_RANGE_UDP",
					Value: "61099 61100",
				},
			},
			Ports: []corev1.ContainerPort{
				{
					Name:          "restapi",
					Protocol:      "TCP",
					ContainerPort: 8082,
				},
				{
					Name:          "rpc",
					Protocol:      "TCP",
					ContainerPort: 65519,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP1,
					Protocol:      "TCP",
					ContainerPort: 61001,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP2,
					Protocol:      "TCP",
					ContainerPort: 61002,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP3,
					Protocol:      "TCP",
					ContainerPort: 61003,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCP4,
					Protocol:      "TCP",
					ContainerPort: 61004,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU1,
					Protocol:      "TCP",
					ContainerPort: 61005,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU2,
					Protocol:      "TCP",
					ContainerPort: 61006,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU3,
					Protocol:      "TCP",
					ContainerPort: 61007,
				},
				{
					Name:          quarkconfig.SNATActiveOpenTCPNSU4,
					Protocol:      "TCP",
					ContainerPort: 61008,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP1,
					Protocol:      "UDP",
					ContainerPort: 61001,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP2,
					Protocol:      "UDP",
					ContainerPort: 61002,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP3,
					Protocol:      "UDP",
					ContainerPort: 61003,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDP4,
					Protocol:      "UDP",
					ContainerPort: 61004,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU1,
					Protocol:      "UDP",
					ContainerPort: 61005,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU2,
					Protocol:      "UDP",
					ContainerPort: 61006,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU3,
					Protocol:      "UDP",
					ContainerPort: 61007,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPNSU4,
					Protocol:      "UDP",
					ContainerPort: 61008,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPDIG1,
					Protocol:      "UDP",
					ContainerPort: 61009,
				},
				{
					Name:          quarkconfig.SNATActiveOpenUDPDIG2,
					Protocol:      "UDP",
					ContainerPort: 61010,
				},
			},
			VolumeMounts: []corev1.VolumeMount{
				c.makeMrootVolumeMount(name),
				{
					Name:      "secd-dns-volume-emptydir",
					MountPath: "/etc/resolv.conf",
					SubPath:   "resolv.conf",
				},
			},
		}

		if storagePool == nil {
			secdContainer.Env = append(secdContainer.Env,
				corev1.EnvVar{
					Name:  "QUARK_PROTOCOL",
					Value: protocol,
				},
				corev1.EnvVar{
					Name:  "QUARK_VOLUME_UUID",
					Value: quarkVolumeUUID,
				},
				corev1.EnvVar{
					Name:  "QUARK_VOLUME_NAME",
					Value: quarkVolumeName,
				},
				corev1.EnvVar{
					Name:  "QUARK_VOLUME_PATH",
					Value: volume.Spec.VolumePath,
				},
			)
		} else {
			secdContainer.Env = append(secdContainer.Env,
				corev1.EnvVar{
					Name:  "QUARK_STORAGE_POOL_UUID",
					Value: name,
				},
				corev1.EnvVar{
					Name:  "QUARK_CLUSTER_UUID",
					Value: name,
				},
				corev1.EnvVar{
					Name:  "QUARK_UUID",
					Value: quarkUUID,
				},
				corev1.EnvVar{
					Name:  "QUARK_NODE_UUID",
					Value: quarkUUID,
				},
				corev1.EnvVar{
					Name:  "QUARK_DBLADE_UUID",
					Value: quarkUUID,
				},
				corev1.EnvVar{
					Name:  "QUARK_NBLADE_UUID",
					Value: nbladeUUID,
				},
				corev1.EnvVar{
					Name:  "QUARK_VSERVER_UUID",
					Value: vserverUUID,
				},
			)
		}
		spec.Containers = append(spec.Containers, secdContainer)
	}

	if c.EnableCCPD {
		ccpdContainer := corev1.Container{
			Name:            quarkconfig.QuarkContainerCCPD,
			Image:           imageLookup[quarkconfig.QuarkContainerCCPD],
			ImagePullPolicy: corev1.PullPolicy(c.ImagePullPolicy),
			Command:         []string{"/bin/bash", "-c"},
			Args:            []string{"PATH=$PATH:/scripts start_ccpd.sh"},
			VolumeMounts: []corev1.VolumeMount{
				c.makeMrootVolumeMount(name),
			},
			Env: []corev1.EnvVar{
				{
					Name:  "QUARK_UUID",
					Value: quarkUUID,
				},
				{
					Name:  "QUARK_STORAGE_POOL_UUID",
					Value: storagePool.Name,
				},
				{
					Name:  "QUARK_NODE_UUID",
					Value: quarkUUID,
				},
				{
					Name:  "QUARK_DBLADE_UUID",
					Value: quarkUUID,
				},
				{
					Name:  "QUARK_NBLADE_UUID",
					Value: nbladeUUID,
				},
				{
					Name:  "QUARK_VSERVER_UUID",
					Value: vserverUUID,
				},
				{
					Name:  "QUARK_CLUSTER_UUID",
					Value: storagePool.Name,
				},
				{
					Name: "QUARK_POD_NAMESPACE",
					ValueFrom: &corev1.EnvVarSource{
						FieldRef: &corev1.ObjectFieldSelector{
							FieldPath: "metadata.namespace",
						},
					},
				},
				{
					Name: "QUARK_POD_IP",
					ValueFrom: &corev1.EnvVarSource{
						FieldRef: &corev1.ObjectFieldSelector{
							FieldPath: "status.podIP",
						},
					},
				},
				{
					Name:  "QUARK_SERVICE_MESH_ENABLED",
					Value: strconv.FormatBool(c.EnableServiceMesh),
				},
			},
			// TODO: Define resources
			// TODO: Define liveness probe
		}

		spec.Containers = append(spec.Containers, ccpdContainer)
	}

	if c.EnableMultiAttach {
		spec.Affinity = &corev1.Affinity{
			PodAffinity: &corev1.PodAffinity{
				RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
					{
						LabelSelector: getStorageLabelSelector(volume, storagePool),
						TopologyKey:   "kubernetes.io/hostname",
					},
				},
			},
		}
	} else {
		spec.Affinity = &corev1.Affinity{
			PodAntiAffinity: &corev1.PodAntiAffinity{
				RequiredDuringSchedulingIgnoredDuringExecution: []corev1.PodAffinityTerm{
					{
						LabelSelector: getQuarkLabelSelector(volume, storagePool),
						TopologyKey:   "kubernetes.io/hostname",
					},
				},
			},
		}
	}
	dmapContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerDMAP)
	if dmapContainer == nil {
		return nil, fmt.Errorf("failed to get container spec")
	}
	dmapContainer.Command = []string{"/bin/bash", "-c"}
	dmapContainer.Args = dmapArgs
	dmapContainer.Env = []corev1.EnvVar{
		{
			// Not sure if this is needed, but we use a hardcoded garbage value
			Name:  "SERIAL_NO_GEN_ENV",
			Value: "123456",
		},
		{
			// DMS_CPUS_ENV - This is the number of pthreads used for SK cpus. A minimum of 4 is required. For alpha, this will be 8.
			Name:  "DMS_CPUS_ENV",
			Value: fmt.Sprintf("%d", dmsCPUs),
		},
		{
			Name: "QUARK_CPU_LIMIT",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "limits.cpu",
				},
			},
		},
		{
			Name: "QUARK_MEM_LIMIT",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "limits.memory",
				},
			},
		},
		{
			Name: "QUARK_CPU_REQ",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "requests.cpu",
				},
			},
		},
		{
			Name: "QUARK_MEM_REQ",
			ValueFrom: &corev1.EnvVarSource{
				ResourceFieldRef: &corev1.ResourceFieldSelector{
					Resource: "requests.memory",
				},
			},
		},
		{
			Name: "QUARK_POD_IP",
			ValueFrom: &corev1.EnvVarSource{
				FieldRef: &corev1.ObjectFieldSelector{
					FieldPath: "status.podIP",
				},
			},
		},
		{
			Name:  "QUARK_LOAD_BALANCER_IP",
			Value: exportAddress,
		},
		{
			Name:  "QUARK_NFS_ACTIVE_OPEN_PORT",
			Value: "61000",
		},
		{
			Name:  "QUARK_MAX_PVSIZE",
			Value: strconv.FormatInt(c.MaxPvcSize, 10),
		},
		{
			Name:  "QUARK_NVRAM_ID",
			Value: quarkNvramID,
		},
		{
			Name:  "QUARK_DARK",
			Value: c.DarkContent,
		},
		{
			Name:  "QUARK_DISK_PACKING_ENABLED",
			Value: strconv.FormatBool(diskPackingEnabled),
		},
		{
			Name:  "QUARK_DMAP_NVRAM_SIZE",
			Value: strconv.FormatInt(nvramSize, 10),
		},
		{
			Name:  "QUARK_ETCD_ENDPOINT",
			Value: "https://" + quarkconfig.QuarkEtcdClusterEndpoint,
		},
		{
			Name:  "QUARK_IOPS_PER_TIB",
			Value: fmt.Sprintf("%d", quarkIOPSPerTiB),
		},
		{
			Name:  "QUARK_MIBS_PER_TIB",
			Value: fmt.Sprintf("%d", quarkThroughput),
		},
		{
			// this value is in bytes
			Name:  "QUARK_CAPACITY_LIMIT",
			Value: fmt.Sprintf("%d", quarkCapacityLimit.Value()),
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_MONITOR_FILE",
			Value: quarkconfig.QuarkUpgradeMonitorFile,
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_VARIANT",
			Value: c.QuarkVariant,
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_ETCD_KEY",
			Value: quarkconfig.QuarkUpgradeEtcdKey,
		},
		{
			// DO NOT MODIFY, used for upgrade and bootargs
			Name:  "QUARK_USE_ETCD_BOOTARGS",
			Value: quarkconfig.QuarkUseEtcdBootargs,
		},
		{
			Name:  "QUARK_DISK_AUTO_PROVISION_ENABLED",
			Value: strconv.FormatBool(c.thinProvisioningEnabled),
		},
		{
			Name:  "QUARK_CCPD_ENABLED",
			Value: strconv.FormatBool(c.EnableCCPD),
		},
		{
			Name:  "CLOUD",
			Value: c.Cloud,
		},
	}

	// Env var used for Azure live-resize
	if c.Cloud == quarkconfig.CloudAzure {
		dmapContainer.Env = append(
			dmapContainer.Env,
			corev1.EnvVar{
				Name:  "DISK_CNT_PER_RG",
				Value: quarkconfig.AzureDiskCountPerRG,
			},
		)
	}

	// XXX Why was this using c.UseStoragePool instead of c.EnableMultiVolume
	// Value: strconv.FormatBool(c.EnableMultiVolume),
	// Since UseStoragePool would always be true, just turn it on
	if storagePool != nil {
		dmapContainer.Env = append(
			dmapContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_MULTIVOL_DMAP_ENABLED",
				Value: "true",
			},
			corev1.EnvVar{
				Name:  "QUARK_STORAGE_POOL_UUID",
				Value: storagePool.Name,
			},
			corev1.EnvVar{
				Name:  "QUARK_CLUSTER_UUID",
				Value: storagePool.Name,
			},
			corev1.EnvVar{
				Name:  "QUARK_UUID",
				Value: quarkUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_NODE_UUID",
				Value: quarkUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_DBLADE_UUID",
				Value: quarkUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_NBLADE_UUID",
				Value: nbladeUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_VSERVER_UUID",
				Value: vserverUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_ETCD_KEY_FORMAT_VERSION",
				Value: quarkconfig.QuarkEtcdKeyVersion,
			},
		)
	}

	if storagePool == nil {
		dmapContainer.Env = append(
			dmapContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_NAME",
				Value: quarkVolumeName,
			},
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_PATH",
				Value: volume.Spec.VolumePath,
			},
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_SIZE",
				Value: quarkVolumeSize,
			},
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_UUID",
				Value: quarkVolumeUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_PROTOCOL",
				Value: protocol,
			},
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_UNIX_PERM",
				Value: "0770",
			},
		)
		if volume.Spec.CloudSnapshot != "" {
			dmapContainer.Env = append(
				dmapContainer.Env,
				corev1.EnvVar{
					Name:  "QUARK_VOLUME_TYPE",
					Value: "DP",
				},
				corev1.EnvVar{
					Name:  "QUARK_PHOTON_COPY_CACHE_SIZE",
					Value: fmt.Sprintf("%d", volume.Status.RestoreCacheSize.Value()),
				},
			)
		}
	}

	if protocol != "cifs" || storagePool != nil {
		dmapContainer.Ports = []corev1.ContainerPort{
			{
				Name:          "rpcbind",
				Protocol:      "TCP",
				ContainerPort: 111,
			},
			{
				Name:          "mount",
				Protocol:      "TCP",
				ContainerPort: 635,
			},
			{
				Name:          "nfs",
				Protocol:      "TCP",
				ContainerPort: 2049,
			},
			{
				Name:          "nfsmntd1",
				Protocol:      "TCP",
				ContainerPort: 4045,
			},
			{
				Name:          "nfsmntd2",
				Protocol:      "TCP",
				ContainerPort: 4046,
			},
			{
				Name:          "activeopen",
				Protocol:      "TCP",
				ContainerPort: 61000,
			},
			{
				Name:          "rpcbindu",
				Protocol:      "UDP",
				ContainerPort: 111,
			},
			{
				Name:          "mountu",
				Protocol:      "UDP",
				ContainerPort: 635,
			},
			{
				Name:          "nfsu",
				Protocol:      "UDP",
				ContainerPort: 2049,
			},
			{
				Name:          "nfsmntd1u",
				Protocol:      "UDP",
				ContainerPort: 4045,
			},
			{
				Name:          "nfsmntd2u",
				Protocol:      "UDP",
				ContainerPort: 4046,
			},
			{
				Name:          "rquotau",
				Protocol:      "UDP",
				ContainerPort: 875,
			},
			{
				Name:          "activeopenu",
				Protocol:      "UDP",
				ContainerPort: 61000,
			},
		}
	}
	if protocol == "cifs" || storagePool != nil {
		dmapProtocolPorts = append(dmapContainer.Ports,
			corev1.ContainerPort{
				Name:          "cifs",
				Protocol:      "TCP",
				ContainerPort: 445,
			},
		)
	}

	restapiport := int32(8080)
	dmapContainer.Ports = []corev1.ContainerPort{
		{
			Name:          "restapitest",
			Protocol:      "TCP",
			ContainerPort: restapiport,
		},
	}

	dmapContainer.Ports = append(dmapContainer.Ports, dmapProtocolPorts...)

	dmapContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(name),
	}

	_, err = c.KubeClientset.CoreV1().Secrets(namespace).Get(context.TODO(),
		quarkconfig.RootCertName, metav1.GetOptions{})
	if err == nil {
		dmapRootCertVolume := corev1.Volume{
			Name: getRootCertVolName(namespace, name),
			VolumeSource: corev1.VolumeSource{
				Secret: &corev1.SecretVolumeSource{
					SecretName: quarkconfig.RootCertName,
				},
			},
		}
		spec.Volumes = append(spec.Volumes, dmapRootCertVolume)
		mountPath := "/usr/local/share/root-certs"
		secretVolume := corev1.VolumeMount{
			Name:      getRootCertVolName(namespace, name),
			ReadOnly:  true,
			MountPath: mountPath,
		}
		dmapContainer.VolumeMounts = append(dmapContainer.VolumeMounts, secretVolume)
		dmapContainer.Env = append(dmapContainer.Env, corev1.EnvVar{
			Name:  "C2C_CERT_FILE",
			Value: fmt.Sprintf("%s/%s", mountPath, "roots.pem"),
		})
	} else if errors.IsNotFound(err) {
		// If no secret provided, use non-TLS for cloud snapshots
	} else {
		return nil, fmt.Errorf("error reading %s secret. %s", quarkconfig.RootCertName, err.Error())
	}

	//Mount etcd client secret
	_, err = c.KubeClientset.CoreV1().Secrets(namespace).Get(context.TODO(),
		quarkconfig.EtcdClientSecretName, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s secret. %s", quarkconfig.EtcdClientSecretName, err.Error())
	}
	dmapEtcdClientSecretVolume := corev1.Volume{
		Name: getEtcdClientSecretVolName(name),
		VolumeSource: corev1.VolumeSource{
			Secret: &corev1.SecretVolumeSource{
				SecretName: quarkconfig.EtcdClientSecretName,
			},
		},
	}
	spec.Volumes = append(spec.Volumes, dmapEtcdClientSecretVolume)
	secretVolume := corev1.VolumeMount{
		Name:      getEtcdClientSecretVolName(name),
		ReadOnly:  true,
		MountPath: "/usr/local/share/ca-certificates",
	}
	dmapContainer.VolumeMounts = append(dmapContainer.VolumeMounts, secretVolume)

	podrickContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerPodrick)
	if podrickContainer == nil {
		return nil, fmt.Errorf("failed to get container spec")
	}
	podrickContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(name),
	}
	podrickContainer.Args = []string{
		"--namespace=" + namespace,
		"--cloud=" + c.Cloud,
		"--nodevol=" + c.Nodevol,
		fmt.Sprintf("--debug=%t", c.Debug),
	}

	if storagePool != nil {
		podrickContainer.Args = append(podrickContainer.Args, "--storagePoolUUID="+storagePool.Name)
	}
	podrickContainer.Env = []corev1.EnvVar{
		{
			Name:  "QUARK_DARK",
			Value: c.DarkContent,
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_MONITOR_FILE",
			Value: quarkconfig.QuarkUpgradeMonitorFile,
		},
		{
			Name: "QUARK_POD_NAME",
			ValueFrom: &corev1.EnvVarSource{
				FieldRef: &corev1.ObjectFieldSelector{
					FieldPath: "metadata.name",
				},
			},
		},
		{
			Name:  "QUARK_VARIANT",
			Value: c.QuarkVariant,
		},
		{
			Name: "QUARK_NODE_NAME",
			ValueFrom: &corev1.EnvVarSource{
				FieldRef: &corev1.ObjectFieldSelector{
					FieldPath: "spec.nodeName",
				},
			},
		},
		{
			// DO NOT MODIFY, used for upgrade
			Name:  "QUARK_UPGRADE_ETCD_KEY",
			Value: quarkconfig.QuarkUpgradeEtcdKey,
		},
		{
			Name:  "QUARK_MIN_PVC_SIZE",
			Value: fmt.Sprintf("%d", c.MinPvcSize),
		},
		{
			Name:  "QUARK_IOPS_PER_TB",
			Value: fmt.Sprintf("%v", quarkIOPSPerTiB),
		},
		{
			Name:  "QUARK_MBPS_PER_TB",
			Value: fmt.Sprintf("%v", quarkThroughput),
		},
		{
			Name:  "QUARK_IOPS_LIMIT",
			Value: fmt.Sprintf("%v", quarkIopsLimit),
		},
		{
			Name:  "QUARK_TPUT_LIMIT",
			Value: fmt.Sprintf("%v", quarkTputLimit),
		},
	}
	if storagePool != nil {
		podrickContainer.Env = append(
			podrickContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_MULTI_VOLUME",
				Value: strconv.FormatBool(c.EnableMultiVolume),
			},
			corev1.EnvVar{
				Name:  "QUARK_STORAGE_POOL_ENABLED",
				Value: strconv.FormatBool(true),
			},
		)
	}
	//Add port to Podrick container
	podrickMetricsint, _ := strconv.Atoi(quarkconfig.PodrickMetricsPort)
	podrickContainer.Ports =
		[]corev1.ContainerPort{
			{
				Name:          "quark-metrics",
				ContainerPort: int32(podrickMetricsint),
			},
		}

	loggingContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerLogging)
	if loggingContainer == nil {
		return nil, fmt.Errorf("failed to get container spec")
	}
	loggingContainer.Env = []corev1.EnvVar{
		{
			Name:  "QUARK_OWNER_ID",
			Value: ownerID,
		},
		{
			Name:  "LOGGING_OUTPUT_NAME",
			Value: c.LoggingOutput,
		},
	}

	if storagePool == nil {
		loggingContainer.Env = append(
			loggingContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_UUID",
				Value: quarkVolumeUUID,
			},
		)
	} else {
		loggingContainer.Env = append(
			loggingContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_STORAGE_POOL_UUID",
				Value: storagePool.Name,
			},
		)
	}

	loggingContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(name),
		{
			Name:      getLoggingVolName(namespace, name),
			MountPath: "/fluent-bit/etc",
		},
	}

	// Configure supportability container
	supportabilityContainer := util.GetContainerSpec(spec.Containers, quarkconfig.QuarkContainerSupportability)
	if supportabilityContainer == nil {
		return nil, fmt.Errorf("failed to get supportability container spec")
	}
	supportabilityContainer.Env = []corev1.EnvVar{
		{
			Name:  "QUARK_DARK",
			Value: c.DarkContent,
		},
		{
			Name:  "QUARK_OWNER_ID",
			Value: ownerID,
		},
		{
			Name:  "QUARK_VARIANT",
			Value: c.QuarkVariant,
		},
	}
	if storagePool == nil {
		supportabilityContainer.Env = append(
			supportabilityContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_UUID",
				Value: quarkVolumeUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_VOLUME_NAME",
				Value: quarkVolumeName,
			},
			corev1.EnvVar{
				Name:  "QUARK_PROTOCOL",
				Value: protocol,
			},
		)
	} else {
		supportabilityContainer.Env = append(
			supportabilityContainer.Env,
			corev1.EnvVar{
				Name:  "QUARK_STORAGE_POOL_ENABLED",
				Value: strconv.FormatBool(true),
			},
			corev1.EnvVar{
				Name:  "QUARK_STORAGE_POOL_UUID",
				Value: storagePool.Name,
			},
			corev1.EnvVar{
				Name:  "QUARK_CLUSTER_UUID",
				Value: storagePool.Name,
			},
			corev1.EnvVar{
				Name:  "QUARK_NODE_UUID",
				Value: quarkUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_DBLADE_UUID",
				Value: quarkUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_NBLADE_UUID",
				Value: nbladeUUID,
			},
			corev1.EnvVar{
				Name:  "QUARK_VSERVER_UUID",
				Value: vserverUUID,
			},
		)
	}

	supportabilityContainer.VolumeMounts = []corev1.VolumeMount{
		c.makeMrootVolumeMount(name),
	}

	// Start only exporter if using Nodevol
	if strings.EqualFold(c.Nodevol, "true") {
		supportabilityContainer.Command = []string{"/bin/bash", "-c"}
		supportabilityContainer.Args = []string{"/usr/local/bin/start_container exporter"}
	}

	// Mount quarkconfig.SupportabilitySecretName (supportability-sa-key) secret,
	// Set the service account key as GOOGLE_APPLICATION_CREDENTIALS env var
	_, err = c.KubeClientset.CoreV1().Secrets(namespace).Get(context.TODO(),
		quarkconfig.SupportabilitySecretName, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s secret. %s", quarkconfig.SupportabilitySecretName, err.Error())
	}

	supportabilitySecretVolume := corev1.Volume{
		Name: "supportability-secret-vol",
		VolumeSource: corev1.VolumeSource{
			Secret: &corev1.SecretVolumeSource{
				SecretName: quarkconfig.SupportabilitySecretName,
			},
		},
	}
	spec.Volumes = append(spec.Volumes, supportabilitySecretVolume)

	mountPath := "/var/secrets/supportability"
	secretVolumeMount := corev1.VolumeMount{
		Name:      "supportability-secret-vol",
		ReadOnly:  true,
		MountPath: mountPath,
	}
	supportabilityContainer.VolumeMounts = append(supportabilityContainer.VolumeMounts, secretVolumeMount)
	supportabilityContainer.Env = append(supportabilityContainer.Env, corev1.EnvVar{
		Name:  "GOOGLE_APPLICATION_CREDENTIALS",
		Value: fmt.Sprintf("%s/%s", mountPath, "supportability-sa.json"),
	})

	// Set env variables from Config.SupportabilityAssociationConfig(supportability-config) configmap
	_, err = c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(context.TODO(),
		quarkconfig.SupportabilityAssociationConfig, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s configmap. %s", quarkconfig.SupportabilityAssociationConfig, err.Error())
	}
	supportabilityContainer.EnvFrom = []corev1.EnvFromSource{
		{
			ConfigMapRef: &corev1.ConfigMapEnvSource{
				LocalObjectReference: corev1.LocalObjectReference{Name: quarkconfig.SupportabilityAssociationConfig},
			},
		},
	}
	// Logging container will have same association Env variables as supportability container
	loggingContainer.EnvFrom = []corev1.EnvFromSource{
		{
			ConfigMapRef: &corev1.ConfigMapEnvSource{
				LocalObjectReference: corev1.LocalObjectReference{Name: quarkconfig.SupportabilityAssociationConfig},
			},
		},
	}

	// Mount Config.QuarkExporterConfig(QuarkExporterConfig) config map as volume if present
	_, err = c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(context.TODO(),
		quarkconfig.QuarkExporterConfig, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s configmap. %s", quarkconfig.QuarkExporterConfig, err.Error())
	}
	exporterVolume := corev1.Volume{
		Name: getExporterVolName(namespace, name),
		VolumeSource: corev1.VolumeSource{
			ConfigMap: &corev1.ConfigMapVolumeSource{
				LocalObjectReference: corev1.LocalObjectReference{
					Name: quarkconfig.QuarkExporterConfig,
				},
			},
		},
	}
	spec.Volumes = append(spec.Volumes, exporterVolume)
	ExporterVolumeMount := corev1.VolumeMount{
		Name:      getExporterVolName(namespace, name),
		MountPath: "/opt/exporter_config",
	}
	supportabilityContainer.VolumeMounts = append(supportabilityContainer.VolumeMounts, ExporterVolumeMount)

	var quarkVolume corev1.Volume
	// Incase of nodevol config, then quarkvol is sub directory inside the nodevol directory on the host node.
	// Here attaching the nodevol root directory (quarkconfig.NodevolHostpath) itself to the quark pod using
	// "HostPathDirectory" option so that it will make sure the pod will not start until the nodevol path is ready.
	// The quarkvol pod containers will access those directories using the Subpath mechanism
	if strings.EqualFold(c.Nodevol, "true") {
		source := corev1.HostPathDirectory
		quarkVolume = corev1.Volume{
			Name: "quarkvol",
			VolumeSource: corev1.VolumeSource{
				HostPath: &corev1.HostPathVolumeSource{
					Type: &source,
					Path: quarkconfig.NodevolHostpath,
				},
			},
		}
	} else {
		quarkVolume = corev1.Volume{
			Name: "quarkvol",
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: GetQuarkVolPvcName(namespace, name),
				},
			},
		}
	}
	spec.Volumes = append(spec.Volumes, quarkVolume)

	// Check if fluentbit-sidecar-config is present
	_, err = c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(context.TODO(),
		quarkconfig.FluentbitSidecarConfig, metav1.GetOptions{})
	if err != nil {
		return nil, fmt.Errorf("error reading %s configmap. %s", quarkconfig.FluentbitSidecarConfig, err.Error())
	}
	// Create volume for logging container from ConfigMap
	loggingVolume := corev1.Volume{
		Name: getLoggingVolName(namespace, name),
		VolumeSource: corev1.VolumeSource{
			ConfigMap: &corev1.ConfigMapVolumeSource{
				LocalObjectReference: corev1.LocalObjectReference{
					Name: quarkconfig.FluentbitSidecarConfig,
				},
			},
		},
	}
	spec.Volumes = append(spec.Volumes, loggingVolume)

	// Cloud-specific customizations
	switch c.Cloud {
	case quarkconfig.CloudCTL:
		// For CTL deployments, add the label and toleration to the Quark pod.
		spec.Tolerations = []corev1.Toleration{
			{quarkconfig.NodeLabel, corev1.TolerationOpEqual, "true", corev1.TaintEffectNoSchedule, nil},
		}
		spec.NodeSelector = map[string]string{quarkconfig.NodeLabel: "true"}
	}

	deploy.Name = GetQuarkDeployName(namespace, name)
	deploy.Labels = labels
	return deploy, nil
}

func (c *Controller) makeInitialQuarkDeployment(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	quarkVolumeUUID string,
	ownerID string,
	vserverUUID string,
	nbladeUUID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) (*appsv1.Deployment, error) {
	return c.makeQuarkDeployment(volume, storagePool, serviceLevel, dataCount, quarkUUID,
		quarkNvramID, quarkVolumeUUID, ownerID, vserverUUID, nbladeUUID, exportAddress,
		quarkconfig.NegativePriorityClass, nodeSelector, false, diskPackingEnabled)
}

func (c *Controller) makeRevisedQuarkDeployment(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	dataCount int,
	quarkUUID string,
	quarkNvramID string,
	quarkVolumeUUID string,
	ownerID string,
	vserverUUID string,
	nbladeUUID string,
	exportAddress string,
	nodeSelector map[string]string,
	diskPackingEnabled bool,
) (*appsv1.Deployment, error) {
	if _, ok := nodeSelector[quarkconfig.ZoneTopologyKey]; !ok && c.regionalMeta != nil {
		// We're upgrading a volume to a storage pool; need to make sure that the
		// volume remains in place. The volume should have come online at least once,
		// so zone should not be empty.
		if storagePool != nil {
			nodeSelector[quarkconfig.ZoneTopologyKey] = storagePool.Status.Zone
		} else {
			nodeSelector[quarkconfig.ZoneTopologyKey] = volume.Status.Zone
		}
	}
	return c.makeQuarkDeployment(volume, storagePool, serviceLevel, dataCount, quarkUUID,
		quarkNvramID, quarkVolumeUUID, ownerID, vserverUUID, nbladeUUID, exportAddress,
		"", nodeSelector, true, diskPackingEnabled)
}

func (c *Controller) makeADClientTCPLoadBalancerService(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	ip string,
) *corev1.Service {

	// If protcol is either all or tcp , we add the tcp ports

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(nil, storagePool),
		Spec: corev1.ServiceSpec{
			Type:                     corev1.ServiceTypeLoadBalancer,
			Ports:                    getGlobalAccessADDNSTCPPorts(),
			LoadBalancerIP:           ip,
			ExternalTrafficPolicy:    "Local",
			Selector:                 getSelectorNetappObj(nil, storagePool),
			PublishNotReadyAddresses: true,
		},
	}
	service.Name = getADDNSTCPLoadBalancerServiceName(storagePool.Namespace,
		storagePool.Name)
	service.Labels = c.getLabelForNetappObj(nil, storagePool)
	c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet,
		storagePool.Spec.GlobalAccess)
	return service
}

func getMainServiceTCPSNATAnnotations(
	protocol string, handleCIFS bool,
) string {
	annotations := []string{}
	if (protocol == "cifs" || protocol == "all") && handleCIFS {
		annotations = append(annotations, fmt.Sprintf("%s:%s",
			quarkconfig.SNATProtocolTCP,
			quarkconfig.SNATActiveOpenCIFSTCPPortRange))
	}
	if protocol != "cifs" {
		annotations = append(annotations, fmt.Sprintf("%s:%s",
			quarkconfig.SNATProtocolTCP,
			quarkconfig.SNATActiveOpenNFSTCPPortRange))
	}
	return strings.Join(annotations, ",")
}

// Generate a TCP load balancer service definition
func (c *Controller) makeTCPLoadBalancerService(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, ip string, internal bool, protocol string) *corev1.Service {
	var ports []corev1.ServicePort
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)

	if protocol == "cifs" || protocol == "all" {
		if internal {
			ports = getCifsTcpServicePorts()
		} else {
			ports = []corev1.ServicePort{
				{
					Name:     "cifs",
					Protocol: "TCP",
					Port:     445,
				},
			}
		}
	}

	// If protcol is either all or tcp , we add the tcp ports
	if protocol != "cifs" {
		ports = append(ports, getNfsTcpServicePorts()...)
	}

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: corev1.ServiceSpec{
			Type:                     corev1.ServiceTypeLoadBalancer,
			Ports:                    ports,
			LoadBalancerIP:           ip,
			ExternalTrafficPolicy:    "Local",
			Selector:                 getSelectorNetappObj(volume, storagePool),
			PublishNotReadyAddresses: true,
		},
	}
	service.Name = GetTCPLoadBalancerServiceName(namespace, name)
	service.Labels = c.getLabelForNetappObj(volume, storagePool)
	if internal {
		if snatAnnotations := getMainServiceTCPSNATAnnotations(protocol,
			(storagePool == nil || !storagePool.Spec.GlobalAccess),
		); snatAnnotations != "" {
			service.Annotations = map[string]string{
				quarkconfig.SNATPortRangeAnnotation: snatAnnotations,
			}
		}
		c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet,
			c.IlbGlobalAccess)
	}
	return service
}

// Generate a UDP load balancer service definition
func (c *Controller) makeADClientUDPLoadBalancerService(
	storagePool *quarkv1alpha1.NetAppStoragePool,
	ip string,
) *corev1.Service {

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(nil, storagePool),
		Spec: corev1.ServiceSpec{
			Type:                     corev1.ServiceTypeLoadBalancer,
			Ports:                    getCifsUDPServicePorts(),
			LoadBalancerIP:           ip,
			ExternalTrafficPolicy:    "Local",
			Selector:                 getSelectorNetappObj(nil, storagePool),
			PublishNotReadyAddresses: true,
		},
	}
	service.Name = getADDNSUDPLoadBalancerServiceName(storagePool.Namespace,
		storagePool.Name)
	service.Labels = c.getLabelForNetappObj(nil, storagePool)
	c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet,
		storagePool.Spec.GlobalAccess)
	return service
}

func getMainServiceUDPSNATAnnotations(
	protocol string, handleCIFS bool,
) string {
	annotations := []string{}
	if (protocol == "cifs" || protocol == "all") && handleCIFS {
		annotations = append(annotations, fmt.Sprintf("%s:%s",
			quarkconfig.SNATProtocolUDP,
			quarkconfig.SNATActiveOpenCIFSUDPPortRange))
	}
	if protocol != "cifs" {
		annotations = append(annotations, fmt.Sprintf("%s:%s",
			quarkconfig.SNATProtocolUDP,
			quarkconfig.SNATActiveOpenNFSUDPPortRange))
	}
	return strings.Join(annotations, ",")
}

// Generate a UDP load balancer service definition
func (c *Controller) makeUDPLoadBalancerService(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, ip string, internal bool, protocol string) *corev1.Service {
	var ports []corev1.ServicePort
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)

	if (protocol == "cifs" || protocol == "all") &&
		(storagePool == nil || !storagePool.Spec.GlobalAccess) {
		if internal {
			ports = getCifsUDPServicePorts()
		}
	}

	if protocol != "cifs" {
		ports = append(ports, getNfsUDPServicePorts()...)
	}
	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: corev1.ServiceSpec{
			Type:                     corev1.ServiceTypeLoadBalancer,
			Ports:                    ports,
			LoadBalancerIP:           ip,
			ExternalTrafficPolicy:    "Local",
			Selector:                 getSelectorNetappObj(volume, storagePool),
			PublishNotReadyAddresses: true,
		},
	}
	service.Name = GetUDPLoadBalancerServiceName(namespace, name)
	service.Labels = c.getLabelForNetappObj(volume, storagePool)
	if internal {
		if snatAnnotations := getMainServiceUDPSNATAnnotations(protocol,
			(storagePool == nil || !storagePool.Spec.GlobalAccess),
		); snatAnnotations != "" {
			service.Annotations = map[string]string{
				quarkconfig.SNATPortRangeAnnotation: snatAnnotations,
			}
		}
		c.cloudClient.AddCloudAnnotations(service, c.IlbSubnet,
			c.IlbGlobalAccess)
	}
	return service
}

// Generate a service definition
// TODO: Do we need to set externalTrafficPolicy to Local?? for non-GKE service
func (c *Controller) makeService(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, protocol string) *corev1.Service {
	var ports []corev1.ServicePort
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	if protocol == "cifs" || protocol == "all" {
		ports = append(ports, getCifsTcpServicePorts()...)
		ports = append(ports, getCifsUDPServicePorts()...)
	}
	if protocol != "cifs" {
		ports = append(ports, getNfsTcpServicePorts()...)
		ports = append(ports, getNfsUDPServicePorts()...)
	}

	service := &corev1.Service{
		ObjectMeta: MakeObjectMeta(volume, storagePool),
		Spec: corev1.ServiceSpec{
			Ports:                    ports,
			Selector:                 getSelectorNetappObj(volume, storagePool),
			PublishNotReadyAddresses: true,
		},
	}
	service.Name = getServiceName(namespace, name)
	service.Labels = c.getLabelForNetappObj(volume, storagePool)

	// go ahead and add SNAT annotations to add iptable rules
	// specific to the TCP/61000 & UDP/61000
	var annotations []string
	if protocol == "cifs" || protocol == "all" {
		annotations = append(annotations, "tcp:"+quarkconfig.SNATActiveOpenCIFSTCPPortRange)
		annotations = append(annotations, "udp:"+quarkconfig.SNATActiveOpenCIFSUDPPortRange)
	}
	if protocol != "cifs" {
		annotations = append(annotations, "tcp:"+quarkconfig.SNATActiveOpenNFSTCPPortRange)
		annotations = append(annotations, "udp:"+quarkconfig.SNATActiveOpenNFSUDPPortRange)
	}
	if len(annotations) > 0 {
		service.Annotations = map[string]string{
			quarkconfig.SNATPortRangeAnnotation: strings.Join(annotations, ","),
		}
	}
	return service
}

func getNfsUDPServicePorts() []corev1.ServicePort {
	// TODO: The 61* ports can be reduced, once we get rid of legacy volumes
	return []corev1.ServicePort{
		{
			Name:     "rpcbindu",
			Protocol: "UDP",
			Port:     111,
		},
		{
			Name:     "nfsmntd2u",
			Protocol: "UDP",
			Port:     4046,
		},
		{
			Name:     "activeopenu",
			Protocol: "UDP",
			Port:     61000,
		},
	}
}

func getNfsTcpServicePorts() []corev1.ServicePort {
	// TODO: The 61* ports can be reduced, once we get rid of legacy volumes
	return []corev1.ServicePort{
		{
			Name:     "rpcbind",
			Protocol: "TCP",
			Port:     111,
		},
		{
			Name:     "mount",
			Protocol: "TCP",
			Port:     635,
		},
		{
			Name:     "nfs",
			Protocol: "TCP",
			Port:     2049,
		},
		{
			Name:     "nfsmntd1",
			Protocol: "TCP",
			Port:     4045,
		},
		{
			Name:     "nfsmntd2",
			Protocol: "TCP",
			Port:     4046,
		},
		{
			Name:     "activeopen",
			Protocol: "TCP",
			Port:     61000,
		},
	}
}

func getADDNSTCPPorts() []corev1.ServicePort {
	return []corev1.ServicePort{

		{
			Name:     quarkconfig.SNATActiveOpenTCP1,
			Protocol: "TCP",
			Port:     61001,
		},
		{
			Name:     quarkconfig.SNATActiveOpenTCP2,
			Protocol: "TCP",
			Port:     61002,
		},
		{
			Name:     quarkconfig.SNATActiveOpenTCP3,
			Protocol: "TCP",
			Port:     61003,
		},
		{
			Name:     quarkconfig.SNATActiveOpenTCP4,
			Protocol: "TCP",
			Port:     61004,
		},
		{
			Name:     quarkconfig.SNATActiveOpenTCPNSU1,
			Protocol: "TCP",
			Port:     61005,
		},
	}
}

func getGlobalAccessADDNSTCPPorts() []corev1.ServicePort {
	return append(getADDNSTCPPorts(), corev1.ServicePort{
		// This is needed to trigger the all ports rule
		Name:     quarkconfig.SNATActiveOpenTCPNSU2,
		Protocol: "TCP",
		Port:     61006,
	},
	)
}

func getCifsTcpServicePorts() []corev1.ServicePort {
	// TODO: The 61* ports can be reduced, once we get rid of legacy volumes
	return append(
		[]corev1.ServicePort{
			{
				Name:     "cifs",
				Protocol: "TCP",
				Port:     445,
			},
		},
		getADDNSTCPPorts()...,
	)
}

func getCifsUDPServicePorts() []corev1.ServicePort {
	// TODO: The 61* ports can be reduced, once we get rid of legacy volumes

	return []corev1.ServicePort{

		{
			Name:     quarkconfig.SNATActiveOpenUDP1,
			Protocol: "UDP",
			Port:     61001,
		},
		{
			Name:     quarkconfig.SNATActiveOpenUDP2,
			Protocol: "UDP",
			Port:     61002,
		},
		{
			Name:     quarkconfig.SNATActiveOpenUDP3,
			Protocol: "UDP",
			Port:     61003,
		},
		{
			Name:     quarkconfig.SNATActiveOpenUDP4,
			Protocol: "UDP",
			Port:     61004,
		},
		{
			Name:     quarkconfig.SNATActiveOpenUDPNSU1,
			Protocol: "UDP",
			Port:     61005,
		},
		{
			Name:     quarkconfig.SNATActiveOpenUDPNSU2,
			Protocol: "UDP",
			Port:     61006,
		},
	}

}

// Generate multi protocol service definition
func (c *Controller) makeMultiProtocolILB(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, ip string, protocol string) *corev1.Service {
	var ports []corev1.ServicePort
	svc := c.makeService(volume, storagePool, protocol)

	// Make the svc ILB type
	svc.Spec.Type = corev1.ServiceTypeLoadBalancer
	svc.Spec.LoadBalancerIP = ip
	svc.Spec.ExternalTrafficPolicy = "Local"

	// Overwriting ports to the correct ports that are needed for azure
	if protocol == "cifs" || protocol == "all" {
		ports = []corev1.ServicePort{
			{
				Name:     "cifs",
				Protocol: "TCP",
				Port:     445,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP1,
				Protocol: "TCP",
				Port:     61001,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP2,
				Protocol: "TCP",
				Port:     61002,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP3,
				Protocol: "TCP",
				Port:     61003,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCP4,
				Protocol: "TCP",
				Port:     61004,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU1,
				Protocol: "TCP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU2,
				Protocol: "TCP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU3,
				Protocol: "TCP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenTCPNSU4,
				Protocol: "TCP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP1,
				Protocol: "UDP",
				Port:     61001,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP2,
				Protocol: "UDP",
				Port:     61002,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP3,
				Protocol: "UDP",
				Port:     61003,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDP4,
				Protocol: "UDP",
				Port:     61004,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU1,
				Protocol: "UDP",
				Port:     61005,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU2,
				Protocol: "UDP",
				Port:     61006,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU3,
				Protocol: "UDP",
				Port:     61007,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPNSU4,
				Protocol: "UDP",
				Port:     61008,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG1,
				Protocol: "UDP",
				Port:     61009,
			},
			{
				Name:     quarkconfig.SNATActiveOpenUDPDIG2,
				Protocol: "UDP",
				Port:     61010,
			},
		}
	}
	if protocol != "cifs" {
		ports = append(ports, []corev1.ServicePort{
			{
				Name:     "rpcbind",
				Protocol: "TCP",
				Port:     111,
			},
			{
				Name:     "mount",
				Protocol: "TCP",
				Port:     635,
			},
			{
				Name:     "nfs",
				Protocol: "TCP",
				Port:     2049,
			},
			{
				Name:     "nfsmntd1",
				Protocol: "TCP",
				Port:     4045,
			},
			{
				Name:     "activeopen",
				Protocol: "TCP",
				Port:     61000,
			},
			{
				Name:     "rpcbindu",
				Protocol: "UDP",
				Port:     111,
			},
			{
				Name:     "activeopenu",
				Protocol: "UDP",
				Port:     61000,
			},
			{
				Name:     "nfsmntd2",
				Protocol: "TCP",
				Port:     4046,
			},
			{
				Name:     "nfsmntd2u",
				Protocol: "UDP",
				Port:     4046,
			},
		}...)
	}
	svc.Spec.Ports = ports

	c.cloudClient.AddCloudAnnotations(svc, c.IlbSubnet,
		c.IlbGlobalAccess ||
			(storagePool != nil && storagePool.Spec.GlobalAccess))
	return svc
}

// Generate standard metadata
func MakeObjectMeta(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) metav1.ObjectMeta {
	if storagePool != nil {
		return metav1.ObjectMeta{
			Namespace: storagePool.Namespace,
			OwnerReferences: []metav1.OwnerReference{
				*metav1.NewControllerRef(storagePool, schema.GroupVersionKind{
					Group:   quarkv1alpha1.SchemeGroupVersion.Group,
					Version: quarkv1alpha1.SchemeGroupVersion.Version,
					Kind:    "NetAppStoragePool",
				}),
			},
		}
	}
	return metav1.ObjectMeta{
		Namespace: volume.Namespace,
		OwnerReferences: []metav1.OwnerReference{
			*metav1.NewControllerRef(volume, schema.GroupVersionKind{
				Group:   quarkv1alpha1.SchemeGroupVersion.Group,
				Version: quarkv1alpha1.SchemeGroupVersion.Version,
				Kind:    "NetAppVolume",
			}),
		},
	}
}

// Generate the list of block devices to put in the container spec template
func (c *Controller) makeVolumeDevices(dataCount int, diskPackingEnabled bool) []corev1.VolumeDevice {
	var (
		diskCount int
		indexInc  int
	)
	if diskPackingEnabled {
		// No separate nvram disk
		diskCount = dataCount
		indexInc = 0
	} else {
		// Need a separate nvram disk
		diskCount = dataCount + 1
		indexInc = 1
	}
	devices := make([]corev1.VolumeDevice, diskCount)
	if !diskPackingEnabled {
		devices[0] = corev1.VolumeDevice{
			Name:       "nvram",
			DevicePath: "/dev/dms/nvram",
		}
	}
	for i := 0; i < dataCount; i++ {
		devices[i+indexInc] = corev1.VolumeDevice{
			Name:       fmt.Sprintf("data%d", i+1),
			DevicePath: fmt.Sprintf("/dev/dms/data%d", i+1),
		}
	}
	return devices
}

// Make a new data VolumeDevice for new data volume
func makeDataVolumeDevice(deviceCount int) corev1.VolumeDevice {
	device := corev1.VolumeDevice{
		Name:       fmt.Sprintf("data%d", deviceCount),
		DevicePath: fmt.Sprintf("/dev/dms/data%d", deviceCount),
	}
	return device
}

// Generate the list of volumes to put in the pod spec template
func (c *Controller) makeVolumes(namespace string, name string, protocol string, dataCount int, diskPackingEnabled bool) []corev1.Volume {
	var (
		diskCount int
		indexInc  int
	)
	if diskPackingEnabled {
		// No separate nvram disk
		diskCount = dataCount
		indexInc = 0
	} else {
		// Need a separate nvram disk
		diskCount = dataCount + 1
		indexInc = 1
	}
	volumes := make([]corev1.Volume, diskCount)
	if !diskPackingEnabled {
		volumes[0] = corev1.Volume{
			Name: "nvram",
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: getNvramPvcName(namespace, name),
				},
			},
		}
	}
	for i := 0; i < dataCount; i++ {
		volumes[i+indexInc] = corev1.Volume{
			Name: fmt.Sprintf("data%d", i+1),
			VolumeSource: corev1.VolumeSource{
				PersistentVolumeClaim: &corev1.PersistentVolumeClaimVolumeSource{
					ClaimName: getDataPvcName(namespace, name, i),
				},
			},
		}
	}
	// Used to be if protocol == "cifs" || c.UseStoragePool {
	dnsEmptyDirVolume := corev1.Volume{
		Name: "secd-dns-volume-emptydir",
		VolumeSource: corev1.VolumeSource{
			EmptyDir: &corev1.EmptyDirVolumeSource{},
		},
	}
	volumes = append(volumes, dnsEmptyDirVolume)
	return volumes
}

// Check if resize is required based on the disklayout
func isResizeRequired(diskLayout map[string]float64, dataPVCs []*corev1.PersistentVolumeClaim) bool {
	for _, pvc := range dataPVCs {
		requestedsize := pvc.Spec.Resources.Requests[corev1.ResourceStorage]
		if requestedsize.Value() != int64(diskLayout[pvc.Name]) {
			return true
		}
	}
	return false
}

// Function to calculate number of non-data volumes in the quark deployment
func countNonDataVolumes(volumes []corev1.Volume) int {
	count := 0
	for _, volume := range volumes {
		if strings.HasPrefix(volume.Name, "data") {
			continue
		} else {
			count++
		}
	}
	return count
}

// Create a selector that finds data PVCs for a netappObj
func (c *Controller) getDataPvcSelector(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) labels.Selector {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.TypeLabel] = "data"
	if storagePool != nil {
		selectorSet[quarkconfig.StoragePoolLabel] = storagePool.Name
	} else {
		selectorSet[quarkconfig.VolumeLabel] = volume.Name
	}
	return labels.SelectorFromSet(selectorSet)
}

// Create a selector that finds Quark pods for a volume
func GetQuarkPodSelector(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) labels.Selector {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.AppLabel] = "quark"
	if storagePool == nil {
		selectorSet[quarkconfig.VolumeLabel] = volume.Name
	} else {
		selectorSet[quarkconfig.StoragePoolLabel] = storagePool.Name
	}
	return labels.SelectorFromSet(selectorSet)
}

// Generate the name of the QuarkVol PVC for a volume
func GetQuarkVolPvcName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-quarkvol", namespace, name)
}

// Generate the name of the NVRAM PVC for a volume
func getNvramPvcName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-nvram", namespace, name)
}

// Generate the name of the one of the data PVCs for a volume
func getDataPvcName(namespace string, name string, n int) string {
	return fmt.Sprintf("%s-%s-data%d", namespace, name, n+1)
}

// Generate the name for logging container volume
func getLoggingVolName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-logging", namespace, name)
}

// Low-level function to generate a Quark deployment name.
// This is here so Podrick can construct it from its input parameters
// (namespace + volumeUUID) without having to take it as an explicit input
// or to look it up. NHC also uses this to quickly find the deployment name
// for a given pod.
func GetQuarkDeployNameForComponents(namespace, volumeUUID string) string {
	return fmt.Sprintf("%s-%s", namespace, volumeUUID)
}

// Generate the name of the Quark deployment for a volume
func GetQuarkDeployName(namespace string, name string) string {
	return GetQuarkDeployNameForComponents(namespace, name)
}

func GetDmapLivenessProbe(variant string) *corev1.Probe {
	initialDelaySec := livenessProbeInitialDelaySeconds

	// For internal deployments, set initialDelay
	// of liveness probe to 120sec
	if variant == quarkconfig.InternalVariant {
		initialDelaySec = int32(120)
	}

	return &corev1.Probe{
		ProbeHandler: corev1.ProbeHandler{
			Exec: &corev1.ExecAction{
				Command: []string{
					"/bin/bash",
					"-ec",
					"/usr/local/bin/dmap_liveness_probe.sh",
					"-t",
					strconv.FormatInt(int64(livenessProbeTimeoutSeconds), 10),
				},
			},
		},
		InitialDelaySeconds: initialDelaySec,
		PeriodSeconds:       livenessProbePeriodSeconds,
		TimeoutSeconds:      livenessProbeTimeoutSeconds,
	}
}

func GetDmapPrestop() *corev1.LifecycleHandler {
	return &corev1.LifecycleHandler{
		Exec: &corev1.ExecAction{
			Command: []string{
				"/bin/bash",
				"-ec",
				"/usr/local/bin/pre_stop.sh",
			},
		},
	}
}

// TODO: Delete this code after we have upgraded all services to ILB subsetting services
func (c *Controller) recreateServices(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) error {
	namespace := volume.Namespace
	// Set Deletion propagation to foreground so that it waits for the object to be deleted
	deletePolicy := metav1.DeletePropagationForeground
	deleteOption := metav1.DeleteOptions{
		PropagationPolicy: &deletePolicy,
	}
	name := volume.Name
	serviceList := []string{fmt.Sprintf("%s-%s-udp", namespace, name), fmt.Sprintf("%s-%s-udp-aux", namespace, name),
		fmt.Sprintf("%s-%s-tcp-aux", namespace, name), fmt.Sprintf("%s-%s-tcp", namespace, name)}
	// Dmap is halted we can proceed with service recreation
	// Get services belonging to the volume
	// Delete UDP services first and then TCP , while recreating create TCP services first
	// Check for Deletion timestamp on the service , if Present skip deleting
	for _, svcName := range serviceList {
		svc, err := c.servicesLister.Services(namespace).Get(svcName)
		if err != nil && !errors.IsNotFound(err) {
			return fmt.Errorf("unable to get service %s: %v", svcName, err)
		}
		if svc != nil && svc.DeletionTimestamp == nil && !errors.IsNotFound(err) {
			err := c.KubeClientset.CoreV1().Services(namespace).Delete(context.TODO(),
				svcName, deleteOption)
			if err != nil && !errors.IsNotFound(err) {
				return fmt.Errorf("unable to Delete service %s: %v", svcName, err)
			}
			log.Debugf("Deleted service %s", svcName)
		}
	}

	// Call sync service creation
	_, _, _, err := c.syncExportAddress(volume, storagePool)
	return err
}

func GetDmapArgs() []string {
	return []string{"PATH=$PATH:/scripts start_dms_main.py"}
}

// Generate the name of the storage deployment for a volume
func getStorageDeployName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-storage", namespace, name)
}

// Generate the name of the TCP load balancer service for a volume
func getADDNSTCPLoadBalancerServiceName(namespace string, name string) string {
	if len(namespace) > 11 {
		namespace = namespace[0:11]
	}
	return fmt.Sprintf("%s-%s-sp-tcp-ad", namespace, name)
}

// Generate the name of the TCP load balancer service for a volume
func GetTCPLoadBalancerServiceName(namespace string, name string) string {
	if len(namespace) > 11 {
		namespace = namespace[0:11]
	}
	return fmt.Sprintf("%s-%s-sp-tcp", namespace, name)
}

// Generate the name of the UDP load balancer service for a volume
func getADDNSUDPLoadBalancerServiceName(namespace string, name string) string {
	if len(namespace) > 11 {
		namespace = namespace[0:11]
	}
	return fmt.Sprintf("%s-%s-sp-udp-ad", namespace, name)
}

// Generate the name of the UDP load balancer service for a volume
func GetUDPLoadBalancerServiceName(namespace string, name string) string {
	if len(namespace) > 11 {
		namespace = namespace[0:11]
	}
	return fmt.Sprintf("%s-%s-sp-udp", namespace, name)
}

// Generate the name for logging container volume
func getExporterVolName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-exporter", namespace, name)
}

func getAddrName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-addr", namespace, name)
}

func getADClientAddrName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-ad-client", namespace, name)
}

func getServiceName(namespace string, name string) string {
	if len(namespace) > 10 {
		namespace = namespace[0:10]
	}
	return fmt.Sprintf("%s-%s-svc", namespace, name)
}

func getRootCertVolName(namespace string, name string) string {
	return fmt.Sprintf("%s-%s-cert-vol", namespace, name)
}

func getEtcdClientSecretVolName(name string) string {
	return fmt.Sprintf("etcd-client-%s-cert-vol", name)
}

func HasFinalizer(finalizers []string, finalizerStr string) bool {
	for _, finalizer := range finalizers {
		if finalizer == finalizerStr {
			return true
		}
	}
	return false
}

func getServiceLevelBucket(
	size int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (*quarkv1alpha1.QuarkPodBucket, error) {
	for _, quarkPodBucket := range serviceLevel.Spec.QuarkPodBuckets {
		minVolSize := quarkPodBucket.MinVolumeSize.Value()
		maxVolSize := quarkPodBucket.MaxVolumeSize.Value()
		if size > minVolSize && size <= maxVolSize {
			return &quarkPodBucket, nil
		}
	}
	return nil, fmt.Errorf("could not determine service level bucket")
}

// Logic to extract node selector based on service-level/volume-size requested.
func getServiceLevelNodeSelector(
	size int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (map[string]string, error) {
	quarkPodBucket, err := getServiceLevelBucket(size, serviceLevel)
	if err != nil {
		return nil, err
	}
	nodeSelector := util.NodeSelectorStringToMap(quarkPodBucket.NodeSelector)
	log.WithFields(log.Fields{
		"NodeSelector": nodeSelector,
	}).Debug("Determined service level node selector.")
	return nodeSelector, nil
}

// Logic to extract number of CPUs required based on service-level/volume-size requested.
func (c *Controller) getCPUs(
	netappObjName string,
	size int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	containername string,
) (resource.Quantity, resource.Quantity, error) {
	request := resource.Quantity{}
	limit := resource.Quantity{}
	quarkPodBucket, _ := getServiceLevelBucket(size, serviceLevel)
	if quarkPodBucket != nil {
		for _, containerResource := range quarkPodBucket.ContainerResources {
			if containerResource.ContainerName == containername {
				request = containerResource.CPUsRequest
				limit = containerResource.CPUsLimit
				log.WithFields(log.Fields{
					"netappObj":                   netappObjName,
					"serviceLevel":                serviceLevel.Name,
					"bucket":                      quarkPodBucket.Name,
					containername + "CPUsRequest": request,
					containername + "CPUsLimit":   limit,
				}).Debug("Determined " + containername + " Container CPUs request and limit.")
				return request, limit, nil
			}
		}
	}
	return request, limit, fmt.Errorf("could not determine CPUs")
}

// Logic to extract memory required based on service-level/volume-size requested.
func (c *Controller) getMem(
	netappObjName string,
	size int64,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
	containername string,
) (resource.Quantity, resource.Quantity, error) {
	request := resource.Quantity{}
	limit := resource.Quantity{}
	quarkPodBucket, _ := getServiceLevelBucket(size, serviceLevel)
	if quarkPodBucket != nil {
		for _, containerResource := range quarkPodBucket.ContainerResources {
			if containerResource.ContainerName == containername {
				request = containerResource.MemRequest
				limit = containerResource.MemLimit
				log.WithFields(log.Fields{
					"netappObj":                  netappObjName,
					"serviceLevel":               serviceLevel.Name,
					"bucket":                     quarkPodBucket.Name,
					containername + "MemRequest": request,
					containername + "MemLimit":   limit,
				}).Debug("Determined " + containername + " Container Memory request and limit")
				return request, limit, nil
			}
		}
	}
	return request, limit, fmt.Errorf("could not determine memory")
}

// Logic to extract NVRAM Size based on service-level/volume-size requested.
func (c *Controller) getNvramSize(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (int64, error) {
	var size int64
	if storagePool == nil && volume.Status.NvramSize != 0 {
		// Just use the nvram size stored in NetAppVolume CR if it's set
		return volume.Status.NvramSize, nil
	} else if storagePool != nil && storagePool.Status.NvramSize != 0 {
		return storagePool.Status.NvramSize, nil
	}

	_, name := GetNamespaceNameForNetappObj(volume, storagePool)
	if storagePool != nil {
		size = storagePool.Spec.Size.Value()
	} else {
		size = volume.Spec.Size.Value()
	}

	// Need to look it up in the serviceLevel
	quarkPodBucket, _ := getServiceLevelBucket(size, serviceLevel)
	if quarkPodBucket != nil {
		nvramSize := quarkPodBucket.NvramSize.Value()
		log.WithFields(log.Fields{
			"netappObj":    name,
			"serviceLevel": serviceLevel.Name,
			"bucket":       quarkPodBucket.Name,
			"nvramSize":    nvramSize,
		}).Debug("Determined Quark pod NVRAM size.")
		return nvramSize, nil
	}
	return 0, fmt.Errorf("could not determine quark pod NVRAM size")
}

// Extract DMS_CPUS value based on service-level/storage pool size requested
func (c *Controller) getDmsCPUs(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (int64, error) {
	var size int64

	_, name := GetNamespaceNameForNetappObj(volume, storagePool)
	if storagePool != nil {
		size = storagePool.Spec.Size.Value()
	} else {
		size = volume.Spec.Size.Value()
	}

	// Need to look it up in the serviceLevel
	quarkPodBucket, _ := getServiceLevelBucket(size, serviceLevel)
	if quarkPodBucket != nil {
		dmsCPUs := quarkPodBucket.DmsCPUs.Value()
		log.WithFields(log.Fields{
			"netappObj":    name,
			"serviceLevel": serviceLevel.Name,
			"bucket":       quarkPodBucket.Name,
			"dmsCPUs":      dmsCPUs,
		}).Debug("Determined Quark pod DMS CPUs.")
		return dmsCPUs, nil
	}
	return 0, fmt.Errorf("could not determine quark pod dms CPUs value")
}

// Logic to extract scheduler name based on service-level requested.
func (c *Controller) getSchedulerName(
	serviceLevel *quarkv1alpha1.NetAppServiceLevel,
) (string, error) {
	if len(serviceLevel.Spec.SchedulerName) > 0 {
		return serviceLevel.Spec.SchedulerName, nil
	}
	return quarkconfig.DefaultScheduler, nil
}

// Mark volume as offline if it is already online. When pod restarts, podrick will mark volume as online
func MakeNetAppVolumeConditionOffline(volume *quarkv1alpha1.NetAppVolume, msg string) []quarkv1alpha1.NetAppVolumeCondition {
	conditions := volume.Status.Conditions

	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.NetAppVolumeOnline && conditions[i].Status == corev1.ConditionTrue {
			timestamp := time.Now()
			conditions[i].LastHeartbeatTime = metav1.NewTime(timestamp)
			conditions[i].LastTransitionTime = metav1.NewTime(timestamp)
			conditions[i].Status = corev1.ConditionFalse
			conditions[i].Reason = quarkconfig.NetAppVolumeNotReady
			conditions[i].Message = msg
			log.WithFields(log.Fields{
				"volume":     volume.Name,
				"conditions": conditions,
			}).Debug("Updated conditions.")
			break
		}
	}
	return conditions
}

// Mark storagePool as offline if it is already online. When pod restarts, podrick will mark storagePool as online
func MakeNetAppStoragePoolConditionOffline(storagePool *quarkv1alpha1.NetAppStoragePool, msg string) []quarkv1alpha1.NetAppStoragePoolCondition {
	conditions := storagePool.Status.Conditions

	for i := range conditions {
		if conditions[i].Type == quarkv1alpha1.NetAppStoragePoolOnline && conditions[i].Status == corev1.ConditionTrue {
			timestamp := time.Now()
			conditions[i].LastHeartbeatTime = metav1.NewTime(timestamp)
			conditions[i].LastTransitionTime = metav1.NewTime(timestamp)
			conditions[i].Status = corev1.ConditionFalse
			conditions[i].Reason = quarkconfig.NetAppStoragePoolNotReady
			conditions[i].Message = msg
			log.WithFields(log.Fields{
				"storagePool": storagePool.Name,
				"conditions":  conditions,
			}).Debug("Updated conditions.")
			break
		}
	}
	return conditions
}

func existsVolResizeErrorCondition(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeResizeError) != nil
}

func existsVolResizedCondition(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeResized) != nil
}

func existsVolOnlineCondition(volume *quarkv1alpha1.NetAppVolume) bool {
	return util.GetNetAppVolumeCondition(volume, quarkv1alpha1.NetAppVolumeOnline) != nil
}

func existsSpOnlineCondition(storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	return util.GetNetAppStoragePoolCondition(storagePool, quarkv1alpha1.NetAppStoragePoolOnline) != nil
}

func existsSpAddressAllocationErrorCondition(
	storagePool *quarkv1alpha1.NetAppStoragePool,
) bool {
	return util.GetNetAppStoragePoolCondition(storagePool,
		quarkv1alpha1.NetAppStoragePoolAddressAllocationError) != nil
}

// create a lookup  table for container image tags
func (c *Controller) getImageTags(quarkVersion string) map[string]string {
	if quarkVersion == "" {
		lookupMap := map[string]string{
			quarkconfig.QuarkContainerDMAP:           c.DmapImageName,
			quarkconfig.QuarkContainerPodrick:        c.PodrickImageName,
			quarkconfig.QuarkContainerSECD:           c.SecdImageName,
			quarkconfig.QuarkContainerEtcd:           c.EtcdImageName,
			quarkconfig.QuarkContainerInit:           c.InitContainerImageName,
			quarkconfig.QuarkContainerLogging:        c.LoggingImageName,
			quarkconfig.QuarkContainerSupportability: c.SupportabilityImageName,
		}

		if c.EnableCCPD {
			lookupMap[quarkconfig.QuarkContainerCCPD] = c.CCPDImageName
		}
		return lookupMap
	}
	registryString := strings.Split(c.DmapImageName, "/quark_dmap")[0]
	specQuarkVersion := quarkVersion

	lookupMap := map[string]string{
		quarkconfig.QuarkContainerDMAP:           fmt.Sprintf("%s/quark_dmap:%s", registryString, specQuarkVersion),
		quarkconfig.QuarkContainerPodrick:        fmt.Sprintf("%s/podrick:%s", registryString, specQuarkVersion),
		quarkconfig.QuarkContainerSECD:           fmt.Sprintf("%s/quark_secd:%s", registryString, specQuarkVersion),
		quarkconfig.QuarkContainerEtcd:           c.EtcdImageName,
		quarkconfig.QuarkContainerInit:           c.InitContainerImageName,
		quarkconfig.QuarkContainerLogging:        c.LoggingImageName,
		quarkconfig.QuarkContainerSupportability: fmt.Sprintf("%s/quark_supportability:%s", registryString, specQuarkVersion),
	}
	if c.EnableCCPD {
		lookupMap[quarkconfig.QuarkContainerCCPD] = fmt.Sprintf("%s/quark_ccpd:%s", registryString, specQuarkVersion)
	}
	return lookupMap
}

// create an upgrade configmap for the given volume
func (c *Controller) createQuarkPodUpgradeConfigMap(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool, deployYaml string) error {
	var specQuarkVersion string
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	specQuarkVersion = storagePool.Spec.QuarkVersion
	cfgMap := &corev1.ConfigMap{
		ObjectMeta: metav1.ObjectMeta{
			Labels: c.getLabelForNetappObj(volume, storagePool),
		},
		Data: map[string]string{
			quarkconfig.QuarkPodDefinition: deployYaml,
			quarkconfig.QuarkVersion:       specQuarkVersion,
		},
	}
	cfgMap.Name = GetQuarkDeployName(namespace, name)
	_, err := c.KubeClientset.CoreV1().ConfigMaps(namespace).Create(context.TODO(),
		cfgMap, metav1.CreateOptions{})
	return err
}

// get the upgrade configmap for the given volume
func (c *Controller) getUpgradeConfigMap(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (*corev1.ConfigMap, error) {
	namespace, name := GetNamespaceNameForNetappObj(volume, storagePool)
	cfgMap, err := c.KubeClientset.CoreV1().ConfigMaps(namespace).Get(context.TODO(),
		GetQuarkDeployName(namespace, name), metav1.GetOptions{})
	if err != nil {
		log.WithFields(log.Fields{
			"netappObj": name,
			"error":     err,
		}).Error("Failed to get the upgrade configmap")
		return nil, err
	}
	return cfgMap, nil
}

// delete configmap
func (c *Controller) deleteConfigMap(cfgName string, cfgNameSpace string) error {
	err := c.KubeClientset.CoreV1().ConfigMaps(cfgNameSpace).Delete(context.TODO(),
		cfgName, metav1.DeleteOptions{})
	if err != nil && !errors.IsNotFound(err) {
		log.WithFields(log.Fields{
			"volume": cfgName,
			"error":  err,
		}).Error("Failed to delete the configmap")
		return err
	}
	return nil
}

// Function that returns max of two integer values
func Max(x, y int64) int64 {
	if x < y {
		return y
	}
	return x
}

// Delete the Quark pod that manages the ntapvol
func (c *Controller) deleteQuarkPod(volume *quarkv1alpha1.NetAppVolume) error {
	log.WithFields(log.Fields{
		"volume": volume.Name,
	}).Info("Deleting Quark Pod")

	podList, err := c.podsLister.Pods(volume.Namespace).List(GetQuarkPodSelector(volume, nil))
	if err != nil {
		return err
	}
	for _, pod := range podList {
		err := c.KubeClientset.CoreV1().Pods(volume.Namespace).Delete(context.TODO(),
			pod.Name, metav1.DeleteOptions{})
		return err
	}
	return nil
}

// Update an environment value
func setEnvVar(envs []corev1.EnvVar, key string, value string) []corev1.EnvVar {
	for i := range envs {
		if envs[i].Name == key {
			envs[i].Value = value
			break
		}
	}
	return envs
}

func isPre1_0VersionedVolume(volume *quarkv1alpha1.NetAppVolume) bool {
	return (volume.Status.QubeVersion == "" && volume.Status.Size.Value() != 0)
}

// getLabelForNetappObj returns label based on storage pool and volume
func (c *Controller) getLabelForNetappObj(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (label map[string]string) {
	if storagePool != nil {
		label = map[string]string{
			quarkconfig.StoragePoolLabel: storagePool.Name,
		}
	} else {
		label = map[string]string{
			quarkconfig.VolumeLabel: volume.Name,
		}
	}
	return label
}

// getSelectorNetappObj returns selector based on storage pool or netappvolume
func getSelectorNetappObj(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) (selector map[string]string) {
	if storagePool != nil {
		selector = map[string]string{
			quarkconfig.AppLabel:         "quark",
			quarkconfig.StoragePoolLabel: storagePool.Name,
		}
	} else {
		selector = map[string]string{
			quarkconfig.AppLabel:    "quark",
			quarkconfig.VolumeLabel: volume.Name,
		}
	}
	return selector
}

func getStorageLabelSelector(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) *metav1.LabelSelector {
	if storagePool != nil {
		return &metav1.LabelSelector{
			MatchLabels: map[string]string{
				quarkconfig.AppLabel:         "storage",
				quarkconfig.StoragePoolLabel: storagePool.Name,
			},
		}
	}
	return &metav1.LabelSelector{
		MatchLabels: map[string]string{
			quarkconfig.AppLabel:    "storage",
			quarkconfig.VolumeLabel: volume.Name,
		},
	}
}

func getQuarkLabelSelector(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) *metav1.LabelSelector {
	if storagePool != nil {
		return &metav1.LabelSelector{
			MatchLabels: map[string]string{
				quarkconfig.AppLabel:         "quark",
				quarkconfig.StoragePoolLabel: storagePool.Name,
			},
		}
	}
	return &metav1.LabelSelector{
		MatchLabels: map[string]string{
			quarkconfig.AppLabel:    "quark",
			quarkconfig.VolumeLabel: volume.Name,
		},
	}
}

func NetappVolumesByStoragePool(lister quarklisters.NetAppVolumeLister, namespace, storagePoolUUID string) ([]*quarkv1alpha1.NetAppVolume, error) {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.StoragePoolLabel] = storagePoolUUID
	volumes, err := lister.NetAppVolumes(namespace).List(labels.SelectorFromSet(selectorSet))
	if err != nil {
		return nil, err
	}
	return volumes, nil
}

func NetappClusterPeeringsByStoragePool(lister quarklisters.NetAppClusterPeeringLister, namespace, storagePoolUUID string) ([]*quarkv1alpha1.NetAppClusterPeering, error) {
	selectorSet := make(labels.Set)
	selectorSet[quarkconfig.StoragePoolLabel] = storagePoolUUID
	clusterPeerings, err := lister.NetAppClusterPeerings(namespace).List(labels.SelectorFromSet(selectorSet))
	if err != nil {
		return nil, err
	}
	return clusterPeerings, nil
}

func (c *Controller) isControlledByNetappObj(obj metav1.Object, volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) bool {
	if storagePool != nil {
		return metav1.IsControlledBy(obj, storagePool)
	} else {
		return metav1.IsControlledBy(obj, volume)
	}
}

func (c *Controller) updatePvcToStoragePool(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) error {
	// updating data pvc label and owner ref
	dataPvcs, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).List(c.getDataPvcSelector(volume, nil))
	if err != nil {
		log.WithFields(log.Fields{
			"volume": volume.Name,
			"error":  err,
		}).Error("Unable to get dataPvc")
		return err
	}
	for _, dataPvc := range dataPvcs {
		dataPvc.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = storagePool.Name
		dataPvc.ObjectMeta.OwnerReferences = MakeObjectMeta(volume, storagePool).OwnerReferences
		if _, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(volume.Namespace).Update(context.TODO(),
			dataPvc, metav1.UpdateOptions{}); err != nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"error":  err,
			}).Error("failed to update dataPvc")
			return err
		}
	}

	// update nvram pvc for diskpacking disabled config
	if !c.UseDiskPacking {
		nvramPvc, err := c.pvcsLister.PersistentVolumeClaims(volume.Namespace).Get(getNvramPvcName(volume.Namespace, volume.Name))
		if err != nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"error":  err,
			}).Error("Unable to get nvram pvc")
			return err
		}
		nvramPvc.ObjectMeta.Labels[quarkconfig.StoragePoolLabel] = storagePool.Name
		nvramPvc.ObjectMeta.OwnerReferences = MakeObjectMeta(volume, storagePool).OwnerReferences
		if _, err := c.KubeClientset.CoreV1().PersistentVolumeClaims(volume.Namespace).Update(context.TODO(),
			nvramPvc, metav1.UpdateOptions{}); err != nil {
			log.WithFields(log.Fields{
				"volume": volume.Name,
				"error":  err,
			}).Error("failed to update nvram pvc")
			return err
		}
	}
	return nil
}

// if volume is present check if it is old volume
// if SP is present check if this is coming from upgrade path and check if the volume is still not upgraded
//func (c *Controller) isPreStoragePoolVolume(volume *quarkv1alpha1.NetAppVolume, storagePool *quarkv1alpha1.NetAppStoragePool) bool {
//	if volume != nil {
//		return volume.Status.QubeVersion == "1.0" && volume.Status.Size.Value() != 0 && c.Cloud == quarkconfig.CloudGCP
//	} else {
//		volume, err := c.volumesLister.NetAppVolumes(storagePool.Namespace).Get(storagePool.Name)
//		if err != nil {
//			return false
//		}
//		return volume.Status.QubeVersion == "1.0" && volume.Status.Size.Value() != 0 && c.Cloud == quarkconfig.CloudGCP
//	}
//}

func (c *Controller) cleanUpEtcdKeysWithRetries(storagePool *quarkv1alpha1.NetAppStoragePool) error {
	var err error
	retryCount := 0
	etcdDeleteKeyPrefix := "/v" + quarkconfig.QuarkEtcdKeyVersion + "/StoragePool/" + storagePool.Name

	// delete the keys with retries
	for {
		ctx, cancel := context.WithTimeout(context.Background(), 60*time.Second)
		defer cancel()
		resp, err := c.EtcdClient.Delete(ctx, etcdDeleteKeyPrefix, etcdclient.WithPrefix(), etcdclient.WithPrevKV())

		if err == nil {
			// Print the key-value pairs for each deleted key
			for _, kv := range resp.PrevKvs {
				log.WithFields(log.Fields{
					"key":   string(kv.Key),
					"value": string(kv.Value),
				}).Debug("Successfully deleted.")
			}
			break
		} else if retryCount >= quarkconfig.EtcdRetries {
			log.WithFields(log.Fields{
				"contextTimeout":      60 * time.Second,
				"etcdDeleteKeyPrefix": etcdDeleteKeyPrefix,
				"err":                 err,
				"etcdRetryCount":      quarkconfig.EtcdRetries,
			}).Error("Timed out trying to reach etcd.")
			return err
		}

		// Exponential back-off to avoid overwhelming etcd cluster.
		delay := time.Duration(1<<uint(retryCount)) * time.Second
		time.Sleep(delay)

		retryCount++
	}

	//Remove Finalizer after key deletion
	if err = c.removeEtcdKeyFinalizer(
		nil, storagePool, quarkconfig.EtcdKeyFinalizer,
	); err != nil {
		return err
	}

	return nil
}

func (c *Controller) ensureEtcdKeyFinalizer(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	finalizerStr string,
) (*quarkv1alpha1.NetAppStoragePool, error) {
	var err error

	if storagePool.ObjectMeta.DeletionTimestamp != nil {
		return storagePool, fmt.Errorf("Cannot add a finalizer to a deleted storagepool: %s", storagePool.Name)
	}

	volume, storagePool, err = AddFinalizer(c.QuarkClientset, volume,
		storagePool, finalizerStr)
	if err != nil {
		return nil, err
	}

	return storagePool, nil
}

func (c *Controller) removeEtcdKeyFinalizer(
	volume *quarkv1alpha1.NetAppVolume,
	storagePool *quarkv1alpha1.NetAppStoragePool,
	finalizerStr string,
) error {
	var err error

	if _, err = RemoveFinalizer(c.QuarkClientset, volume, storagePool,
		finalizerStr); err != nil {
		return err
	}
	log.WithFields(log.Fields{
		"storagePool":     storagePool.Name,
		"resourceVersion": storagePool.ResourceVersion,
	}).Debug("Removed etcdKey finalizer.")

	return nil
}
